package org.jetbrains.kotlinx.statistics.math3.analysis.interpolation

import org.jetbrains.kotlinx.statistics.math3.analysis.polynomials.PolynomialFunction
import org.jetbrains.kotlinx.statistics.math3.analysis.polynomials.PolynomialSplineFunction
import org.jetbrains.kotlinx.statistics.math3.util.MathArrays

class SplineInterpolator {
    /**
     * Computes an interpolating function for the data set.
     * @param x the arguments for the interpolation points
     * @param y the values for the interpolation points
     * @return a function which interpolates the data set
     * @throws DimensionMismatchException if `x` and `y`
     * @throws DimensionMismatchException if `x` and `y`
     * have different sizes.
     * @throws org.apache.commons.math3.exception.NonMonotonicSequenceException
     * if `x` is not sorted in strict increasing order.
     * @throws NumberIsTooSmallException if the size of `x` is smaller
     * than 3.
     */
    fun interpolate(x: DoubleArray, y: DoubleArray): PolynomialSplineFunction {
        if (x.size != y.size) {
            error("Dimension mismatch ${x.size} != ${y.size}")
        }
        if (x.size < 3) {
            error("Too small value of points number: ${x.size} < 3")
        }
        // Number of intervals.  The number of data points is n + 1.
        val n = x.size - 1
        MathArrays.checkOrder(x)
        // Differences between knot points
        val h = DoubleArray(n)
        for (i in 0 until n) {
            h[i] = x[i + 1] - x[i]
        }
        val mu = DoubleArray(n)
        val z = DoubleArray(n + 1)
        mu[0] = 0.0
        z[0] = 0.0

        for (i in 1 until n) {
            var g = 2.0 * (x[i + 1] - x[i - 1]) - h[i - 1] * mu[i - 1]
            mu[i] = h[i] / g
            z[i] =
                (3.0 * (y[i + 1] * h[i - 1] - y[i] * (x[i + 1] - x[i - 1]) + y[i - 1] * h[i]) /
                        (h[i - 1] * h[i]) - h[i - 1] * z[i - 1]) / g
        }
        // cubic spline coefficients --  b is linear, c quadratic, d is cubic (original y's are constants)
        val b = DoubleArray(n)
        val c = DoubleArray(n + 1)
        val d = DoubleArray(n)
        z[n] = 0.0
        c[n] = 0.0
        for (j in n - 1 downTo 0) {
            c[j] = z[j] - mu[j] * c[j + 1]
            b[j] = (y[j + 1] - y[j]) / h[j] - h[j] * (c[j + 1] + 2.0 * c[j]) / 3.0
            d[j] = (c[j + 1] - c[j]) / (3.0 * h[j])
        }
        val polynomials: Array<PolynomialFunction?> = arrayOfNulls<PolynomialFunction>(n)
        val coefficients = DoubleArray(4)
        for (i in 0 until n) {
            coefficients[0] = y[i]
            coefficients[1] = b[i]
            coefficients[2] = c[i]
            coefficients[3] = d[i]
            polynomials[i] = PolynomialFunction(coefficients)
        }
        return PolynomialSplineFunction(x, polynomials)
    }
}

