/*
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.jetbrains.kotlinx.statistics.math3.distribution

import org.jetbrains.kotlinx.statistics.math3.analysis.UnivariateFunction
import org.jetbrains.kotlinx.statistics.math3.analysis.solvers.UnivariateSolverUtils
import org.jetbrains.kotlinx.statistics.math3.random.RandomGenerator

/**
 * Base class for probability distributions on the reals.
 * Default implementations are provided for some of the methods
 * that do not vary from distribution to distribution.
 *
 * @since 3.0
 */
abstract class AbstractRealDistribution : RealDistribution {
    /**
     * RandomData instance used to generate samples from the distribution.
     *//*
    @Deprecated(
        """As of 3.1, to be removed in 4.0. Please use the
       {@link #random} instance variable instead."""
    )
    protected var randomData: RandomDataImpl =
        RandomDataImpl()*/

    /**
     * RNG instance used to generate samples from the distribution.
     * @since 3.1
     */
    protected val random: RandomGenerator
    /**
     * Returns the solver absolute accuracy for inverse cumulative computation.
     * You can override this method in order to use a Brent solver with an
     * absolute accuracy different from the default.
     *
     * @return the maximum absolute error in inverse cumulative probability estimates
     */
    /** Solver absolute accuracy for inverse cumulative computation  */
    protected open val solverAbsoluteAccuracy = SOLVER_DEFAULT_ABSOLUTE_ACCURACY

   /* @Deprecated(
        """As of 3.1, to be removed in 4.0. Please use
      {@link #AbstractRealDistribution(RandomGenerator)} instead."""
    )
    protected constructor() {
        // Legacy users are only allowed to access the deprecated "randomData".
        // New users are forbidden to use this constructor.
        random = null
    }*/

    /**
     * @param rng Random number generator.
     * @since 3.1
     */
    protected constructor(rng: RandomGenerator) {
        random = rng
    }

    /**
     * {@inheritDoc}
     *
     * The default implementation uses the identity
     *
     * `P(x0 < X <= x1) = P(X <= x1) - P(X <= x0)`
     *
     */
    /*@Deprecated(
        """As of 3.1 (to be removed in 4.0). Please use
      {@link #probability(double,double)} instead."""
    )
    *//*@Throws(
        org.apache.commons.math3.exception.NumberIsTooLargeException::class
    )*//*
    override fun cumulativeProbability(x0: Double, x1: Double): Double {
        return probability(x0, x1)
    }*/

    /**
     * For a random variable `X` whose values are distributed according
     * to this distribution, this method returns `P(x0 < X <= x1)`.
     *
     * @param x0 Lower bound (excluded).
     * @param x1 Upper bound (included).
     * @return the probability that a random variable with this distribution
     * takes a value between `x0` and `x1`, excluding the lower
     * and including the upper endpoint.
     * @throws NumberIsTooLargeException if `x0 > x1`.
     *
     * The default implementation uses the identity
     * `P(x0 < X <= x1) = P(X <= x1) - P(X <= x0)`
     *
     * @since 3.1
     */
    open fun probability(
        x0: Double,
        x1: Double
    ): Double {
        if (x0 > x1) {
            TODO()
            /*throw org.apache.commons.math3.exception.NumberIsTooLargeException(
                org.apache.commons.math3.exception.util.LocalizedFormats.LOWER_ENDPOINT_ABOVE_UPPER_ENDPOINT,
                x0, x1, true
            )*/
        }
        return cumulativeProbability(x1) - cumulativeProbability(x0)
    }

    /**
     * {@inheritDoc}
     *
     * The default implementation returns
     *
     *  * [.getSupportLowerBound] for `p = 0`,
     *  * [.getSupportUpperBound] for `p = 1`.
     *
     */
   // @Throws(org.apache.commons.math3.exception.OutOfRangeException::class)
    override fun inverseCumulativeProbability(p: Double): Double {
        /*
         * IMPLEMENTATION NOTES
         * --------------------
         * Where applicable, use is made of the one-sided Chebyshev inequality
         * to bracket the root. This inequality states that
         * P(X - mu >= k * sig) <= 1 / (1 + k^2),
         * mu: mean, sig: standard deviation. Equivalently
         * 1 - P(X < mu + k * sig) <= 1 / (1 + k^2),
         * F(mu + k * sig) >= k^2 / (1 + k^2).
         *
         * For k = sqrt(p / (1 - p)), we find
         * F(mu + k * sig) >= p,
         * and (mu + k * sig) is an upper-bound for the root.
         *
         * Then, introducing Y = -X, mean(Y) = -mu, sd(Y) = sig, and
         * P(Y >= -mu + k * sig) <= 1 / (1 + k^2),
         * P(-X >= -mu + k * sig) <= 1 / (1 + k^2),
         * P(X <= mu - k * sig) <= 1 / (1 + k^2),
         * F(mu - k * sig) <= 1 / (1 + k^2).
         *
         * For k = sqrt((1 - p) / p), we find
         * F(mu - k * sig) <= p,
         * and (mu - k * sig) is a lower-bound for the root.
         *
         * In cases where the Chebyshev inequality does not apply, geometric
         * progressions 1, 2, 4, ... and -1, -2, -4, ... are used to bracket
         * the root.
         */
        if (p < 0.0 || p > 1.0) {
            TODO() //throw org.apache.commons.math3.exception.OutOfRangeException(p, 0, 1)
        }
        var lowerBound: Double = supportLowerBound
        if (p == 0.0) {
            return lowerBound
        }
        var upperBound: Double = supportUpperBound
        if (p == 1.0) {
            return upperBound
        }
        val mu: Double = numericalMean
        val sig: Double = kotlin.math.sqrt(numericalVariance)
        val chebyshevApplies: Boolean = !(mu.isInfinite() || mu.isNaN() ||
                sig.isInfinite() || sig.isNaN())
        if (lowerBound == Double.NEGATIVE_INFINITY) {
            if (chebyshevApplies) {
                lowerBound = mu - sig * kotlin.math.sqrt((1.0 - p) / p)
            } else {
                lowerBound = -1.0
                while (cumulativeProbability(lowerBound) >= p) {
                    lowerBound *= 2.0
                }
            }
        }
        if (upperBound == Double.POSITIVE_INFINITY) {
            if (chebyshevApplies) {
                upperBound = mu + sig * kotlin.math.sqrt(p / (1.0 - p))
            } else {
                upperBound = 1.0
                while (cumulativeProbability(upperBound) < p) {
                    upperBound *= 2.0
                }
            }
        }
        val toSolve: UnivariateFunction =
            object : UnivariateFunction {
                /** {@inheritDoc}  */
                override fun value(x: Double): Double {
                    return cumulativeProbability(x) - p
                }
            }
        val x: Double = UnivariateSolverUtils.solve(
            toSolve,
            lowerBound,
            upperBound,
            solverAbsoluteAccuracy
        )
        if (!isSupportConnected) {
            /* Test for plateau. */
            val dx = solverAbsoluteAccuracy
            if (x - dx >= supportLowerBound) {
                val px = cumulativeProbability(x)
                if (cumulativeProbability(x - dx) == px) {
                    upperBound = x
                    while (upperBound - lowerBound > dx) {
                        val midPoint = 0.5 * (lowerBound + upperBound)
                        if (cumulativeProbability(midPoint) < px) {
                            lowerBound = midPoint
                        } else {
                            upperBound = midPoint
                        }
                    }
                    return upperBound
                }
            }
        }
        return x
    }

    /** {@inheritDoc}  */
    override fun reseedRandomGenerator(seed: Long) {
        random!!.setSeed(seed)
       // randomData.reSeed(seed)
    }

    /**
     * {@inheritDoc}
     *
     * The default implementation uses the
     * [
 * inversion method.
](http://en.wikipedia.org/wiki/Inverse_transform_sampling) *
     */
    override fun sample(): Double {
        return inverseCumulativeProbability(random.nextDouble())
    }

    /**
     * {@inheritDoc}
     *
     * The default implementation generates the sample by calling
     * [.sample] in a loop.
     */
    override fun sample(sampleSize: Int): DoubleArray {
        if (sampleSize <= 0) {
            TODO() /*throw org.apache.commons.math3.exception.NotStrictlyPositiveException(
                org.apache.commons.math3.exception.util.LocalizedFormats.NUMBER_OF_SAMPLES,
                sampleSize
            )*/
        }
        val out = DoubleArray(sampleSize)
        for (i in 0 until sampleSize) {
            out[i] = sample()
        }
        return out
    }

    /**
     * {@inheritDoc}
     *
     * @return zero.
     * @since 3.1
     */
    override fun probability(x: Double): Double {
        return 0.0
    }

    /**
     * Returns the natural logarithm of the probability density function (PDF) of this distribution
     * evaluated at the specified point `x`. In general, the PDF is the derivative of the
     * [CDF][.cumulativeProbability]. If the derivative does not exist at `x`,
     * then an appropriate replacement should be returned, e.g. `Double.POSITIVE_INFINITY`,
     * `Double.NaN`, or the limit inferior or limit superior of the difference quotient. Note
     * that due to the floating point precision and under/overflow issues, this method will for some
     * distributions be more precise and faster than computing the logarithm of
     * [.density]. The default implementation simply computes the logarithm of
     * `density(x)`.
     *
     * @param x the point at which the PDF is evaluated
     * @return the logarithm of the value of the probability density function at point `x`
     */
    open fun lnDensity(x: Double): Double {
        return kotlin.math.ln(density(x))
    }

    companion object {
        /** Default accuracy.  */
        const val SOLVER_DEFAULT_ABSOLUTE_ACCURACY = 1e-6

        /** Serializable version identifier  */
        private const val serialVersionUID = -38038050983108802L
    }
}
