/*
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.jetbrains.kotlinx.statistics.math3.distribution

import kotlin.jvm.JvmOverloads
import org.jetbrains.kotlinx.statistics.math3.random.RandomGenerator
import org.jetbrains.kotlinx.statistics.math3.random.Well19937c
import org.jetbrains.kotlinx.statistics.math3.special.Erf

/**
 * Implementation of the normal (gaussian) distribution.
 *
 * @see [Normal distribution
 * @see [Normal distribution
](http://mathworld.wolfram.com/NormalDistribution.html)](http://en.wikipedia.org/wiki/Normal_distribution) */
class NormalDistribution @JvmOverloads constructor(
    rng: RandomGenerator,
    mean: Double,
    sd: Double,
    inverseCumAccuracy: Double = DEFAULT_INVERSE_ABSOLUTE_ACCURACY
) : AbstractRealDistribution(rng) {
    /**
     * Access the mean.
     *
     * @return the mean for this distribution.
     */
    /** Mean of this distribution.  */
    val mean: Double
    /**
     * Access the standard deviation.
     *
     * @return the standard deviation for this distribution.
     */
    /** Standard deviation of this distribution.  */
    val standardDeviation: Double

    /** The value of `log(sd) + 0.5*log(2*pi)` stored for faster computation.  */
    private val logStandardDeviationPlusHalfLog2Pi: Double
    /** {@inheritDoc}  */
    /** Inverse cumulative probability accuracy.  */
    protected override val solverAbsoluteAccuracy: Double
    /**
     * Create a normal distribution using the given mean, standard deviation and
     * inverse cumulative distribution accuracy.
     *
     *
     * **Note:** this constructor will implicitly create an instance of
     * [Well19937c] as random generator to be used for sampling only (see
     * [.sample] and [.sample]). In case no sampling is
     * needed for the created distribution, it is advised to pass `null`
     * as random generator via the appropriate constructors to avoid the
     * additional initialisation overhead.
     *
     * @param mean Mean for this distribution.
     * @param sd Standard deviation for this distribution.
     * @param inverseCumAccuracy Inverse cumulative probability accuracy.
     * @throws NotStrictlyPositiveException if `sd <= 0`.
     * @since 2.1
     */
    /**
     * Create a normal distribution using the given mean and standard deviation.
     *
     *
     * **Note:** this constructor will implicitly create an instance of
     * [Well19937c] as random generator to be used for sampling only (see
     * [.sample] and [.sample]). In case no sampling is
     * needed for the created distribution, it is advised to pass `null`
     * as random generator via the appropriate constructors to avoid the
     * additional initialisation overhead.
     *
     * @param mean Mean for this distribution.
     * @param sd Standard deviation for this distribution.
     * @throws NotStrictlyPositiveException if `sd <= 0`.
     */
    /**
     * Create a normal distribution with mean equal to zero and standard
     * deviation equal to one.
     *
     *
     * **Note:** this constructor will implicitly create an instance of
     * [Well19937c] as random generator to be used for sampling only (see
     * [.sample] and [.sample]). In case no sampling is
     * needed for the created distribution, it is advised to pass `null`
     * as random generator via the appropriate constructors to avoid the
     * additional initialisation overhead.
     */
    @JvmOverloads
    constructor(
        mean: Double = 0.0,
        sd: Double = 1.0,
        inverseCumAccuracy: Double = DEFAULT_INVERSE_ABSOLUTE_ACCURACY
    ) : this(Well19937c(), mean, sd, inverseCumAccuracy)
    /**
     * Creates a normal distribution.
     *
     * @param rng Random number generator.
     * @param mean Mean for this distribution.
     * @param sd Standard deviation for this distribution.
     * @param inverseCumAccuracy Inverse cumulative probability accuracy.
     * @throws NotStrictlyPositiveException if `sd <= 0`.
     * @since 3.1
     */
    /**
     * Creates a normal distribution.
     *
     * @param rng Random number generator.
     * @param mean Mean for this distribution.
     * @param sd Standard deviation for this distribution.
     * @throws NotStrictlyPositiveException if `sd <= 0`.
     * @since 3.3
     */
    init {
        if (sd <= 0) {
            TODO()
            /*throw org.apache.commons.math3.exception.NotStrictlyPositiveException(
                org.apache.commons.math3.exception.util.LocalizedFormats.STANDARD_DEVIATION,
                sd
            )*/
        }
        this.mean = mean
        standardDeviation = sd
        logStandardDeviationPlusHalfLog2Pi =
            kotlin.math.ln(sd) + 0.5 * kotlin.math.ln(2 * kotlin.math.PI)
        solverAbsoluteAccuracy = inverseCumAccuracy
    }

    /** {@inheritDoc}  */
    override fun density(x: Double): Double {
        return kotlin.math.exp(lnDensity(x))
    }

    /** {@inheritDoc}  */
    override fun lnDensity(x: Double): Double {
        val x0 = x - mean
        val x1 = x0 / standardDeviation
        return -0.5 * x1 * x1 - logStandardDeviationPlusHalfLog2Pi
    }

    /**
     * {@inheritDoc}
     *
     * If `x` is more than 40 standard deviations from the mean, 0 or 1
     * is returned, as in these cases the actual value is within
     * `Double.MIN_VALUE` of 0 or 1.
     */
    override fun cumulativeProbability(x: Double): Double {
        val dev = x - mean
        return if (kotlin.math.abs(dev) > 40 * standardDeviation) {
            if (dev < 0) 0.0 else 1.0
        } else 0.5 * Erf.erfc(-dev / (standardDeviation * SQRT2))
    }


    /** {@inheritDoc}
     * @since 3.2
     */
    //@Throws(org.apache.commons.math3.exception.OutOfRangeException::class)
    override fun inverseCumulativeProbability(p: Double): Double {
        if (p < 0.0 || p > 1.0) {
            TODO()//throw org.apache.commons.math3.exception.OutOfRangeException(p, 0, 1)
        }
        return mean + standardDeviation * SQRT2 * Erf.erfInv(2 * p - 1)
    }

    /**
     * {@inheritDoc}
     *
     *//*
    @Deprecated("See {@link RealDistribution#cumulativeProbability(double,double)}")
    @Throws(org.apache.commons.math3.exception.NumberIsTooLargeException::class)
    override fun cumulativeProbability(x0: Double, x1: Double): Double {
        return probability(x0, x1)
    }
*/
    /** {@inheritDoc}  */
    //@Throws(org.apache.commons.math3.exception.NumberIsTooLargeException::class)
    override fun probability(
        x0: Double,
        x1: Double
    ): Double {
        if (x0 > x1) {
            TODO()
            /*throw org.apache.commons.math3.exception.NumberIsTooLargeException(
                org.apache.commons.math3.exception.util.LocalizedFormats.LOWER_ENDPOINT_ABOVE_UPPER_ENDPOINT,
                x0, x1, true
            )*/
        }
        val denom = standardDeviation * SQRT2
        val v0 = (x0 - mean) / denom
        val v1 = (x1 - mean) / denom
        return 0.5 * Erf.erf(v0, v1)
    }

    override val numericalMean: Double
        /**
         * {@inheritDoc}
         *
         * For mean parameter `mu`, the mean is `mu`.
         */
        get() = mean
    override val numericalVariance: Double
        /**
         * {@inheritDoc}
         *
         * For standard deviation parameter `s`, the variance is `s^2`.
         */
        get() {
            val s = standardDeviation
            return s * s
        }
    override val supportLowerBound: Double
        /**
         * {@inheritDoc}
         *
         * The lower bound of the support is always negative infinity
         * no matter the parameters.
         *
         * @return lower bound of the support (always
         * `Double.NEGATIVE_INFINITY`)
         */
        get() = Double.NEGATIVE_INFINITY
    override val supportUpperBound: Double
        /**
         * {@inheritDoc}
         *
         * The upper bound of the support is always positive infinity
         * no matter the parameters.
         *
         * @return upper bound of the support (always
         * `Double.POSITIVE_INFINITY`)
         */
        get() = Double.POSITIVE_INFINITY

    override val isSupportConnected: Boolean
        /**
         * {@inheritDoc}
         *
         * The support of this distribution is connected.
         *
         * @return `true`
         */
        get() = true

    /** {@inheritDoc}  */
    override fun sample(): Double {
        return standardDeviation * random.nextGaussian() + mean
    }

    companion object {
        /**
         * Default inverse cumulative probability accuracy.
         * @since 2.1
         */
        const val DEFAULT_INVERSE_ABSOLUTE_ACCURACY = 1e-9

        /** Serializable version identifier.  */
        private const val serialVersionUID = 8589540077390120676L

        /** (2)  */
        private val SQRT2: Double = kotlin.math.sqrt(2.0)
    }
}
