/*
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.jetbrains.kotlinx.statistics.math3.distribution

import org.jetbrains.kotlinx.statistics.math3.random.RandomGenerator
import org.jetbrains.kotlinx.statistics.math3.random.Well19937c
import org.jetbrains.kotlinx.statistics.math3.special.Beta
import org.jetbrains.kotlinx.statistics.math3.special.Gamma
import kotlin.jvm.JvmOverloads
import kotlin.math.PI
import kotlin.math.exp
import kotlin.math.ln

/**
 * Implementation of Student's t-distribution.
 *
 * @see "[&apos;Student's t-distribution
 * @see "[Student's t-distribution
](http://mathworld.wolfram.com/Studentst-Distribution.html)](http://en.wikipedia.org/wiki/Student&apos;s_t-distribution) */
class TDistribution @JvmOverloads constructor(
    rng: RandomGenerator,
    degreesOfFreedom: Double,
    inverseCumAccuracy: Double = DEFAULT_INVERSE_ABSOLUTE_ACCURACY
) :AbstractRealDistribution(rng) {
    /**
     * Access the degrees of freedom.
     *
     * @return the degrees of freedom.
     */
    /** The degrees of freedom.  */
    val degreesOfFreedom: Double
    /** {@inheritDoc}  */
    /** Inverse cumulative probability accuracy.  */
    protected override val solverAbsoluteAccuracy: Double

    /** Static computation factor based on degreesOfFreedom.  */
    private val factor: Double
    /**
     * Create a t distribution using the given degrees of freedom and the
     * specified inverse cumulative probability absolute accuracy.
     *
     *
     * **Note:** this constructor will implicitly create an instance of
     * [Well19937c] as random generator to be used for sampling only (see
     * [.sample] and [.sample]). In case no sampling is
     * needed for the created distribution, it is advised to pass `null`
     * as random generator via the appropriate constructors to avoid the
     * additional initialisation overhead.
     *
     * @param degreesOfFreedom Degrees of freedom.
     * @param inverseCumAccuracy the maximum absolute error in inverse
     * cumulative probability estimates
     * (defaults to [.DEFAULT_INVERSE_ABSOLUTE_ACCURACY]).
     * @throws NotStrictlyPositiveException if `degreesOfFreedom <= 0`
     * @since 2.1
     */
    /**
     * Create a t distribution using the given degrees of freedom.
     *
     *
     * **Note:** this constructor will implicitly create an instance of
     * [Well19937c] as random generator to be used for sampling only (see
     * [.sample] and [.sample]). In case no sampling is
     * needed for the created distribution, it is advised to pass `null`
     * as random generator via the appropriate constructors to avoid the
     * additional initialisation overhead.
     *
     * @param degreesOfFreedom Degrees of freedom.
     * @throws NotStrictlyPositiveException if `degreesOfFreedom <= 0`
     */
    @JvmOverloads
    constructor(
        degreesOfFreedom: Double,
        inverseCumAccuracy: Double = DEFAULT_INVERSE_ABSOLUTE_ACCURACY
    ) : this(Well19937c(), degreesOfFreedom, inverseCumAccuracy)
    /**
     * Creates a t distribution.
     *
     * @param rng Random number generator.
     * @param degreesOfFreedom Degrees of freedom.
     * @param inverseCumAccuracy the maximum absolute error in inverse
     * cumulative probability estimates
     * (defaults to [.DEFAULT_INVERSE_ABSOLUTE_ACCURACY]).
     * @throws NotStrictlyPositiveException if `degreesOfFreedom <= 0`
     * @since 3.1
     */
    /**
     * Creates a t distribution.
     *
     * @param rng Random number generator.
     * @param degreesOfFreedom Degrees of freedom.
     * @throws NotStrictlyPositiveException if `degreesOfFreedom <= 0`
     * @since 3.3
     */
    init {
        if (degreesOfFreedom <= 0) {
            TODO()/*throw org.apache.commons.math3.exception.NotStrictlyPositiveException(
                org.apache.commons.math3.exception.util.LocalizedFormats.DEGREES_OF_FREEDOM,
                degreesOfFreedom
            )*/
        }
        this.degreesOfFreedom = degreesOfFreedom
        solverAbsoluteAccuracy = inverseCumAccuracy
        val n = degreesOfFreedom
        val nPlus1Over2 = (n + 1) / 2
        factor =
            Gamma.lnGamma(nPlus1Over2) - 0.5 * (ln(
                PI
            ) + ln(n)) - Gamma.lnGamma(n / 2)
    }

    /** {@inheritDoc}  */
    override fun density(x: Double): Double {
        return exp(lnDensity(x))
    }

    /** {@inheritDoc}  */
    override fun lnDensity(x: Double): Double {
        val n = degreesOfFreedom
        val nPlus1Over2 = (n + 1) / 2
        return factor - nPlus1Over2 * ln(1 + x * x / n)
    }

    /** {@inheritDoc}  */
    override fun cumulativeProbability(x: Double): Double {
        val ret: Double = if (x == 0.0) {
            0.5
        } else {
            val t: Double = Beta.regularizedBeta(
                degreesOfFreedom / (degreesOfFreedom + x * x),
                0.5 * degreesOfFreedom,
                0.5
            )
            if (x < 0.0) {
                0.5 * t
            } else {
                1.0 - 0.5 * t
            }
        }
        return ret
    }


    override val numericalMean: Double
        /**
         * {@inheritDoc}
         *
         * For degrees of freedom parameter `df`, the mean is
         *
         *  * if `df > 1` then `0`,
         *  * else undefined (`Double.NaN`).
         *
         */
        get() {
            val df = degreesOfFreedom
            return if (df > 1) {
                0.0
            } else Double.NaN
        }
    override val numericalVariance: Double
        /**
         * {@inheritDoc}
         *
         * For degrees of freedom parameter `df`, the variance is
         *
         *  * if `df > 2` then `df / (df - 2)`,
         *  * if `1 < df <= 2` then positive infinity
         * (`Double.POSITIVE_INFINITY`),
         *  * else undefined (`Double.NaN`).
         *
         */
        get() {
            val df = degreesOfFreedom
            if (df > 2) {
                return df / (df - 2)
            }
            return if (df > 1 && df <= 2) {
                Double.POSITIVE_INFINITY
            } else Double.NaN
        }
    override val supportLowerBound: Double
        /**
         * {@inheritDoc}
         *
         * The lower bound of the support is always negative infinity no matter the
         * parameters.
         *
         * @return lower bound of the support (always
         * `Double.NEGATIVE_INFINITY`)
         */
        get() = Double.NEGATIVE_INFINITY
    override val supportUpperBound: Double
        /**
         * {@inheritDoc}
         *
         * The upper bound of the support is always positive infinity no matter the
         * parameters.
         *
         * @return upper bound of the support (always
         * `Double.POSITIVE_INFINITY`)
         */
        get() = Double.POSITIVE_INFINITY
    val isSupportLowerBoundInclusive: Boolean
        /** {@inheritDoc}  */
        get() = false
    val isSupportUpperBoundInclusive: Boolean
        /** {@inheritDoc}  */
        get() = false
    override val isSupportConnected: Boolean
        /**
         * {@inheritDoc}
         *
         * The support of this distribution is connected.
         *
         * @return `true`
         */
        get() = true

    companion object {
        /**
         * Default inverse cumulative probability accuracy.
         * @since 2.1
         */
        const val DEFAULT_INVERSE_ABSOLUTE_ACCURACY = 1e-9

        /** Serializable version identifier  */
        private const val serialVersionUID = -5852615386664158222L

        fun density(x: Double, degreesOfFreedom: Double): Double {
            return TDistribution(degreesOfFreedom).density(x)
        }

        fun cumulativeProbability(x: Double, degreesOfFreedom: Double): Double {
            val ret: Double = if (x == 0.0) {
                0.5
            } else {
                val t: Double = Beta.regularizedBeta(
                    degreesOfFreedom / (degreesOfFreedom + x * x),
                    0.5 * degreesOfFreedom,
                    0.5
                )
                if (x < 0.0) {
                    0.5 * t
                } else {
                    1.0 - 0.5 * t
                }
            }
            return ret
        }
    }
}
