/*
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.jetbrains.kotlinx.statistics.math3.random

import kotlin.math.*

/** Base class for random number generators that generates bits streams.
 *
 * @since 2.0
 */
abstract class BitsStreamGenerator : RandomGenerator {
    /** Next gaussian.  */
    private var nextGaussian: Double

    /**
     * Creates a new random number generator.
     */
    init {
        nextGaussian = Double.NaN
    }

    /** {@inheritDoc}  */
    abstract override fun setSeed(seed: Int)

    /** {@inheritDoc}  */
    abstract override fun setSeed(seed: IntArray?)

    /** {@inheritDoc}  */
    abstract override fun setSeed(seed: Long)

    /** Generate next pseudorandom number.
     *
     * This method is the core generation algorithm. It is used by all the
     * public generation methods for the various primitive types [ ][.nextBoolean], [.nextBytes], [.nextDouble],
     * [.nextFloat], [.nextGaussian], [.nextInt],
     * [.next] and [.nextLong].
     * @param bits number of random bits to produce
     * @return random bits generated
     */
    protected abstract fun next(bits: Int): Int

    /** {@inheritDoc}  */
    override fun nextBoolean(): Boolean {
        return next(1) != 0
    }

    /** {@inheritDoc}  */
    override fun nextDouble(): Double {
        val high = next(26).toLong() shl 26
        val low = next(26)
        return (high or low.toLong()).toDouble() * 2.0.pow(-52)
    }

    /** {@inheritDoc}  */
    override fun nextFloat(): Float {
        return next(23).toFloat() * 2.0.pow(-23).toFloat()
    }

    /** {@inheritDoc}  */
    override fun nextGaussian(): Double {
        val random: Double
        if (nextGaussian.isNaN()) {
            // generate a new pair of gaussian numbers
            val x = nextDouble()
            val y = nextDouble()
            val alpha: Double = 2 * PI * x
            val r: Double = sqrt(-2 * ln(y))
            random = r * cos(alpha)
            nextGaussian = r * sin(alpha)
        } else {
            // use the second element of the pair already generated
            random = nextGaussian
            nextGaussian = Double.NaN
        }
        return random
    }

    /** {@inheritDoc}  */
    override fun nextInt(): Int {
        return next(32)
    }

    /**
     * {@inheritDoc}
     *
     * This default implementation is copied from Apache Harmony
     * java.util.Random (r929253).
     *
     *
     * Implementation notes:
     *  * If n is a power of 2, this method returns
     * `(int) ((n * (long) next(31)) >> 31)`.
     *
     *  * If n is not a power of 2, what is returned is `next(31) % n`
     * with `next(31)` values rejected (i.e. regenerated) until a
     * value that is larger than the remainder of `Integer.MAX_VALUE / n`
     * is generated. Rejection of this initial segment is necessary to ensure
     * a uniform distribution.
     */
    //@Throws(java.lang.IllegalArgumentException::class)
    override fun nextInt(n: Int): Int {
        if (n > 0) {
            if (n and -n == n) {
                return (n * next(31).toLong() shr 31).toInt()
            }
            var bits: Int
            var `val`: Int
            do {
                bits = next(31)
                `val` = bits % n
            } while (bits - `val` + (n - 1) < 0)
            return `val`
        }
        throw TODO()// NotStrictlyPositiveException(n)
    }

    /** {@inheritDoc}  */
    override fun nextLong(): Long {
        val high = next(32).toLong() shl 32
        val low = next(32).toLong() and 0xffffffffL
        return high or low
    }

    /**
     * Returns a pseudorandom, uniformly distributed `long` value
     * between 0 (inclusive) and the specified value (exclusive), drawn from
     * this random number generator's sequence.
     *
     * @param n the bound on the random number to be returned.  Must be
     * positive.
     * @return  a pseudorandom, uniformly distributed `long`
     * value between 0 (inclusive) and n (exclusive).
     * @throws IllegalArgumentException  if n is not positive.
     */
    //@Throws(java.lang.IllegalArgumentException::class)
    fun nextLong(n: Long): Long {
        if (n > 0) {
            var bits: Long
            var `val`: Long
            do {
                bits = next(31).toLong() shl 32
                bits = bits or (next(32).toLong() and 0xffffffffL)
                `val` = bits % n
            } while (bits - `val` + (n - 1) < 0)
            return `val`
        }
        throw TODO()// NotStrictlyPositiveException(n)
    }

    /**
     * Clears the cache used by the default implementation of
     * [.nextGaussian].
     */
    fun clear() {
        nextGaussian = Double.NaN
    }

    /**
     * Generates random bytes and places them into a user-supplied array.
     *
     *
     *
     * The array is filled with bytes extracted from random integers.
     * This implies that the number of random bytes generated may be larger than
     * the length of the byte array.
     *
     *
     * @param bytes Array in which to put the generated bytes. Cannot be `null`.
     */
    override fun nextBytes(bytes: ByteArray) {
        nextBytesFill(bytes, 0, bytes.size)
    }

    /**
     * Generates random bytes and places them into a user-supplied array.
     *
     *
     *
     * The array is filled with bytes extracted from random integers.
     * This implies that the number of random bytes generated may be larger than
     * the length of the byte array.
     *
     *
     * @param bytes Array in which to put the generated bytes. Cannot be `null`.
     * @param start Index at which to start inserting the generated bytes.
     * @param len Number of bytes to insert.
     * @throws OutOfRangeException if `start < 0` or `start >= bytes.length`.
     * @throws OutOfRangeException if `len < 0` or `len > bytes.length - start`.
     */
    fun nextBytes(
        bytes: ByteArray,
        start: Int,
        len: Int
    ) {
        if (start < 0 ||
            start >= bytes.size
        ) {
            throw TODO()// OutOfRangeException(start, 0, bytes.size)
        }
        if (len < 0 ||
            len > bytes.size - start
        ) {
            throw TODO() // OutOfRangeException(len, 0, bytes.size - start)
        }
        nextBytesFill(bytes, start, len)
    }

    /**
     * Generates random bytes and places them into a user-supplied array.
     *
     *
     *
     * The array is filled with bytes extracted from random integers.
     * This implies that the number of random bytes generated may be larger than
     * the length of the byte array.
     *
     *
     * @param bytes Array in which to put the generated bytes. Cannot be `null`.
     * @param start Index at which to start inserting the generated bytes.
     * @param len Number of bytes to insert.
     */
    private fun nextBytesFill(
        bytes: ByteArray,
        start: Int,
        len: Int
    ) {
        var index = start // Index of first insertion.

        // Index of first insertion plus multiple 4 part of length (i.e. length
        // with two least significant bits unset).
        val indexLoopLimit = index + (len and 0x7ffffffc)

        // Start filling in the byte array, 4 bytes at a time.
        while (index < indexLoopLimit) {
            val random = next(32)
            bytes[index++] = random.toByte()
            bytes[index++] = (random ushr 8).toByte()
            bytes[index++] = (random ushr 16).toByte()
            bytes[index++] = (random ushr 24).toByte()
        }
        val indexLimit = start + len // Index of last insertion + 1.

        // Fill in the remaining bytes.
        if (index < indexLimit) {
            var random = next(32)
            while (true) {
                bytes[index++] = random.toByte()
                random = if (index < indexLimit) {
                    random ushr 8
                } else {
                    break
                }
            }
        }
    }

    companion object {
        /** Serializable version identifier  */
        private const val serialVersionUID = 20130104L
    }
}
