/*
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.jetbrains.kotlinx.statistics.math3.special

import org.jetbrains.kotlinx.statistics.math3.util.ContinuedFraction
import kotlin.jvm.JvmOverloads
import kotlin.math.*

/**
 *
 *
 * This is a utility class that provides computation methods related to the
 * Beta family of functions.
 *
 *
 *
 * Implementation of [.lnBeta] is based on the
 * algorithms described in
 *
 *  * [Didonato and Morris
 * (1986)](http://dx.doi.org/10.1145/22721.23109), *Computation of the Incomplete Gamma Function Ratios
 * and their Inverse*, TOMS 12(4), 377-393,
 *  * [Didonato and Morris
 * (1992)](http://dx.doi.org/10.1145/131766.131776), *Algorithm 708: Significant Digit Computation of the
 * Incomplete Beta Function Ratios*, TOMS 18(3), 360-373,
 *
 * and implemented in the
 * [NSWC Library of Mathematical Functions](http://www.dtic.mil/docs/citations/ADA476840),
 * available
 * [here](http://www.ualberta.ca/CNS/RESEARCH/Software/NumericalNSWC/site.html).
 * This library is "approved for public release", and the
 * [Copyright guidance](http://www.dtic.mil/dtic/pdf/announcements/CopyrightGuidance.pdf)
 * indicates that unless otherwise stated in the code, all FORTRAN functions in
 * this library are license free. Since no such notice appears in the code these
 * functions can safely be ported to Commons-Math.
 *
 *
 *
 */
object Beta {
    /** Maximum allowed numerical error.  */
    private const val DEFAULT_EPSILON = 1E-14

    /** The constant value of ½ln 2π.  */
    private const val HALF_ln_TWO_PI = .9189385332046727

    /**
     *
     *
     * The coefficients of the series expansion of the Δ function. This function
     * is defined as follows
     *
     * <center>Δ(x) = ln Γ(x) - (x - 0.5) ln a + a - 0.5 ln 2π,</center>
     *
     *
     * see equation (23) in Didonato and Morris (1992). The series expansion,
     * which applies for x ≥ 10, reads
     *
     * <pre>
     * 14
     * ====
     * 1  \                2 n
     * Δ(x) = ---  >    d  (10 / x)
     * x  /      n
     * ====
     * n = 0
     * <pre>
    </pre></pre> */
    private val DELTA = doubleArrayOf(
        .833333333333333333333333333333E-01,
        -.277777777777777777777777752282E-04,
        .793650793650793650791732130419E-07,
        -.595238095238095232389839236182E-09,
        .841750841750832853294451671990E-11,
        -.191752691751854612334149171243E-12,
        .641025640510325475730918472625E-14,
        -.295506514125338232839867823991E-15,
        .179643716359402238723287696452E-16,
        -.139228964661627791231203060395E-17,
        .133802855014020915603275339093E-18,
        -.154246009867966094273710216533E-19,
        .197701992980957427278370133333E-20,
        -.234065664793997056856992426667E-21,
        .171348014966398575409015466667E-22
    )

    /**
     * Returns the regularized beta function I(x, a, b).
     *
     * @param x the value.
     * @param a Parameter `a`.
     * @param b Parameter `b`.
     * @param maxIterations Maximum number of "iterations" to complete.
     * @return the regularized beta function I(x, a, b)
     * @throws org.apache.commons.math3.exception.MaxCountExceededException
     * if the algorithm fails to converge.
     */
    fun regularizedBeta(
        x: Double,
        a: Double, b: Double,
        maxIterations: Int
    ): Double {
        return regularizedBeta(x, a, b, DEFAULT_EPSILON, maxIterations)
    }
    /**
     * Returns the regularized beta function I(x, a, b).
     *
     * The implementation of this method is based on:
     *
     *  *
     * [
 * Regularized Beta Function](http://mathworld.wolfram.com/RegularizedBetaFunction.html).
     *  *
     * [
 * Regularized Beta Function](http://functions.wolfram.com/06.21.10.0001.01).
     *
     *
     * @param x the value.
     * @param a Parameter `a`.
     * @param b Parameter `b`.
     * @param epsilon When the absolute value of the nth item in the
     * series is less than epsilon the approximation ceases to calculate
     * further elements in the series.
     * @param maxIterations Maximum number of "iterations" to complete.
     * @return the regularized beta function I(x, a, b)
     * @throws org.apache.commons.math3.exception.MaxCountExceededException
     * if the algorithm fails to converge.
     */
    /**
     * Returns the
     * [
 * regularized beta function](http://mathworld.wolfram.com/RegularizedBetaFunction.html) I(x, a, b).
     *
     * @param x Value.
     * @param a Parameter `a`.
     * @param b Parameter `b`.
     * @return the regularized beta function I(x, a, b).
     * @throws org.apache.commons.math3.exception.MaxCountExceededException
     * if the algorithm fails to converge.
     */
    /**
     * Returns the
     * [
 * regularized beta function](http://mathworld.wolfram.com/RegularizedBetaFunction.html) I(x, a, b).
     *
     * @param x Value.
     * @param a Parameter `a`.
     * @param b Parameter `b`.
     * @param epsilon When the absolute value of the nth item in the
     * series is less than epsilon the approximation ceases to calculate
     * further elements in the series.
     * @return the regularized beta function I(x, a, b)
     * @throws org.apache.commons.math3.exception.MaxCountExceededException
     * if the algorithm fails to converge.
     */
    @JvmOverloads
    fun regularizedBeta(
        x: Double,
        a: Double, b: Double,
        epsilon: Double = DEFAULT_EPSILON, maxIterations: Int = Int.MAX_VALUE
    ): Double {
        val ret: Double
        if (x.isNaN() ||
            a.isNaN() ||
            b.isNaN() || x < 0 || x > 1 || a <= 0 || b <= 0
        ) {
            ret = Double.NaN
        } else if (x > (a + 1) / (2 + b + a) &&
            1 - x <= (b + 1) / (2 + b + a)
        ) {
            ret = 1 - regularizedBeta(1 - x, b, a, epsilon, maxIterations)
        } else {
            val fraction: ContinuedFraction =
                object : ContinuedFraction() {
                    /** {@inheritDoc}  */
                    protected override fun getB(n: Int, x: Double): Double {
                        val m: Double
                        return if (n % 2 == 0) { // even
                            m = n / 2.0
                            m * (b - m) * x /
                                    ((a + 2 * m - 1) * (a + 2 * m))
                        } else {
                            m = (n - 1.0) / 2.0
                            -((a + m) * (a + b + m) * x) /
                                    ((a + 2 * m) * (a + 2 * m + 1.0))
                        }
                    }

                    /** {@inheritDoc}  */
                    protected override fun getA(n: Int, x: Double): Double {
                        return 1.0
                    }
                }
            ret = exp(
                a * ln(x) + b * ln1p(-x) -
                        ln(a) - lnBeta(a, b)
            ) *
                    1.0 / fraction.evaluate(x, epsilon, maxIterations)
        }
        return ret
    }


    /**
     * Returns the value of ln Γ(a + b) for 1 ≤ a, b ≤ 2. Based on the
     * *NSWC Library of Mathematics Subroutines* double precision
     * implementation, `DGSMLN`. In `BetaTest.testlnGammaSum()`,
     * this private method is accessed through reflection.
     *
     * @param a First argument.
     * @param b Second argument.
     * @return the value of `ln(Gamma(a + b))`.
     * @throws OutOfRangeException if `a` or `b` is lower than
     * `1.0` or greater than `2.0`.
     */
    //@Throws(org.apache.commons.math3.exception.OutOfRangeException::class)
    private fun lnGammaSum(a: Double, b: Double): Double {
        if (a < 1.0 || a > 2.0) {
            TODO()//throw org.apache.commons.math3.exception.OutOfRangeException(a, 1.0, 2.0)
        }
        if (b < 1.0 || b > 2.0) {
            TODO() //throw org.apache.commons.math3.exception.OutOfRangeException(b, 1.0, 2.0)
        }
        val x = a - 1.0 + (b - 1.0)
        return if (x <= 0.5) {
            Gamma.lnGamma1p(1.0 + x)
        } else if (x <= 1.5) {
            Gamma.lnGamma1p(x) + ln1p(x)
        } else {
            Gamma.lnGamma1p(x - 1.0) + ln(x * (1.0 + x))
        }
    }

    /**
     * Returns the value of ln[Γ(b) / Γ(a + b)] for a ≥ 0 and b ≥ 10. Based on
     * the *NSWC Library of Mathematics Subroutines* double precision
     * implementation, `DLGDIV`. In
     * `BetaTest.testlnGammaMinuslnGammaSum()`, this private method is
     * accessed through reflection.
     *
     * @param a First argument.
     * @param b Second argument.
     * @return the value of `ln(Gamma(b) / Gamma(a + b))`.
     * @throws NumberIsTooSmallException if `a < 0.0` or `b < 10.0`.
     */
    //@Throws(org.apache.commons.math3.exception.NumberIsTooSmallException::class)
    private fun lnGammaMinuslnGammaSum(
        a: Double,
        b: Double
    ): Double {
        if (a < 0.0) {
            TODO()//throw org.apache.commons.math3.exception.NumberIsTooSmallException(a, 0.0, true)
        }
        if (b < 10.0) {
            TODO()//throw org.apache.commons.math3.exception.NumberIsTooSmallException(b, 10.0, true)
        }

        /*
         * d = a + b - 0.5
         */
        val d: Double
        val w: Double
        if (a <= b) {
            d = b + (a - 0.5)
            w = deltaMinusDeltaSum(a, b)
        } else {
            d = a + (b - 0.5)
            w = deltaMinusDeltaSum(b, a)
        }
        val u: Double = d * ln1p(a / b)
        val v: Double = a * (ln(b) - 1.0)
        return if (u <= v) w - u - v else w - v - u
    }

    /**
     * Returns the value of Δ(b) - Δ(a + b), with 0 ≤ a ≤ b and b ≥ 10. Based
     * on equations (26), (27) and (28) in Didonato and Morris (1992).
     *
     * @param a First argument.
     * @param b Second argument.
     * @return the value of `Delta(b) - Delta(a + b)`
     * @throws OutOfRangeException if `a < 0` or `a > b`
     * @throws NumberIsTooSmallException if `b < 10`
     */
    /*@Throws(
        org.apache.commons.math3.exception.OutOfRangeException::class,
        org.apache.commons.math3.exception.NumberIsTooSmallException::class
    )*/
    private fun deltaMinusDeltaSum(
        a: Double,
        b: Double
    ): Double {
        if (a < 0 || a > b) {
           TODO()// throw org.apache.commons.math3.exception.OutOfRangeException(a, 0, b)
        }
        if (b < 10) {
            TODO()//throw org.apache.commons.math3.exception.NumberIsTooSmallException(b, 10, true)
        }
        val h = a / b
        val p = h / (1.0 + h)
        val q = 1.0 / (1.0 + h)
        val q2 = q * q
        /*
         * s[i] = 1 + q + ... - q**(2 * i)
         */
        val s = DoubleArray(DELTA.size)
        s[0] = 1.0
        for (i in 1 until s.size) {
            s[i] = 1.0 + (q + q2 * s[i - 1])
        }
        /*
         * w = Delta(b) - Delta(a + b)
         */
        val sqrtT = 10.0 / b
        val t = sqrtT * sqrtT
        var w = DELTA[DELTA.size - 1] * s[s.size - 1]
        for (i in DELTA.size - 2 downTo 0) {
            w = t * w + DELTA[i] * s[i]
        }
        return w * p / b
    }

    /**
     * Returns the value of Δ(p) + Δ(q) - Δ(p + q), with p, q ≥ 10. Based on
     * the *NSWC Library of Mathematics Subroutines* double precision
     * implementation, `DBCORR`. In
     * `BetaTest.testSumDeltaMinusDeltaSum()`, this private method is
     * accessed through reflection.
     *
     * @param p First argument.
     * @param q Second argument.
     * @return the value of `Delta(p) + Delta(q) - Delta(p + q)`.
     * @throws NumberIsTooSmallException if `p < 10.0` or `q < 10.0`.
     */
    private fun sumDeltaMinusDeltaSum(
        p: Double,
        q: Double
    ): Double {
        if (p < 10.0) {
            throw TODO()// org.apache.commons.math3.exception.NumberIsTooSmallException(p, 10.0, true)
        }
        if (q < 10.0) {
            throw TODO()// org.apache.commons.math3.exception.NumberIsTooSmallException(q, 10.0, true)
        }
        val a: Double = min(p, q)
        val b: Double = max(p, q)
        val sqrtT = 10.0 / a
        val t = sqrtT * sqrtT
        var z = DELTA[DELTA.size - 1]
        for (i in DELTA.size - 2 downTo 0) {
            z = t * z + DELTA[i]
        }
        return z / a + deltaMinusDeltaSum(a, b)
    }

    /**
     * Returns the value of ln B(p, q) for 0 ≤ x ≤ 1 and p, q > 0. Based on the
     * *NSWC Library of Mathematics Subroutines* implementation,
     * `DBETLN`.
     *
     * @param p First argument.
     * @param q Second argument.
     * @return the value of `ln(Beta(p, q))`, `NaN` if
     * `p <= 0` or `q <= 0`.
     */
    fun lnBeta(p: Double, q: Double): Double {
        if (p.isNaN() || q.isNaN() || p <= 0.0 || q <= 0.0) {
            return Double.NaN
        }
        val a: Double = min(p, q)
        val b: Double = max(p, q)
        return if (a >= 10.0) {
            val w = sumDeltaMinusDeltaSum(a, b)
            val h = a / b
            val c = h / (1.0 + h)
            val u: Double = -(a - 0.5) * ln(c)
            val v: Double = b * ln1p(h)
            if (u <= v) {
                -0.5 * ln(b) + HALF_ln_TWO_PI + w - u - v
            } else {
                -0.5 * ln(b) + HALF_ln_TWO_PI + w - v - u
            }
        } else if (a > 2.0) {
            if (b > 1000.0) {
                val n: Int = floor(a - 1.0).toInt()
                var prod = 1.0
                var ared = a
                for (i in 0 until n) {
                    ared -= 1.0
                    prod *= ared / (1.0 + ared / b)
                }
                ln(prod) - n * ln(b) +
                        (Gamma.lnGamma(ared) +
                                lnGammaMinuslnGammaSum(ared, b))
            } else {
                var prod1 = 1.0
                var ared = a
                while (ared > 2.0) {
                    ared -= 1.0
                    val h = ared / b
                    prod1 *= h / (1.0 + h)
                }
                if (b < 10.0) {
                    var prod2 = 1.0
                    var bred = b
                    while (bred > 2.0) {
                        bred -= 1.0
                        prod2 *= bred / (ared + bred)
                    }
                    ln(prod1) +
                            ln(prod2) +
                            (Gamma.lnGamma(ared) +
                                    (Gamma.lnGamma(bred) -
                                            lnGammaSum(ared, bred)))
                } else {
                    ln(prod1) +
                            Gamma.lnGamma(ared) +
                            lnGammaMinuslnGammaSum(ared, b)
                }
            }
        } else if (a >= 1.0) {
            if (b > 2.0) {
                if (b < 10.0) {
                    var prod = 1.0
                    var bred = b
                    while (bred > 2.0) {
                        bred -= 1.0
                        prod *= bred / (a + bred)
                    }
                    ln(prod) +
                            (Gamma.lnGamma(a) +
                                    (Gamma.lnGamma(bred) -
                                            lnGammaSum(a, bred)))
                } else {
                    Gamma.lnGamma(a) +
                            lnGammaMinuslnGammaSum(a, b)
                }
            } else {
                Gamma.lnGamma(a) +
                        Gamma.lnGamma(b) -
                        lnGammaSum(a, b)
            }
        } else {
            if (b >= 10.0) {
                Gamma.lnGamma(a) +
                        lnGammaMinuslnGammaSum(a, b)
            } else {
                // The following command is the original NSWC implementation.
                // return Gamma.lnGamma(a) +
                // (Gamma.lnGamma(b) - Gamma.lnGamma(a + b));
                // The following command turns out to be more accurate.
                ln(
                    Gamma.gamma(a) * Gamma.gamma(b) /
                            Gamma.gamma(a + b)
                )
            }
        }
    }
}
