/*
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.jetbrains.kotlinx.statistics.math3.special

import org.jetbrains.kotlinx.statistics.math3.util.ContinuedFraction
import org.jetbrains.kotlinx.statistics.math3.util.rint
import kotlin.jvm.JvmOverloads
import kotlin.math.*

/**
 *
 *
 * This is a utility class that provides computation methods related to the
 *  (Gamma) family of functions.
 *
 *
 *
 * Implementation of [.invGamma1pm1] and
 * [.lnGamma1p] is based on the algorithms described in
 *
 *  * [Didonato and Morris
 * (1986)](http://dx.doi.org/10.1145/22721.23109), *Computation of the Incomplete Gamma Function Ratios and
 * their Inverse*, TOMS 12(4), 377-393,
 *  * [Didonato and Morris
 * (1992)](http://dx.doi.org/10.1145/131766.131776), *Algorithm 708: Significant Digit Computation of the
 * Incomplete Beta Function Ratios*, TOMS 18(3), 360-373,
 *
 * and implemented in the
 * [NSWC Library of Mathematical Functions](http://www.dtic.mil/docs/citations/ADA476840),
 * available
 * [here](http://www.ualberta.ca/CNS/RESEARCH/Software/NumericalNSWC/site.html).
 * This library is "approved for public release", and the
 * [Copyright guidance](http://www.dtic.mil/dtic/pdf/announcements/CopyrightGuidance.pdf)
 * indicates that unless otherwise stated in the code, all FORTRAN functions in
 * this library are license free. Since no such notice appears in the code these
 * functions can safely be ported to Commons-Math.
 *
 *
 */
object Gamma {
    /**
     * [Euler-Mascheroni constant](http://en.wikipedia.org/wiki/Euler-Mascheroni_constant)
     * @since 2.0
     */
    const val GAMMA = 0.577215664901532860606512090082

    /**
     * The value of the `g` constant in the Lanczos approximation, see
     * [.lanczos].
     * @since 3.1
     */
    const val LANCZOS_G = 607.0 / 128.0

    /** Maximum allowed numerical error.  */
    private const val DEFAULT_EPSILON = 10e-15

    /** Lanczos coefficients  */
    private val LANCZOS = doubleArrayOf(
        0.99999999999999709182,
        57.156235665862923517,
        -59.597960355475491248,
        14.136097974741747174,
        -0.49191381609762019978,
        .33994649984811888699e-4,
        .46523628927048575665e-4,
        -.98374475304879564677e-4,
        .15808870322491248884e-3,
        -.21026444172410488319e-3,
        .21743961811521264320e-3,
        -.16431810653676389022e-3,
        .84418223983852743293e-4,
        -.26190838401581408670e-4,
        .36899182659531622704e-5
    )

    /** Avoid repeated computation of ln of 2 PI in lnGamma  */
    private val HALF_ln_2_PI: Double = 0.5 * ln(2.0 * PI)

    /** The constant value of (2).  */
    private const val SQRT_TWO_PI = 2.506628274631000502
    // limits for switching algorithm in digamma
    /** C limit.  */
    private const val C_LIMIT = 49.0

    /** S limit.  */
    private const val S_LIMIT = 1e-5
    /*
     * Constants for the computation of double invGamma1pm1(double).
     * Copied from DGAM1 in the NSWC library.
     */
    /** The constant `A0` defined in `DGAM1`.  */
    private const val INV_GAMMA1P_M1_A0 = .611609510448141581788E-08

    /** The constant `A1` defined in `DGAM1`.  */
    private const val INV_GAMMA1P_M1_A1 = .624730830116465516210E-08

    /** The constant `B1` defined in `DGAM1`.  */
    private const val INV_GAMMA1P_M1_B1 = .203610414066806987300E+00

    /** The constant `B2` defined in `DGAM1`.  */
    private const val INV_GAMMA1P_M1_B2 = .266205348428949217746E-01

    /** The constant `B3` defined in `DGAM1`.  */
    private const val INV_GAMMA1P_M1_B3 = .493944979382446875238E-03

    /** The constant `B4` defined in `DGAM1`.  */
    private const val INV_GAMMA1P_M1_B4 = -.851419432440314906588E-05

    /** The constant `B5` defined in `DGAM1`.  */
    private const val INV_GAMMA1P_M1_B5 = -.643045481779353022248E-05

    /** The constant `B6` defined in `DGAM1`.  */
    private const val INV_GAMMA1P_M1_B6 = .992641840672773722196E-06

    /** The constant `B7` defined in `DGAM1`.  */
    private const val INV_GAMMA1P_M1_B7 = -.607761895722825260739E-07

    /** The constant `B8` defined in `DGAM1`.  */
    private const val INV_GAMMA1P_M1_B8 = .195755836614639731882E-09

    /** The constant `P0` defined in `DGAM1`.  */
    private const val INV_GAMMA1P_M1_P0 = .6116095104481415817861E-08

    /** The constant `P1` defined in `DGAM1`.  */
    private const val INV_GAMMA1P_M1_P1 = .6871674113067198736152E-08

    /** The constant `P2` defined in `DGAM1`.  */
    private const val INV_GAMMA1P_M1_P2 = .6820161668496170657918E-09

    /** The constant `P3` defined in `DGAM1`.  */
    private const val INV_GAMMA1P_M1_P3 = .4686843322948848031080E-10

    /** The constant `P4` defined in `DGAM1`.  */
    private const val INV_GAMMA1P_M1_P4 = .1572833027710446286995E-11

    /** The constant `P5` defined in `DGAM1`.  */
    private const val INV_GAMMA1P_M1_P5 = -.1249441572276366213222E-12

    /** The constant `P6` defined in `DGAM1`.  */
    private const val INV_GAMMA1P_M1_P6 = .4343529937408594255178E-14

    /** The constant `Q1` defined in `DGAM1`.  */
    private const val INV_GAMMA1P_M1_Q1 = .3056961078365221025009E+00

    /** The constant `Q2` defined in `DGAM1`.  */
    private const val INV_GAMMA1P_M1_Q2 = .5464213086042296536016E-01

    /** The constant `Q3` defined in `DGAM1`.  */
    private const val INV_GAMMA1P_M1_Q3 = .4956830093825887312020E-02

    /** The constant `Q4` defined in `DGAM1`.  */
    private const val INV_GAMMA1P_M1_Q4 = .2692369466186361192876E-03

    /** The constant `C` defined in `DGAM1`.  */
    private const val INV_GAMMA1P_M1_C = -.422784335098467139393487909917598E+00

    /** The constant `C0` defined in `DGAM1`.  */
    private const val INV_GAMMA1P_M1_C0 = .577215664901532860606512090082402E+00

    /** The constant `C1` defined in `DGAM1`.  */
    private const val INV_GAMMA1P_M1_C1 = -.655878071520253881077019515145390E+00

    /** The constant `C2` defined in `DGAM1`.  */
    private const val INV_GAMMA1P_M1_C2 = -.420026350340952355290039348754298E-01

    /** The constant `C3` defined in `DGAM1`.  */
    private const val INV_GAMMA1P_M1_C3 = .166538611382291489501700795102105E+00

    /** The constant `C4` defined in `DGAM1`.  */
    private const val INV_GAMMA1P_M1_C4 = -.421977345555443367482083012891874E-01

    /** The constant `C5` defined in `DGAM1`.  */
    private const val INV_GAMMA1P_M1_C5 = -.962197152787697356211492167234820E-02

    /** The constant `C6` defined in `DGAM1`.  */
    private const val INV_GAMMA1P_M1_C6 = .721894324666309954239501034044657E-02

    /** The constant `C7` defined in `DGAM1`.  */
    private const val INV_GAMMA1P_M1_C7 = -.116516759185906511211397108401839E-02

    /** The constant `C8` defined in `DGAM1`.  */
    private const val INV_GAMMA1P_M1_C8 = -.215241674114950972815729963053648E-03

    /** The constant `C9` defined in `DGAM1`.  */
    private const val INV_GAMMA1P_M1_C9 = .128050282388116186153198626328164E-03

    /** The constant `C10` defined in `DGAM1`.  */
    private const val INV_GAMMA1P_M1_C10 = -.201348547807882386556893914210218E-04

    /** The constant `C11` defined in `DGAM1`.  */
    private const val INV_GAMMA1P_M1_C11 = -.125049348214267065734535947383309E-05

    /** The constant `C12` defined in `DGAM1`.  */
    private const val INV_GAMMA1P_M1_C12 = .113302723198169588237412962033074E-05

    /** The constant `C13` defined in `DGAM1`.  */
    private const val INV_GAMMA1P_M1_C13 = -.205633841697760710345015413002057E-06

    /**
     *
     *
     * Returns the value of ln&nbsp;(x) for x&nbsp;&gt;&nbsp;0.
     *
     *
     *
     * For x  8, the implementation is based on the double precision
     * implementation in the *NSWC Library of Mathematics Subroutines*,
     * `DGAMLN`. For x &gt; 8, the implementation is based on
     *
     *
     *  * [Gamma
 * Function](http://mathworld.wolfram.com/GammaFunction.html), equation (28).
     *  * [
 * Lanczos Approximation](http://mathworld.wolfram.com/LanczosApproximation.html), equations (1) through (5).
     *  * [Paul Godfrey, A note on
 * the computation of the convergent Lanczos complex Gamma
 * approximation](http://my.fit.edu/~gabdo/gamma.txt)
     *
     *
     * @param x Argument.
     * @return the value of `ln(Gamma(x))`, `Double.NaN` if
     * `x <= 0.0`.
     */
    fun lnGamma(x: Double): Double {
        val ret: Double
        if (x.isNaN() || x <= 0.0) {
            ret = Double.NaN
        } else if (x < 0.5) {
            return lnGamma1p(x) - ln(x)
        } else if (x <= 2.5) {
            return lnGamma1p(x - 0.5 - 0.5)
        } else if (x <= 8.0) {
            val n = floor(x - 1.5).toInt()
            var prod = 1.0
            for (i in 1..n) {
                prod *= x - i
            }
            return lnGamma1p(x - (n + 1)) + ln(prod)
        } else {
            val sum = lanczos(x)
            val tmp = x + LANCZOS_G + .5
            ret = (x + .5) * ln(tmp) - tmp +
                    HALF_ln_2_PI + ln(sum / x)
        }
        return ret
    }
    /**
     * Returns the regularized gamma function P(a, x).
     *
     * The implementation of this method is based on:
     *
     *  *
     * [
 * Regularized Gamma Function](http://mathworld.wolfram.com/RegularizedGammaFunction.html), equation (1)
     *
     *  *
     * [
 * Incomplete Gamma Function](http://mathworld.wolfram.com/IncompleteGammaFunction.html), equation (4).
     *
     *  *
     * [
 * Confluent Hypergeometric Function of the First Kind](http://mathworld.wolfram.com/ConfluentHypergeometricFunctionoftheFirstKind.html), equation (1).
     *
     *
     *
     * @param a the a parameter.
     * @param x the value.
     * @param epsilon When the absolute value of the nth item in the
     * series is less than epsilon the approximation ceases to calculate
     * further elements in the series.
     * @param maxIterations Maximum number of "iterations" to complete.
     * @return the regularized gamma function P(a, x)
     * @throws MaxCountExceededException if the algorithm fails to converge.
     */
    /**
     * Returns the regularized gamma function P(a, x).
     *
     * @param a Parameter.
     * @param x Value.
     * @return the regularized gamma function P(a, x).
     * @throws MaxCountExceededException if the algorithm fails to converge.
     */
    @JvmOverloads
    fun regularizedGammaP(
        a: Double,
        x: Double,
        epsilon: Double = DEFAULT_EPSILON,
        maxIterations: Int = Int.MAX_VALUE
    ): Double {
        val ret: Double
        if (a.isNaN() || x.isNaN() || a <= 0.0 || x < 0.0) {
            ret = Double.NaN
        } else if (x == 0.0) {
            ret = 0.0
        } else if (x >= a + 1) {
            // use regularizedGammaQ because it should converge faster in this
            // case.
            ret = 1.0 - regularizedGammaQ(a, x, epsilon, maxIterations)
        } else {
            // calculate series
            var n = 0.0 // current element index
            var an = 1.0 / a // n-th element in the series
            var sum = an // partial sum
            while (abs(an / sum) > epsilon && n < maxIterations && sum < Double.POSITIVE_INFINITY) {
                // compute next element in the series
                n += 1.0
                an *= x / (a + n)

                // update partial sum
                sum += an
            }
            ret = if (n >= maxIterations) {
                throw TODO()// MaxCountExceededException(maxIterations)
            } else if (sum.isInfinite()) {
                1.0
            } else {
                exp(-x + a * ln(x) - lnGamma(a)) * sum
            }
        }
        return ret
    }
    /**
     * Returns the regularized gamma function Q(a, x) = 1 - P(a, x).
     *
     * The implementation of this method is based on:
     *
     *  *
     * [
 * Regularized Gamma Function](http://mathworld.wolfram.com/RegularizedGammaFunction.html), equation (1).
     *
     *  *
     * [
 * Regularized incomplete gamma function: Continued fraction representations
 * (formula 06.08.10.0003)](http://functions.wolfram.com/GammaBetaErf/GammaRegularized/10/0003/)
     *
     *
     *
     * @param a the a parameter.
     * @param x the value.
     * @param epsilon When the absolute value of the nth item in the
     * series is less than epsilon the approximation ceases to calculate
     * further elements in the series.
     * @param maxIterations Maximum number of "iterations" to complete.
     * @return the regularized gamma function P(a, x)
     * @throws MaxCountExceededException if the algorithm fails to converge.
     */
    /**
     * Returns the regularized gamma function Q(a, x) = 1 - P(a, x).
     *
     * @param a the a parameter.
     * @param x the value.
     * @return the regularized gamma function Q(a, x)
     * @throws MaxCountExceededException if the algorithm fails to converge.
     */
    @JvmOverloads
    fun regularizedGammaQ(
        a: Double,
        x: Double,
        epsilon: Double = DEFAULT_EPSILON,
        maxIterations: Int = Int.MAX_VALUE
    ): Double {
        var ret: Double
        if (a.isNaN() || x.isNaN() || a <= 0.0 || x < 0.0) {
            ret = Double.NaN
        } else if (x == 0.0) {
            ret = 1.0
        } else if (x < a + 1.0) {
            // use regularizedGammaP because it should converge faster in this
            // case.
            ret = 1.0 - regularizedGammaP(a, x, epsilon, maxIterations)
        } else {
            // create continued fraction
            val cf: ContinuedFraction = object : ContinuedFraction() {
                /** {@inheritDoc}  */
                protected override fun getA(n: Int, x: Double): Double {
                    return 2.0 * n + 1.0 - a + x
                }

                /** {@inheritDoc}  */
                protected override fun getB(n: Int, x: Double): Double {
                    return n * (a - n)
                }
            }
            ret = 1.0 / cf.evaluate(x, epsilon, maxIterations)
            ret = exp(-x + a * ln(x) - lnGamma(a)) * ret
        }
        return ret
    }

    /**
     *
     * Computes the digamma function of x.
     *
     *
     * This is an independently written implementation of the algorithm described in
     * Jose Bernardo, Algorithm AS 103: Psi (Digamma) Function, Applied Statistics, 1976.
     *
     *
     * Some of the constants have been changed to increase accuracy at the moderate expense
     * of run-time.  The result should be accurate to within 10^-8 absolute tolerance for
     * x >= 10^-5 and within 10^-8 relative tolerance for x > 0.
     *
     *
     * Performance for large negative values of x will be quite expensive (proportional to
     * |x|).  Accuracy for negative values of x should be about 10^-8 absolute for results
     * less than 10^5 and 10^-8 relative for results larger than that.
     *
     * @param x Argument.
     * @return digamma(x) to within 10-8 relative or absolute error whichever is smaller.
     * @see [Digamma](http://en.wikipedia.org/wiki/Digamma_function)
     *
     * @see [Bernardo&apos;s original article ](http://www.uv.es/~bernardo/1976AppStatist.pdf)
     *
     * @since 2.0
     */
    fun digamma(x: Double): Double {
        if (x.isNaN() || x.isInfinite()) {
            return x
        }
        if (x > 0 && x <= S_LIMIT) {
            // use method 5 from Bernardo AS103
            // accurate to O(x)
            return -GAMMA - 1 / x
        }
        if (x >= C_LIMIT) {
            // use method 4 (accurate to O(1/x^8)
            val inv = 1 / (x * x)
            //            1       1        1         1
            // ln(x) -  --- - ------ + ------- - -------
            //           2 x   12 x^2   120 x^4   252 x^6
            return ln(x) - 0.5 / x - inv * (1.0 / 12 + inv * (1.0 / 120 - inv / 252))
        }
        return digamma(x + 1) - 1 / x
    }

    /**
     * Computes the trigamma function of x.
     * This function is derived by taking the derivative of the implementation
     * of digamma.
     *
     * @param x Argument.
     * @return trigamma(x) to within 10-8 relative or absolute error whichever is smaller
     * @see [Trigamma](http://en.wikipedia.org/wiki/Trigamma_function)
     *
     * @see Gamma.digamma
     * @since 2.0
     */
    fun trigamma(x: Double): Double {
        if (x.isNaN() || x.isInfinite()) {
            return x
        }
        if (x > 0 && x <= S_LIMIT) {
            return 1 / (x * x)
        }
        if (x >= C_LIMIT) {
            val inv = 1 / (x * x)
            //  1    1      1       1       1
            //  - + ---- + ---- - ----- + -----
            //  x      2      3       5       7
            //      2 x    6 x    30 x    42 x
            return 1 / x + inv / 2 + inv / x * (1.0 / 6 - inv * (1.0 / 30 + inv / 42))
        }
        return trigamma(x + 1) + 1 / (x * x)
    }

    /**
     *
     *
     * Returns the Lanczos approximation used to compute the gamma function.
     * The Lanczos approximation is related to the Gamma function by the
     * following equation
     * <center>
     * `gamma(x) = sqrt(2 * pi) / x * (x + g + 0.5) ^ (x + 0.5)
     * * exp(-x - g - 0.5) * lanczos(x)`,
    </center> *
     * where `g` is the Lanczos constant.
     *
     *
     * @param x Argument.
     * @return The Lanczos approximation.
     * @see [Lanczos Approximation](http://mathworld.wolfram.com/LanczosApproximation.html)
     * equations
     * @since 3.1
     */
    fun lanczos(x: Double): Double {
        var sum = 0.0
        for (i in LANCZOS.size - 1 downTo 1) {
            sum += LANCZOS[i] / (x + i)
        }
        return sum + LANCZOS[0]
    }

    /**
     * Returns the value of 1 / (1 + x) - 1 for -0&#46;5  x
     * 1&#46;5. This implementation is based on the double precision
     * implementation in the *NSWC Library of Mathematics Subroutines*,
     * `DGAM1`.
     *
     * @param x Argument.
     * @return The value of `1.0 / Gamma(1.0 + x) - 1.0`.
     * @throws NumberIsTooSmallException if `x < -0.5`
     * @throws NumberIsTooLargeException if `x > 1.5`
     * @since 3.1
     */
    fun invGamma1pm1(x: Double): Double {
        if (x < -0.5) {
            throw TODO()// NumberIsTooSmallException(x, -0.5, true)
        }
        if (x > 1.5) {
            throw TODO()// NumberIsTooLargeException(x, 1.5, true)
        }
        val ret: Double
        val t = if (x <= 0.5) x else x - 0.5 - 0.5
        if (t < 0.0) {
            val a = INV_GAMMA1P_M1_A0 + t * INV_GAMMA1P_M1_A1
            var b = INV_GAMMA1P_M1_B8
            b = INV_GAMMA1P_M1_B7 + t * b
            b = INV_GAMMA1P_M1_B6 + t * b
            b = INV_GAMMA1P_M1_B5 + t * b
            b = INV_GAMMA1P_M1_B4 + t * b
            b = INV_GAMMA1P_M1_B3 + t * b
            b = INV_GAMMA1P_M1_B2 + t * b
            b = INV_GAMMA1P_M1_B1 + t * b
            b = 1.0 + t * b
            var c = INV_GAMMA1P_M1_C13 + t * (a / b)
            c = INV_GAMMA1P_M1_C12 + t * c
            c = INV_GAMMA1P_M1_C11 + t * c
            c = INV_GAMMA1P_M1_C10 + t * c
            c = INV_GAMMA1P_M1_C9 + t * c
            c = INV_GAMMA1P_M1_C8 + t * c
            c = INV_GAMMA1P_M1_C7 + t * c
            c = INV_GAMMA1P_M1_C6 + t * c
            c = INV_GAMMA1P_M1_C5 + t * c
            c = INV_GAMMA1P_M1_C4 + t * c
            c = INV_GAMMA1P_M1_C3 + t * c
            c = INV_GAMMA1P_M1_C2 + t * c
            c = INV_GAMMA1P_M1_C1 + t * c
            c = INV_GAMMA1P_M1_C + t * c
            ret = if (x > 0.5) {
                t * c / x
            } else {
                x * (c + 0.5 + 0.5)
            }
        } else {
            var p = INV_GAMMA1P_M1_P6
            p = INV_GAMMA1P_M1_P5 + t * p
            p = INV_GAMMA1P_M1_P4 + t * p
            p = INV_GAMMA1P_M1_P3 + t * p
            p = INV_GAMMA1P_M1_P2 + t * p
            p = INV_GAMMA1P_M1_P1 + t * p
            p = INV_GAMMA1P_M1_P0 + t * p
            var q = INV_GAMMA1P_M1_Q4
            q = INV_GAMMA1P_M1_Q3 + t * q
            q = INV_GAMMA1P_M1_Q2 + t * q
            q = INV_GAMMA1P_M1_Q1 + t * q
            q = 1.0 + t * q
            var c = INV_GAMMA1P_M1_C13 + p / q * t
            c = INV_GAMMA1P_M1_C12 + t * c
            c = INV_GAMMA1P_M1_C11 + t * c
            c = INV_GAMMA1P_M1_C10 + t * c
            c = INV_GAMMA1P_M1_C9 + t * c
            c = INV_GAMMA1P_M1_C8 + t * c
            c = INV_GAMMA1P_M1_C7 + t * c
            c = INV_GAMMA1P_M1_C6 + t * c
            c = INV_GAMMA1P_M1_C5 + t * c
            c = INV_GAMMA1P_M1_C4 + t * c
            c = INV_GAMMA1P_M1_C3 + t * c
            c = INV_GAMMA1P_M1_C2 + t * c
            c = INV_GAMMA1P_M1_C1 + t * c
            c = INV_GAMMA1P_M1_C0 + t * c
            ret = if (x > 0.5) {
                t / x * (c - 0.5 - 0.5)
            } else {
                x * c
            }
        }
        return ret
    }

    /**
     * Returns the value of ln (1 + x) for -0&#46;5  x  1&#46;5.
     * This implementation is based on the double precision implementation in
     * the *NSWC Library of Mathematics Subroutines*, `DGMLN1`.
     *
     * @param x Argument.
     * @return The value of `ln(Gamma(1 + x))`.
     * @throws NumberIsTooSmallException if `x < -0.5`.
     * @throws NumberIsTooLargeException if `x > 1.5`.
     * @since 3.1
     */
    //@Throws(NumberIsTooSmallException::class, NumberIsTooLargeException::class)
    fun lnGamma1p(x: Double): Double {
        if (x < -0.5) {
            throw TODO()// NumberIsTooSmallException(x, -0.5, true)
        }
        if (x > 1.5) {
            throw TODO()// NumberIsTooLargeException(x, 1.5, true)
        }
        return -ln1p(invGamma1pm1(x))
    }

    /**
     * Returns the value of Γ(x). Based on the *NSWC Library of
     * Mathematics Subroutines* double precision implementation,
     * `DGAMMA`.
     *
     * @param x Argument.
     * @return the value of `Gamma(x)`.
     * @since 3.1
     */
    fun gamma(x: Double): Double {
        if (x == rint(x) && x <= 0.0) {
            return Double.NaN
        }
        val ret: Double
        val absX: Double = abs(x)
        if (absX <= 20.0) {
            if (x >= 1.0) {
                /*
                 * From the recurrence relation
                 * Gamma(x) = (x - 1) * ... * (x - n) * Gamma(x - n),
                 * then
                 * Gamma(t) = 1 / [1 + invGamma1pm1(t - 1)],
                 * where t = x - n. This means that t must satisfy
                 * -0.5 <= t - 1 <= 1.5.
                 */
                var prod = 1.0
                var t = x
                while (t > 2.5) {
                    t -= 1.0
                    prod *= t
                }
                ret = prod / (1.0 + invGamma1pm1(t - 1.0))
            } else {
                /*
                 * From the recurrence relation
                 * Gamma(x) = Gamma(x + n + 1) / [x * (x + 1) * ... * (x + n)]
                 * then
                 * Gamma(x + n + 1) = 1 / [1 + invGamma1pm1(x + n)],
                 * which requires -0.5 <= x + n <= 1.5.
                 */
                var prod = x
                var t = x
                while (t < -0.5) {
                    t += 1.0
                    prod *= t
                }
                ret = 1.0 / (prod * (1.0 + invGamma1pm1(t)))
            }
        } else {
            val y = absX + LANCZOS_G + 0.5
            val gammaAbs: Double = SQRT_TWO_PI / absX *
                    y.pow(absX + 0.5) *
                    exp(-y) * lanczos(absX)
            ret = if (x > 0.0) {
                gammaAbs
            } else {
                /*
                 * From the reflection formula
                 * Gamma(x) * Gamma(1 - x) * sin(pi * x) = pi,
                 * and the recurrence relation
                 * Gamma(1 - x) = -x * Gamma(-x),
                 * it is found
                 * Gamma(x) = -pi / [x * sin(pi * x) * Gamma(-x)].
                 */
                -PI /
                        (x * sin(PI * x) * gammaAbs)
            }
        }
        return ret
    }
}
