/*
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.jetbrains.kotlinx.statistics.math3.stat

import org.jetbrains.kotlinx.statistics.math3.stat.descriptive.DescriptiveStatistics
import org.jetbrains.kotlinx.statistics.math3.stat.descriptive.UnivariateStatistic
import org.jetbrains.kotlinx.statistics.math3.stat.descriptive.moment.Mean
import org.jetbrains.kotlinx.statistics.math3.stat.descriptive.moment.Variance
import org.jetbrains.kotlinx.statistics.math3.stat.descriptive.summary.*

/**
 * StatUtils provides static methods for computing statistics based on data
 * stored in double[] arrays.
 *
 */
object StatUtils {

    /** mean  */
    private val MEAN: UnivariateStatistic = Mean()

    fun mean(values: DoubleArray): Double {
        return MEAN.evaluate(values)
    }

    /** variance  */
    private val VARIANCE: Variance = Variance()


    fun variance(values: DoubleArray): Double {
        return VARIANCE.evaluate(values)
    }
    /*
    /** sum  */
    private val SUM: UnivariateStatistic = Sum()

    /** sumSq  */
    private val SUM_OF_SQUARES: UnivariateStatistic = SumOfSquares()

    /** prod  */
    private val PRODUCT: UnivariateStatistic = Product()

    /** sumLog  */
    private val SUM_OF_LOGS: UnivariateStatistic = SumOfLogs()

    /** min  */
    private val MIN: UnivariateStatistic = Min()

    /** max  */
    private val MAX: UnivariateStatistic = Max()

    /** mean  */
    private val MEAN: UnivariateStatistic = Mean()

    /** variance  */
    private val VARIANCE: Variance = Variance()

    /** percentile  */
    private val PERCENTILE: Percentile = Percentile()

    /** geometric mean  */
    private val GEOMETRIC_MEAN: GeometricMean = GeometricMean()

    /**
     * Returns the sum of the values in the input array, or
     * `Double.NaN` if the array is empty.
     *
     *
     * Throws `IllegalArgumentException` if the input array
     * is null.
     *
     * @param values  array of values to sum
     * @return the sum of the values or `Double.NaN` if the array
     * is empty
     * @throws MathIllegalArgumentException if the array is null
     */
    @Throws(MathIllegalArgumentException::class)
    fun sum(values: DoubleArray?): Double {
        return SUM.evaluate(values)
    }

    /**
     * Returns the sum of the entries in the specified portion of
     * the input array, or `Double.NaN` if the designated subarray
     * is empty.
     *
     *
     * Throws `IllegalArgumentException` if the array is null.
     *
     * @param values the input array
     * @param begin index of the first array element to include
     * @param length the number of elements to include
     * @return the sum of the values or Double.NaN if length = 0
     * @throws MathIllegalArgumentException if the array is null or the array index
     * parameters are not valid
     */
    @Throws(MathIllegalArgumentException::class)
    fun sum(
        values: DoubleArray?, begin: Int,
        length: Int
    ): Double {
        return SUM.evaluate(values, begin, length)
    }

    /**
     * Returns the sum of the squares of the entries in the input array, or
     * `Double.NaN` if the array is empty.
     *
     *
     * Throws `IllegalArgumentException` if the array is null.
     *
     * @param values  input array
     * @return the sum of the squared values or `Double.NaN` if the
     * array is empty
     * @throws MathIllegalArgumentException if the array is null
     */
    @Throws(MathIllegalArgumentException::class)
    fun sumSq(values: DoubleArray?): Double {
        return SUM_OF_SQUARES.evaluate(values)
    }

    /**
     * Returns the sum of the squares of the entries in the specified portion of
     * the input array, or `Double.NaN` if the designated subarray
     * is empty.
     *
     *
     * Throws `IllegalArgumentException` if the array is null.
     *
     * @param values the input array
     * @param begin index of the first array element to include
     * @param length the number of elements to include
     * @return the sum of the squares of the values or Double.NaN if length = 0
     * @throws MathIllegalArgumentException if the array is null or the array index
     * parameters are not valid
     */
    @Throws(MathIllegalArgumentException::class)
    fun sumSq(
        values: DoubleArray?, begin: Int,
        length: Int
    ): Double {
        return SUM_OF_SQUARES.evaluate(values, begin, length)
    }

    /**
     * Returns the product of the entries in the input array, or
     * `Double.NaN` if the array is empty.
     *
     *
     * Throws `IllegalArgumentException` if the array is null.
     *
     * @param values the input array
     * @return the product of the values or Double.NaN if the array is empty
     * @throws MathIllegalArgumentException if the array is null
     */
    @Throws(MathIllegalArgumentException::class)
    fun product(values: DoubleArray?): Double {
        return PRODUCT.evaluate(values)
    }

    /**
     * Returns the product of the entries in the specified portion of
     * the input array, or `Double.NaN` if the designated subarray
     * is empty.
     *
     *
     * Throws `IllegalArgumentException` if the array is null.
     *
     * @param values the input array
     * @param begin index of the first array element to include
     * @param length the number of elements to include
     * @return the product of the values or Double.NaN if length = 0
     * @throws MathIllegalArgumentException if the array is null or the array index
     * parameters are not valid
     */
    @Throws(MathIllegalArgumentException::class)
    fun product(
        values: DoubleArray?, begin: Int,
        length: Int
    ): Double {
        return PRODUCT.evaluate(values, begin, length)
    }

    /**
     * Returns the sum of the natural logs of the entries in the input array, or
     * `Double.NaN` if the array is empty.
     *
     *
     * Throws `IllegalArgumentException` if the array is null.
     *
     *
     * See [org.apache.commons.math3.stat.descriptive.summary.SumOfLogs].
     *
     *
     * @param values the input array
     * @return the sum of the natural logs of the values or Double.NaN if
     * the array is empty
     * @throws MathIllegalArgumentException if the array is null
     */
    @Throws(MathIllegalArgumentException::class)
    fun sumLog(values: DoubleArray?): Double {
        return SUM_OF_LOGS.evaluate(values)
    }

    /**
     * Returns the sum of the natural logs of the entries in the specified portion of
     * the input array, or `Double.NaN` if the designated subarray
     * is empty.
     *
     *
     * Throws `IllegalArgumentException` if the array is null.
     *
     *
     * See [org.apache.commons.math3.stat.descriptive.summary.SumOfLogs].
     *
     *
     * @param values the input array
     * @param begin index of the first array element to include
     * @param length the number of elements to include
     * @return the sum of the natural logs of the values or Double.NaN if
     * length = 0
     * @throws MathIllegalArgumentException if the array is null or the array index
     * parameters are not valid
     */
    @Throws(MathIllegalArgumentException::class)
    fun sumLog(
        values: DoubleArray?, begin: Int,
        length: Int
    ): Double {
        return SUM_OF_LOGS.evaluate(values, begin, length)
    }

    /**
     * Returns the arithmetic mean of the entries in the input array, or
     * `Double.NaN` if the array is empty.
     *
     *
     * Throws `IllegalArgumentException` if the array is null.
     *
     *
     * See [org.apache.commons.math3.stat.descriptive.moment.Mean] for
     * details on the computing algorithm.
     *
     * @param values the input array
     * @return the mean of the values or Double.NaN if the array is empty
     * @throws MathIllegalArgumentException if the array is null
     */
    @Throws(MathIllegalArgumentException::class)
    fun mean(values: DoubleArray?): Double {
        return MEAN.evaluate(values)
    }

    /**
     * Returns the arithmetic mean of the entries in the specified portion of
     * the input array, or `Double.NaN` if the designated subarray
     * is empty.
     *
     *
     * Throws `IllegalArgumentException` if the array is null.
     *
     *
     * See [org.apache.commons.math3.stat.descriptive.moment.Mean] for
     * details on the computing algorithm.
     *
     * @param values the input array
     * @param begin index of the first array element to include
     * @param length the number of elements to include
     * @return the mean of the values or Double.NaN if length = 0
     * @throws MathIllegalArgumentException if the array is null or the array index
     * parameters are not valid
     */
    @Throws(MathIllegalArgumentException::class)
    fun mean(
        values: DoubleArray?, begin: Int,
        length: Int
    ): Double {
        return MEAN.evaluate(values, begin, length)
    }

    /**
     * Returns the geometric mean of the entries in the input array, or
     * `Double.NaN` if the array is empty.
     *
     *
     * Throws `IllegalArgumentException` if the array is null.
     *
     *
     * See [org.apache.commons.math3.stat.descriptive.moment.GeometricMean]
     * for details on the computing algorithm.
     *
     * @param values the input array
     * @return the geometric mean of the values or Double.NaN if the array is empty
     * @throws MathIllegalArgumentException if the array is null
     */
    @Throws(MathIllegalArgumentException::class)
    fun geometricMean(values: DoubleArray?): Double {
        return GEOMETRIC_MEAN.evaluate(values)
    }

    /**
     * Returns the geometric mean of the entries in the specified portion of
     * the input array, or `Double.NaN` if the designated subarray
     * is empty.
     *
     *
     * Throws `IllegalArgumentException` if the array is null.
     *
     *
     * See [org.apache.commons.math3.stat.descriptive.moment.GeometricMean]
     * for details on the computing algorithm.
     *
     * @param values the input array
     * @param begin index of the first array element to include
     * @param length the number of elements to include
     * @return the geometric mean of the values or Double.NaN if length = 0
     * @throws MathIllegalArgumentException if the array is null or the array index
     * parameters are not valid
     */
    @Throws(MathIllegalArgumentException::class)
    fun geometricMean(
        values: DoubleArray?, begin: Int,
        length: Int
    ): Double {
        return GEOMETRIC_MEAN.evaluate(values, begin, length)
    }

    /**
     * Returns the variance of the entries in the input array, or
     * `Double.NaN` if the array is empty.
     *
     *
     * This method returns the bias-corrected sample variance (using `n - 1` in
     * the denominator).  Use [.populationVariance] for the non-bias-corrected
     * population variance.
     *
     *
     * See [org.apache.commons.math3.stat.descriptive.moment.Variance] for
     * details on the computing algorithm.
     *
     *
     * Returns 0 for a single-value (i.e. length = 1) sample.
     *
     *
     * Throws `MathIllegalArgumentException` if the array is null.
     *
     * @param values the input array
     * @return the variance of the values or Double.NaN if the array is empty
     * @throws MathIllegalArgumentException if the array is null
     */
    @Throws(MathIllegalArgumentException::class)
    fun variance(values: DoubleArray?): Double {
        return VARIANCE.evaluate(values)
    }

    /**
     * Returns the variance of the entries in the specified portion of
     * the input array, or `Double.NaN` if the designated subarray
     * is empty.
     *
     *
     * This method returns the bias-corrected sample variance (using `n - 1` in
     * the denominator).  Use [.populationVariance] for the non-bias-corrected
     * population variance.
     *
     *
     * See [org.apache.commons.math3.stat.descriptive.moment.Variance] for
     * details on the computing algorithm.
     *
     *
     * Returns 0 for a single-value (i.e. length = 1) sample.
     *
     *
     * Throws `MathIllegalArgumentException` if the array is null or the
     * array index parameters are not valid.
     *
     * @param values the input array
     * @param begin index of the first array element to include
     * @param length the number of elements to include
     * @return the variance of the values or Double.NaN if length = 0
     * @throws MathIllegalArgumentException if the array is null or the array index
     * parameters are not valid
     */
    @Throws(MathIllegalArgumentException::class)
    fun variance(
        values: DoubleArray?, begin: Int,
        length: Int
    ): Double {
        return VARIANCE.evaluate(values, begin, length)
    }

    /**
     * Returns the variance of the entries in the specified portion of
     * the input array, using the precomputed mean value.  Returns
     * `Double.NaN` if the designated subarray is empty.
     *
     *
     * This method returns the bias-corrected sample variance (using `n - 1` in
     * the denominator).  Use [.populationVariance] for the non-bias-corrected
     * population variance.
     *
     *
     * See [org.apache.commons.math3.stat.descriptive.moment.Variance] for
     * details on the computing algorithm.
     *
     *
     * The formula used assumes that the supplied mean value is the arithmetic
     * mean of the sample data, not a known population parameter.  This method
     * is supplied only to save computation when the mean has already been
     * computed.
     *
     *
     * Returns 0 for a single-value (i.e. length = 1) sample.
     *
     *
     * Throws `MathIllegalArgumentException` if the array is null or the
     * array index parameters are not valid.
     *
     * @param values the input array
     * @param mean the precomputed mean value
     * @param begin index of the first array element to include
     * @param length the number of elements to include
     * @return the variance of the values or Double.NaN if length = 0
     * @throws MathIllegalArgumentException if the array is null or the array index
     * parameters are not valid
     */
    @Throws(MathIllegalArgumentException::class)
    fun variance(
        values: DoubleArray?, mean: Double,
        begin: Int, length: Int
    ): Double {
        return VARIANCE.evaluate(values, mean, begin, length)
    }

    /**
     * Returns the variance of the entries in the input array, using the
     * precomputed mean value.  Returns `Double.NaN` if the array
     * is empty.
     *
     *
     * This method returns the bias-corrected sample variance (using `n - 1` in
     * the denominator).  Use [.populationVariance] for the non-bias-corrected
     * population variance.
     *
     *
     * See [org.apache.commons.math3.stat.descriptive.moment.Variance] for
     * details on the computing algorithm.
     *
     *
     * The formula used assumes that the supplied mean value is the arithmetic
     * mean of the sample data, not a known population parameter.  This method
     * is supplied only to save computation when the mean has already been
     * computed.
     *
     *
     * Returns 0 for a single-value (i.e. length = 1) sample.
     *
     *
     * Throws `MathIllegalArgumentException` if the array is null.
     *
     * @param values the input array
     * @param mean the precomputed mean value
     * @return the variance of the values or Double.NaN if the array is empty
     * @throws MathIllegalArgumentException if the array is null
     */
    @Throws(MathIllegalArgumentException::class)
    fun variance(values: DoubleArray?, mean: Double): Double {
        return VARIANCE.evaluate(values, mean)
    }

    /**
     * Returns the [
 * population variance](http://en.wikibooks.org/wiki/Statistics/Summary/Variance) of the entries in the input array, or
     * `Double.NaN` if the array is empty.
     *
     *
     * See [org.apache.commons.math3.stat.descriptive.moment.Variance] for
     * details on the formula and computing algorithm.
     *
     *
     * Returns 0 for a single-value (i.e. length = 1) sample.
     *
     *
     * Throws `MathIllegalArgumentException` if the array is null.
     *
     * @param values the input array
     * @return the population variance of the values or Double.NaN if the array is empty
     * @throws MathIllegalArgumentException if the array is null
     */
    @Throws(MathIllegalArgumentException::class)
    fun populationVariance(values: DoubleArray?): Double {
        return Variance(false).evaluate(values)
    }

    /**
     * Returns the [
 * population variance](http://en.wikibooks.org/wiki/Statistics/Summary/Variance) of the entries in the specified portion of
     * the input array, or `Double.NaN` if the designated subarray
     * is empty.
     *
     *
     * See [org.apache.commons.math3.stat.descriptive.moment.Variance] for
     * details on the computing algorithm.
     *
     *
     * Returns 0 for a single-value (i.e. length = 1) sample.
     *
     *
     * Throws `MathIllegalArgumentException` if the array is null or the
     * array index parameters are not valid.
     *
     * @param values the input array
     * @param begin index of the first array element to include
     * @param length the number of elements to include
     * @return the population variance of the values or Double.NaN if length = 0
     * @throws MathIllegalArgumentException if the array is null or the array index
     * parameters are not valid
     */
    @Throws(MathIllegalArgumentException::class)
    fun populationVariance(
        values: DoubleArray?, begin: Int,
        length: Int
    ): Double {
        return Variance(false).evaluate(values, begin, length)
    }

    /**
     * Returns the [
 * population variance](http://en.wikibooks.org/wiki/Statistics/Summary/Variance) of the entries in the specified portion of
     * the input array, using the precomputed mean value.  Returns
     * `Double.NaN` if the designated subarray is empty.
     *
     *
     * See [org.apache.commons.math3.stat.descriptive.moment.Variance] for
     * details on the computing algorithm.
     *
     *
     * The formula used assumes that the supplied mean value is the arithmetic
     * mean of the sample data, not a known population parameter.  This method
     * is supplied only to save computation when the mean has already been
     * computed.
     *
     *
     * Returns 0 for a single-value (i.e. length = 1) sample.
     *
     *
     * Throws `MathIllegalArgumentException` if the array is null or the
     * array index parameters are not valid.
     *
     * @param values the input array
     * @param mean the precomputed mean value
     * @param begin index of the first array element to include
     * @param length the number of elements to include
     * @return the population variance of the values or Double.NaN if length = 0
     * @throws MathIllegalArgumentException if the array is null or the array index
     * parameters are not valid
     */
    @Throws(MathIllegalArgumentException::class)
    fun populationVariance(
        values: DoubleArray?, mean: Double,
        begin: Int, length: Int
    ): Double {
        return Variance(false).evaluate(values, mean, begin, length)
    }

    /**
     * Returns the [
 * population variance](http://en.wikibooks.org/wiki/Statistics/Summary/Variance) of the entries in the input array, using the
     * precomputed mean value.  Returns `Double.NaN` if the array
     * is empty.
     *
     *
     * See [org.apache.commons.math3.stat.descriptive.moment.Variance] for
     * details on the computing algorithm.
     *
     *
     * The formula used assumes that the supplied mean value is the arithmetic
     * mean of the sample data, not a known population parameter.  This method
     * is supplied only to save computation when the mean has already been
     * computed.
     *
     *
     * Returns 0 for a single-value (i.e. length = 1) sample.
     *
     *
     * Throws `MathIllegalArgumentException` if the array is null.
     *
     * @param values the input array
     * @param mean the precomputed mean value
     * @return the population variance of the values or Double.NaN if the array is empty
     * @throws MathIllegalArgumentException if the array is null
     */
    @Throws(MathIllegalArgumentException::class)
    fun populationVariance(values: DoubleArray?, mean: Double): Double {
        return Variance(false).evaluate(values, mean)
    }

    /**
     * Returns the maximum of the entries in the input array, or
     * `Double.NaN` if the array is empty.
     *
     *
     * Throws `MathIllegalArgumentException` if the array is null.
     *
     *
     *
     *  * The result is `NaN` iff all values are `NaN`
     * (i.e. `NaN` values have no impact on the value of the statistic).
     *  * If any of the values equals `Double.POSITIVE_INFINITY`,
     * the result is `Double.POSITIVE_INFINITY.`
     *
     *
     * @param values the input array
     * @return the maximum of the values or Double.NaN if the array is empty
     * @throws MathIllegalArgumentException if the array is null
     */
    @Throws(MathIllegalArgumentException::class)
    fun max(values: DoubleArray?): Double {
        return MAX.evaluate(values)
    }

    /**
     * Returns the maximum of the entries in the specified portion of
     * the input array, or `Double.NaN` if the designated subarray
     * is empty.
     *
     *
     * Throws `MathIllegalArgumentException` if the array is null or
     * the array index parameters are not valid.
     *
     *
     *
     *  * The result is `NaN` iff all values are `NaN`
     * (i.e. `NaN` values have no impact on the value of the statistic).
     *  * If any of the values equals `Double.POSITIVE_INFINITY`,
     * the result is `Double.POSITIVE_INFINITY.`
     *
     *
     * @param values the input array
     * @param begin index of the first array element to include
     * @param length the number of elements to include
     * @return the maximum of the values or Double.NaN if length = 0
     * @throws MathIllegalArgumentException if the array is null or the array index
     * parameters are not valid
     */
    @Throws(MathIllegalArgumentException::class)
    fun max(
        values: DoubleArray?, begin: Int,
        length: Int
    ): Double {
        return MAX.evaluate(values, begin, length)
    }

    /**
     * Returns the minimum of the entries in the input array, or
     * `Double.NaN` if the array is empty.
     *
     *
     * Throws `MathIllegalArgumentException` if the array is null.
     *
     *
     *
     *  * The result is `NaN` iff all values are `NaN`
     * (i.e. `NaN` values have no impact on the value of the statistic).
     *  * If any of the values equals `Double.NEGATIVE_INFINITY`,
     * the result is `Double.NEGATIVE_INFINITY.`
     *
     *
     * @param values the input array
     * @return the minimum of the values or Double.NaN if the array is empty
     * @throws MathIllegalArgumentException if the array is null
     */
    @Throws(MathIllegalArgumentException::class)
    fun min(values: DoubleArray?): Double {
        return MIN.evaluate(values)
    }

    /**
     * Returns the minimum of the entries in the specified portion of
     * the input array, or `Double.NaN` if the designated subarray
     * is empty.
     *
     *
     * Throws `MathIllegalArgumentException` if the array is null or
     * the array index parameters are not valid.
     *
     *
     *
     *  * The result is `NaN` iff all values are `NaN`
     * (i.e. `NaN` values have no impact on the value of the statistic).
     *  * If any of the values equals `Double.NEGATIVE_INFINITY`,
     * the result is `Double.NEGATIVE_INFINITY.`
     *
     *
     * @param values the input array
     * @param begin index of the first array element to include
     * @param length the number of elements to include
     * @return the minimum of the values or Double.NaN if length = 0
     * @throws MathIllegalArgumentException if the array is null or the array index
     * parameters are not valid
     */
    @Throws(MathIllegalArgumentException::class)
    fun min(
        values: DoubleArray?, begin: Int,
        length: Int
    ): Double {
        return MIN.evaluate(values, begin, length)
    }

    /**
     * Returns an estimate of the `p`th percentile of the values
     * in the `values` array.
     *
     *
     *
     *  * Returns `Double.NaN` if `values` has length
     * `0`
     *  * Returns (for any value of `p`) `values[0]`
     * if `values` has length `1`
     *  * Throws `IllegalArgumentException` if `values`
     * is null  or p is not a valid quantile value (p must be greater than 0
     * and less than or equal to 100)
     *
     *
     *
     * See [org.apache.commons.math3.stat.descriptive.rank.Percentile] for
     * a description of the percentile estimation algorithm used.
     *
     * @param values input array of values
     * @param p the percentile value to compute
     * @return the percentile value or Double.NaN if the array is empty
     * @throws MathIllegalArgumentException if `values` is null
     * or p is invalid
     */
    @Throws(MathIllegalArgumentException::class)
    fun percentile(values: DoubleArray?, p: Double): Double {
        return PERCENTILE.evaluate(values, p)
    }

    /**
     * Returns an estimate of the `p`th percentile of the values
     * in the `values` array, starting with the element in (0-based)
     * position `begin` in the array and including `length`
     * values.
     *
     *
     *
     *  * Returns `Double.NaN` if `length = 0`
     *  * Returns (for any value of `p`) `values[begin]`
     * if `length = 1 `
     *  * Throws `MathIllegalArgumentException` if `values`
     * is null , `begin` or `length` is invalid, or
     * `p` is not a valid quantile value (p must be greater than 0
     * and less than or equal to 100)
     *
     *
     *
     * See [org.apache.commons.math3.stat.descriptive.rank.Percentile] for
     * a description of the percentile estimation algorithm used.
     *
     * @param values array of input values
     * @param p  the percentile to compute
     * @param begin  the first (0-based) element to include in the computation
     * @param length  the number of array elements to include
     * @return  the percentile value
     * @throws MathIllegalArgumentException if the parameters are not valid or the
     * input array is null
     */
    @Throws(MathIllegalArgumentException::class)
    fun percentile(
        values: DoubleArray?, begin: Int,
        length: Int, p: Double
    ): Double {
        return PERCENTILE.evaluate(values, begin, length, p)
    }


     */
    /**
     * Returns the sum of the (signed) differences between corresponding elements of the
     * input arrays -- i.e., sum(sample1[i] - sample2[i]).
     *
     * @param sample1  the first array
     * @param sample2  the second array
     * @return sum of paired differences
     * @throws DimensionMismatchException if the arrays do not have the same
     * (positive) length.
     * @throws NoDataException if the sample arrays are empty.
     */
    //@Throws(DimensionMismatchException::class, NoDataException::class)
    fun sumDifference(sample1: DoubleArray, sample2: DoubleArray): Double {
        val n = sample1.size
        if (n != sample2.size) {
            throw TODO()// DimensionMismatchException(n, sample2.size)
        }
        if (n <= 0) {
            throw TODO()// NoDataException(LocalizedFormats.INSUFFICIENT_DIMENSION)
        }
        var result = 0.0
        for (i in 0 until n) {
            result += sample1[i] - sample2[i]
        }
        return result
    }

    /**
     * Returns the mean of the (signed) differences between corresponding elements of the
     * input arrays -- i.e., sum(sample1[i] - sample2[i]) / sample1.length.
     *
     * @param sample1  the first array
     * @param sample2  the second array
     * @return mean of paired differences
     * @throws DimensionMismatchException if the arrays do not have the same
     * (positive) length.
     * @throws NoDataException if the sample arrays are empty.
     */
    //@Throws(DimensionMismatchException::class, NoDataException::class)
    fun meanDifference(sample1: DoubleArray, sample2: DoubleArray): Double {
        return sumDifference(sample1, sample2) / sample1.size
    }

    /**
     * Returns the variance of the (signed) differences between corresponding elements of the
     * input arrays -- i.e., var(sample1[i] - sample2[i]).
     *
     * @param sample1  the first array
     * @param sample2  the second array
     * @param meanDifference   the mean difference between corresponding entries
     * @see .meanDifference
     * @return variance of paired differences
     * @throws DimensionMismatchException if the arrays do not have the same
     * length.
     * @throws NumberIsTooSmallException if the arrays length is less than 2.
     */
    //@Throws(DimensionMismatchException::class, NumberIsTooSmallException::class)
    fun varianceDifference(
        sample1: DoubleArray,
        sample2: DoubleArray, meanDifference: Double
    ): Double {
        var sum1 = 0.0
        var sum2 = 0.0
        var diff = 0.0
        val n = sample1.size
        if (n != sample2.size) {
            throw TODO()// DimensionMismatchException(n, sample2.size)
        }
        if (n < 2) {
            throw TODO()// NumberIsTooSmallException(n, 2, true)
        }
        for (i in 0 until n) {
            diff = sample1[i] - sample2[i]
            sum1 += (diff - meanDifference) * (diff - meanDifference)
            sum2 += diff - meanDifference
        }
        return (sum1 - sum2 * sum2 / n) / (n - 1)
    }

    /**
     * Normalize (standardize) the sample, so it is has a mean of 0 and a standard deviation of 1.
     *
     * @param sample Sample to normalize.
     * @return normalized (standardized) sample.
     * @since 2.2
     */
    fun normalize(sample: DoubleArray): DoubleArray {
        val stats = DescriptiveStatistics()

        // Add the data from the series to stats
        for (i in sample.indices) {
            stats.addValue(sample[i])
        }

        // Compute mean and standard deviation
        val mean: Double = stats.mean
        val standardDeviation: Double = stats.standardDeviation

        // initialize the standardizedSample, which has the same length as the sample
        val standardizedSample = DoubleArray(sample.size)
        for (i in sample.indices) {
            // z = (x- mean)/standardDeviation
            standardizedSample[i] = (sample[i] - mean) / standardDeviation
        }
        return standardizedSample
    }
    /*
    /**
     * Returns the sample mode(s).  The mode is the most frequently occurring
     * value in the sample. If there is a unique value with maximum frequency,
     * this value is returned as the only element of the output array. Otherwise,
     * the returned array contains the maximum frequency elements in increasing
     * order.  For example, if `sample` is {0, 12, 5, 6, 0, 13, 5, 17},
     * the returned array will have length two, with 0 in the first element and
     * 5 in the second.
     *
     *
     * NaN values are ignored when computing the mode - i.e., NaNs will never
     * appear in the output array.  If the sample includes only NaNs or has
     * length 0, an empty array is returned.
     *
     * @param sample input data
     * @return array of array of the most frequently occurring element(s) sorted in ascending order.
     * @throws MathIllegalArgumentException if the indices are invalid or the array is null
     * @since 3.3
     */
    @Throws(MathIllegalArgumentException::class)
    fun mode(sample: DoubleArray?): DoubleArray {
        if (sample == null) {
            throw NullArgumentException(LocalizedFormats.INPUT_ARRAY)
        }
        return getMode(sample, 0, sample.size)
    }

    /**
     * Returns the sample mode(s).  The mode is the most frequently occurring
     * value in the sample. If there is a unique value with maximum frequency,
     * this value is returned as the only element of the output array. Otherwise,
     * the returned array contains the maximum frequency elements in increasing
     * order.  For example, if `sample` is {0, 12, 5, 6, 0, 13, 5, 17},
     * the returned array will have length two, with 0 in the first element and
     * 5 in the second.
     *
     *
     * NaN values are ignored when computing the mode - i.e., NaNs will never
     * appear in the output array.  If the sample includes only NaNs or has
     * length 0, an empty array is returned.
     *
     * @param sample input data
     * @param begin index (0-based) of the first array element to include
     * @param length the number of elements to include
     *
     * @return array of array of the most frequently occurring element(s) sorted in ascending order.
     * @throws MathIllegalArgumentException if the indices are invalid or the array is null
     * @since 3.3
     */
    fun mode(sample: DoubleArray?, begin: Int, length: Int): DoubleArray {
        if (sample == null) {
            throw NullArgumentException(LocalizedFormats.INPUT_ARRAY)
        }
        if (begin < 0) {
            throw NotPositiveException(LocalizedFormats.START_POSITION, java.lang.Integer.valueOf(begin))
        }
        if (length < 0) {
            throw NotPositiveException(LocalizedFormats.LENGTH, java.lang.Integer.valueOf(length))
        }
        return getMode(sample, begin, length)
    }

    /**
     * Private helper method.
     * Assumes parameters have been validated.
     * @param values input data
     * @param begin index (0-based) of the first array element to include
     * @param length the number of elements to include
     * @return array of array of the most frequently occurring element(s) sorted in ascending order.
     */
    private fun getMode(values: DoubleArray, begin: Int, length: Int): DoubleArray {
        // Add the values to the frequency table
        val freq = Frequency()
        for (i in begin until begin + length) {
            val value = values[i]
            if (!java.lang.Double.isNaN(value)) {
                freq.addValue(java.lang.Double.valueOf(value))
            }
        }
        val list: List<Comparable<*>> = freq.getMode()
        // Convert the list to an array of primitive double
        val modes = DoubleArray(list.size)
        var i = 0
        for (c in list) {
            modes[i++] = c as Double
        }
        return modes
    }

     */
}
