/*
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.jetbrains.kotlinx.statistics.math3.stat.descriptive

import org.jetbrains.kotlinx.statistics.math3.util.Precision
import org.jetbrains.kotlinx.statistics.math3.exception.MathIllegalArgumentException
import org.jetbrains.kotlinx.statistics.math3.util.MathUtils
import org.jetbrains.kotlinx.statistics.math3.stat.descriptive.StorelessUnivariateStatistic

/**
 *
 * Abstract implementation of the [StorelessUnivariateStatistic] interface.
 *
 *
 * Provides default `evaluate()` and `incrementAll(double[])`
 * implementations.
 *
 *
 * **Note that these implementations are not synchronized.**
 *
 */
abstract class AbstractStorelessUnivariateStatistic : AbstractUnivariateStatistic(), StorelessUnivariateStatistic {
    /**
     * This default implementation calls [.clear], then invokes
     * [.increment] in a loop over the the input array, and then uses
     * [.getResult] to compute the return value.
     *
     *
     * Note that this implementation changes the internal state of the
     * statistic.  Its side effects are the same as invoking [.clear] and
     * then [.incrementAll].
     *
     *
     * Implementations may override this method with a more efficient and
     * possibly more accurate implementation that works directly with the
     * input array.
     *
     *
     * If the array is null, a MathIllegalArgumentException is thrown.
     * @param values input array
     * @return the value of the statistic applied to the input array
     * @throws MathIllegalArgumentException if values is null
     * @see org.apache.commons.math3.stat.descriptive.UnivariateStatistic.evaluate
     */
    override fun evaluate(values: DoubleArray): Double {
        return evaluate(values, 0, values.size)
    }

    /**
     * This default implementation calls [.clear], then invokes
     * [.increment] in a loop over the specified portion of the input
     * array, and then uses [.getResult] to compute the return value.
     *
     *
     * Note that this implementation changes the internal state of the
     * statistic.  Its side effects are the same as invoking [.clear] and
     * then [.incrementAll].
     *
     *
     * Implementations may override this method with a more efficient and
     * possibly more accurate implementation that works directly with the
     * input array.
     *
     *
     * If the array is null or the index parameters are not valid, an
     * MathIllegalArgumentException is thrown.
     * @param values the input array
     * @param begin the index of the first element to include
     * @param length the number of elements to include
     * @return the value of the statistic applied to the included array entries
     * @throws MathIllegalArgumentException if the array is null or the indices are not valid
     * @see org.apache.commons.math3.stat.descriptive.UnivariateStatistic.evaluate
     */
    //@Throws(MathIllegalArgumentException::class)
    override fun evaluate(
        values: DoubleArray, begin: Int,
        length: Int
    ): Double {
        if (test(values, begin, length)) {
            clear()
            incrementAll(values, begin, length)
        }
        return result
    }

    /**
     * {@inheritDoc}
     */
    abstract override fun copy(): StorelessUnivariateStatistic

    /**
     * {@inheritDoc}
     */
    abstract override fun clear()

    /**
     * {@inheritDoc}
     */
    abstract override val result: Double

    /**
     * {@inheritDoc}
     */
    abstract override fun increment(d: Double)

    /**
     * This default implementation just calls [.increment] in a loop over
     * the input array.
     *
     *
     * Throws IllegalArgumentException if the input values array is null.
     *
     * @param values values to add
     * @throws MathIllegalArgumentException if values is null
     * @see org.apache.commons.math3.stat.descriptive.StorelessUnivariateStatistic.incrementAll
     */
    //@Throws(MathIllegalArgumentException::class)
    override fun incrementAll(values: DoubleArray) {
        incrementAll(values, 0, values.size)
    }

    /**
     * This default implementation just calls [.increment] in a loop over
     * the specified portion of the input array.
     *
     *
     * Throws IllegalArgumentException if the input values array is null.
     *
     * @param values  array holding values to add
     * @param begin   index of the first array element to add
     * @param length  number of array elements to add
     * @throws MathIllegalArgumentException if values is null
     * @see org.apache.commons.math3.stat.descriptive.StorelessUnivariateStatistic.incrementAll
     */
    //@Throws(MathIllegalArgumentException::class)
    override fun incrementAll(values: DoubleArray, begin: Int, length: Int) {
        if (test(values, begin, length)) {
            val k = begin + length
            for (i in begin until k) {
                increment(values[i])
            }
        }
    }

    /**
     * Returns true iff `object` is an
     * `AbstractStorelessUnivariateStatistic` returning the same
     * values as this for `getResult()` and `getN()`
     * @param other object to test equality against.
     * @return true if object returns the same value as this
     */
    override fun equals(other: Any?): Boolean {
        if (other === this) {
            return true
        }
        if (other !is AbstractStorelessUnivariateStatistic) {
            return false
        }
        return Precision.equalsIncludingNaN(other.result, result) &&
                Precision.equalsIncludingNaN(other.n.toDouble(), this.n.toDouble())
    }

    /**
     * Returns hash code based on getResult() and getN()
     *
     * @return hash code
     */
    override fun hashCode(): Int {
        return 31 * (31 + MathUtils.hash(result)) + MathUtils.hash(n.toDouble())
    }
}
