/*
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.jetbrains.kotlinx.statistics.math3.stat.descriptive

import org.jetbrains.kotlinx.statistics.math3.exception.MathIllegalArgumentException
import org.jetbrains.kotlinx.statistics.math3.stat.descriptive.moment.*
import org.jetbrains.kotlinx.statistics.math3.stat.descriptive.rank.Max
import org.jetbrains.kotlinx.statistics.math3.stat.descriptive.rank.Min
import org.jetbrains.kotlinx.statistics.math3.stat.descriptive.rank.Percentile
import org.jetbrains.kotlinx.statistics.math3.stat.descriptive.moment.*
import org.jetbrains.kotlinx.statistics.math3.stat.descriptive.rank.*
import org.jetbrains.kotlinx.statistics.math3.util.MathUtils
import org.jetbrains.kotlinx.statistics.math3.util.ResizableDoubleArray
import org.jetbrains.kotlinx.statistics.math3.stat.descriptive.summary.Sum
import org.jetbrains.kotlinx.statistics.math3.stat.descriptive.summary.SumOfSquares
import kotlin.math.sqrt

/**
 * Maintains a dataset of values of a single variable and computes descriptive
 * statistics based on stored data. The [windowSize][.getWindowSize]
 * property sets a limit on the number of values that can be stored in the
 * dataset.  The default value, INFINITE_WINDOW, puts no limit on the size of
 * the dataset.  This value should be used with caution, as the backing store
 * will grow without bound in this case.  For very large datasets,
 * [SummaryStatistics], which does not store the dataset, should be used
 * instead of this class. If `windowSize` is not INFINITE_WINDOW and
 * more values are added than can be stored in the dataset, new values are
 * added in a "rolling" manner, with new values replacing the "oldest" values
 * in the dataset.
 *
 *
 * Note: this class is not threadsafe.  Use
 * [SynchronizedDescriptiveStatistics] if concurrent access from multiple
 * threads is required.
 *
 */
open class DescriptiveStatistics : StatisticalSummary {
    /** hold the window size  */
    protected var windowSize = INFINITE_WINDOW
        set(value)  {
            if (value < 1 && value != INFINITE_WINDOW) {
                TODO()
                /*throw MathIllegalArgumentException(
                    LocalizedFormats.NOT_POSITIVE_WINDOW_SIZE, windowSize
                )*/
            }
            field = value

            // We need to check to see if we need to discard elements
            // from the front of the array.  If the windowSize is less than
            // the current number of elements.
            if (value != INFINITE_WINDOW && value < eDA.getNumElements()) {
                eDA.discardFrontElements(eDA.getNumElements() - value)
            }
        }


    /**
     * Stored data values
     */
    private var eDA: ResizableDoubleArray = ResizableDoubleArray()

    /** Mean statistic implementation - can be reset by setter.  */
    private var meanImpl: UnivariateStatistic = Mean()

    /** Geometric mean statistic implementation - can be reset by setter.  */
    private var geometricMeanImpl: UnivariateStatistic = GeometricMean()

    /** Kurtosis statistic implementation - can be reset by setter.  */
    private var kurtosisImpl: UnivariateStatistic = Kurtosis()

    /** Maximum statistic implementation - can be reset by setter.  */
    private var maxImpl: UnivariateStatistic = Max()

    /** Minimum statistic implementation - can be reset by setter.  */
    private var minImpl: UnivariateStatistic = Min()

    /** Percentile statistic implementation - can be reset by setter.  */
    private var percentileImpl: UnivariateStatistic = Percentile()

    /** Skewness statistic implementation - can be reset by setter.  */
    private var skewnessImpl: UnivariateStatistic = Skewness()

    /** Variance statistic implementation - can be reset by setter.  */
    private var varianceImpl: UnivariateStatistic = Variance()

    /** Sum of squares statistic implementation - can be reset by setter.  */
    private var sumsqImpl: UnivariateStatistic = SumOfSquares()

    /** Sum statistic implementation - can be reset by setter.  */
    private var sumImpl: UnivariateStatistic = Sum()

    /**
     * Construct a DescriptiveStatistics instance with an infinite window
     */
    constructor()

    /**
     * Construct a DescriptiveStatistics instance with the specified window
     *
     * @param window the window size.
     * @throws MathIllegalArgumentException if window size is less than 1 but
     * not equal to [.INFINITE_WINDOW]
     */
    constructor(window: Int) {
        windowSize = window
    }

    /**
     * Construct a DescriptiveStatistics instance with an infinite window
     * and the initial data values in double[] initialDoubleArray.
     * If initialDoubleArray is null, then this constructor corresponds to
     * DescriptiveStatistics()
     *
     * @param initialDoubleArray the initial double[].
     */
    constructor(initialDoubleArray: DoubleArray) {
        eDA = ResizableDoubleArray(initialDoubleArray)
    }

    /**
     * Copy constructor.  Construct a new DescriptiveStatistics instance that
     * is a copy of original.
     *
     * @param original DescriptiveStatistics instance to copy
     * @throws NullArgumentException if original is null
     */
    constructor(original: DescriptiveStatistics) {
        copy(original, this)
    }

    /**
     * Adds the value to the dataset. If the dataset is at the maximum size
     * (i.e., the number of stored elements equals the currently configured
     * windowSize), the first (oldest) element in the dataset is discarded
     * to make room for the new value.
     *
     * @param v the value to be added
     */
    fun addValue(v: Double) {
        if (windowSize != INFINITE_WINDOW) {
            if (n == windowSize.toLong()) {
                eDA.addElementRolling(v)
            } else if (n < windowSize) {
                eDA.addElement(v)
            }
        } else {
            eDA.addElement(v)
        }
    }

    /**
     * Removes the most recent value from the dataset.
     *
     * @throws MathIllegalStateException if there are no elements stored
     */
    //@Throws(MathIllegalStateException::class)
    fun removeMostRecentValue() {
        try {
            eDA.discardMostRecentElements(1)
        } catch (ex: MathIllegalArgumentException) {
            TODO()
            //throw MathIllegalStateException(LocalizedFormats.NO_DATA)
        }
    }

    /**
     * Replaces the most recently stored value with the given value.
     * There must be at least one element stored to call this method.
     *
     * @param v the value to replace the most recent stored value
     * @return replaced value
     * @throws MathIllegalStateException if there are no elements stored
     */
    //@Throws(MathIllegalStateException::class)
    fun replaceMostRecentValue(v: Double): Double {
        return eDA.substituteMostRecentElement(v)
    }

    override val mean: Double
        /**
         * Returns the [
 * arithmetic mean ](http://www.xycoon.com/arithmetic_mean.htm) of the available values
         * @return The mean or Double.NaN if no values have been added.
         */
        get() = apply(meanImpl)
    val geometricMean: Double
        /**
         * Returns the [
 * geometric mean ](http://www.xycoon.com/geometric_mean.htm) of the available values.
         *
         *
         * See [GeometricMean] for details on the computing algorithm.
         *
         * @return The geometricMean, Double.NaN if no values have been added,
         * or if any negative values have been added.
         */
        get() = apply(geometricMeanImpl)
    override val variance: Double
        /**
         * Returns the (sample) variance of the available values.
         *
         *
         * This method returns the bias-corrected sample variance (using `n - 1` in
         * the denominator).  Use [.getPopulationVariance] for the non-bias-corrected
         * population variance.
         *
         * @return The variance, Double.NaN if no values have been added
         * or 0.0 for a single value set.
         */
        get() = apply(varianceImpl)
    val populationVariance: Double
        /**
         * Returns the [
 * population variance](http://en.wikibooks.org/wiki/Statistics/Summary/Variance) of the available values.
         *
         * @return The population variance, Double.NaN if no values have been added,
         * or 0.0 for a single value set.
         */
        get() = apply(Variance(false))
    override val standardDeviation: Double
        /**
         * Returns the standard deviation of the available values.
         * @return The standard deviation, Double.NaN if no values have been added
         * or 0.0 for a single value set.
         */
        get() {
            var stdDev = Double.NaN
            if (n > 0) {
                stdDev = if (n > 1) {
                    sqrt(variance)
                } else {
                    0.0
                }
            }
            return stdDev
        }
    val quadraticMean: Double
        /**
         * Returns the quadratic mean, a.k.a.
         * [
 * root-mean-square](http://mathworld.wolfram.com/Root-Mean-Square.html) of the available values
         * @return The quadratic mean or `Double.NaN` if no values
         * have been added.
         */
        get() {
            val n = n
            return if (n > 0) sqrt(sumsq / n) else Double.NaN
        }
    val skewness: Double
        /**
         * Returns the skewness of the available values. Skewness is a
         * measure of the asymmetry of a given distribution.
         *
         * @return The skewness, Double.NaN if less than 3 values have been added.
         */
        get() = apply(skewnessImpl)
    val kurtosis: Double
        /**
         * Returns the Kurtosis of the available values. Kurtosis is a
         * measure of the "peakedness" of a distribution.
         *
         * @return The kurtosis, Double.NaN if less than 4 values have been added.
         */
        get() = apply(kurtosisImpl)
    override val max: Double
        /**
         * Returns the maximum of the available values
         * @return The max or Double.NaN if no values have been added.
         */
        get() = apply(maxImpl)
    override val min: Double
        /**
         * Returns the minimum of the available values
         * @return The min or Double.NaN if no values have been added.
         */
        get() = apply(minImpl)
    override val n: Long
        /**
         * Returns the number of available values
         * @return The number of available values
         */
        get() = eDA.getNumElements().toLong()
    override val sum: Double
        /**
         * Returns the sum of the values that have been added to Univariate.
         * @return The sum or Double.NaN if no values have been added
         */
        get() = apply(sumImpl)
    val sumsq: Double
        /**
         * Returns the sum of the squares of the available values.
         * @return The sum of the squares or Double.NaN if no
         * values have been added.
         */
        get() = apply(sumsqImpl)

    /**
     * Resets all statistics and storage
     */
    fun clear() {
        eDA.clear()
    }

    /**
     * Returns the maximum number of values that can be stored in the
     * dataset, or INFINITE_WINDOW (-1) if there is no limit.
     *
     * @return The current window size or -1 if its Infinite.
     *//*
    fun getWindowSize(): Int {
        return windowSize
    }*/

    /**
     * WindowSize controls the number of values that contribute to the
     * reported statistics.  For example, if windowSize is set to 3 and the
     * values {1,2,3,4,5} have been added ** in that order** then
     * the *available values* are {3,4,5} and all reported statistics will
     * be based on these values. If `windowSize` is decreased as a result
     * of this call and there are more than the new value of elements in the
     * current dataset, values from the front of the array are discarded to
     * reduce the dataset to `windowSize` elements.
     *
     * @param windowSize sets the size of the window.
     * @throws MathIllegalArgumentException if window size is less than 1 but
     * not equal to [.INFINITE_WINDOW]
     */
    /*@Throws(MathIllegalArgumentException::class)
    fun setWindowSize(windowSize: Int) {
        if (windowSize < 1 && windowSize != INFINITE_WINDOW) {
            TODO()
            *//*throw MathIllegalArgumentException(
                LocalizedFormats.NOT_POSITIVE_WINDOW_SIZE, windowSize
            )*//*
        }
        this.windowSize = windowSize

        // We need to check to see if we need to discard elements
        // from the front of the array.  If the windowSize is less than
        // the current number of elements.
        if (windowSize != INFINITE_WINDOW && windowSize < eDA.getNumElements()) {
            eDA.discardFrontElements(eDA.getNumElements() - windowSize)
        }
    }*/

    val values: DoubleArray
        /**
         * Returns the current set of values in an array of double primitives.
         * The order of addition is preserved.  The returned array is a fresh
         * copy of the underlying data -- i.e., it is not a reference to the
         * stored data.
         *
         * @return returns the current set of numbers in the order in which they
         * were added to this set
         */
        get() = eDA.elements
    val sortedValues: DoubleArray
        /**
         * Returns the current set of values in an array of double primitives,
         * sorted in ascending order.  The returned array is a fresh
         * copy of the underlying data -- i.e., it is not a reference to the
         * stored data.
         * @return returns the current set of
         * numbers sorted in ascending order
         */
        get() {
            val sort = values
            sort.sort()
            return sort
        }

    /**
     * Returns the element at the specified index
     * @param index The Index of the element
     * @return return the element at the specified index
     */
    fun getElement(index: Int): Double {
        return eDA.getElement(index)
    }

    /**
     * Returns an estimate for the pth percentile of the stored values.
     *
     *
     * The implementation provided here follows the first estimation procedure presented
     * [here.](http://www.itl.nist.gov/div898/handbook/prc/section2/prc252.htm)
     *
     *
     * **Preconditions**:
     *  * `0 < p  100` (otherwise an
     * `MathIllegalArgumentException` is thrown)
     *  * at least one value must be stored (returns `Double.NaN
    ` *  otherwise)
     *
     *
     * @param p the requested percentile (scaled from 0 - 100)
     * @return An estimate for the pth percentile of the stored data
     * @throws MathIllegalStateException if percentile implementation has been
     * overridden and the supplied implementation does not support setQuantile
     * @throws MathIllegalArgumentException if p is not a valid quantile
     */
    //@Throws(MathIllegalStateException::class, MathIllegalArgumentException::class)
    fun getPercentile(p: Double): Double {
        (percentileImpl as Percentile).setQuantile(p)
        return apply(percentileImpl)
        TODO()/*
        if (percentileImpl is Percentile) {
            (percentileImpl as Percentile).setQuantile(p)
        } else {
            TODO()
            /*try {
                percentileImpl.getClass()
                    .getMethod(SET_QUANTILE_METHOD_NAME, arrayOf<java.lang.Class>(java.lang.Double.TYPE))
                    .invoke(percentileImpl, arrayOf<Any>(java.lang.Double.valueOf(p)))
            } catch (e1: java.lang.NoSuchMethodException) { // Setter guard should prevent
                throw MathIllegalStateException(
                    LocalizedFormats.PERCENTILE_IMPLEMENTATION_UNSUPPORTED_METHOD,
                    percentileImpl.getClass().getName(), SET_QUANTILE_METHOD_NAME
                )
            } catch (e2: java.lang.IllegalAccessException) {
                throw MathIllegalStateException(
                    LocalizedFormats.PERCENTILE_IMPLEMENTATION_CANNOT_ACCESS_METHOD,
                    SET_QUANTILE_METHOD_NAME, percentileImpl.getClass().getName()
                )
            } catch (e3: java.lang.reflect.InvocationTargetException) {
                throw java.lang.IllegalStateException(e3.cause)
            }*/
        }
        return apply(percentileImpl)
        */
    }

    /**
     * Generates a text report displaying univariate statistics from values
     * that have been added.  Each statistic is displayed on a separate
     * line.
     *
     * @return String with line feeds displaying statistics
     */
    override fun toString(): String {
        TODO()
        /*val outBuffer: java.lang.StringBuilder = java.lang.StringBuilder()
        val endl = "\n"
        outBuffer.append("DescriptiveStatistics:").append(endl)
        outBuffer.append("n: ").append(n).append(endl)
        outBuffer.append("min: ").append(min).append(endl)
        outBuffer.append("max: ").append(max).append(endl)
        outBuffer.append("mean: ").append(mean).append(endl)
        outBuffer.append("std dev: ").append(standardDeviation)
            .append(endl)
        try {
            // No catch for MIAE because actual parameter is valid below
            outBuffer.append("median: ").append(getPercentile(50.0)).append(endl)
        } catch (ex: MathIllegalStateException) {
            outBuffer.append("median: unavailable").append(endl)
        }
        outBuffer.append("skewness: ").append(skewness).append(endl)
        outBuffer.append("kurtosis: ").append(kurtosis).append(endl)
        return outBuffer.toString()*/
    }

    /**
     * Apply the given statistic to the data associated with this set of statistics.
     * @param stat the statistic to apply
     * @return the computed value of the statistic.
     */
    fun apply(stat: UnivariateStatistic): Double {
        // No try-catch or advertised exception here because arguments are guaranteed valid
        return eDA.compute(stat)
    }
    // Implementation getters and setter
    /**
     * Returns the currently configured mean implementation.
     *
     * @return the UnivariateStatistic implementing the mean
     * @since 1.2
     */
    
    fun getMeanImpl(): UnivariateStatistic {
        return meanImpl
    }

    /**
     *
     * Sets the implementation for the mean.
     *
     * @param meanImpl the UnivariateStatistic instance to use
     * for computing the mean
     * @since 1.2
     */
    
    fun setMeanImpl(meanImpl: UnivariateStatistic) {
        this.meanImpl = meanImpl
    }

    /**
     * Returns the currently configured geometric mean implementation.
     *
     * @return the UnivariateStatistic implementing the geometric mean
     * @since 1.2
     */
    
    fun getGeometricMeanImpl(): UnivariateStatistic {
        return geometricMeanImpl
    }

    /**
     *
     * Sets the implementation for the gemoetric mean.
     *
     * @param geometricMeanImpl the UnivariateStatistic instance to use
     * for computing the geometric mean
     * @since 1.2
     */
    
    fun setGeometricMeanImpl(
        geometricMeanImpl: UnivariateStatistic
    ) {
        this.geometricMeanImpl = geometricMeanImpl
    }

    /**
     * Returns the currently configured kurtosis implementation.
     *
     * @return the UnivariateStatistic implementing the kurtosis
     * @since 1.2
     */
    
    fun getKurtosisImpl(): UnivariateStatistic {
        return kurtosisImpl
    }

    /**
     *
     * Sets the implementation for the kurtosis.
     *
     * @param kurtosisImpl the UnivariateStatistic instance to use
     * for computing the kurtosis
     * @since 1.2
     */
    
    fun setKurtosisImpl(kurtosisImpl: UnivariateStatistic) {
        this.kurtosisImpl = kurtosisImpl
    }

    /**
     * Returns the currently configured maximum implementation.
     *
     * @return the UnivariateStatistic implementing the maximum
     * @since 1.2
     */
    
    fun getMaxImpl(): UnivariateStatistic {
        return maxImpl
    }

    /**
     *
     * Sets the implementation for the maximum.
     *
     * @param maxImpl the UnivariateStatistic instance to use
     * for computing the maximum
     * @since 1.2
     */
    
    fun setMaxImpl(maxImpl: UnivariateStatistic) {
        this.maxImpl = maxImpl
    }

    /**
     * Returns the currently configured minimum implementation.
     *
     * @return the UnivariateStatistic implementing the minimum
     * @since 1.2
     */
    
    fun getMinImpl(): UnivariateStatistic {
        return minImpl
    }

    /**
     *
     * Sets the implementation for the minimum.
     *
     * @param minImpl the UnivariateStatistic instance to use
     * for computing the minimum
     * @since 1.2
     */
    
    fun setMinImpl(minImpl: UnivariateStatistic) {
        this.minImpl = minImpl
    }

    /**
     * Returns the currently configured percentile implementation.
     *
     * @return the UnivariateStatistic implementing the percentile
     * @since 1.2
     */
    
    fun getPercentileImpl(): UnivariateStatistic {
        return percentileImpl
    }

    /**
     * Sets the implementation to be used by [.getPercentile].
     * The supplied `UnivariateStatistic` must provide a
     * `setQuantile(double)` method; otherwise
     * `IllegalArgumentException` is thrown.
     *
     * @param percentileImpl the percentileImpl to set
     * @throws MathIllegalArgumentException if the supplied implementation does not
     * provide a `setQuantile` method
     * @since 1.2
     */
    
    /* TODO @Throws(MathIllegalArgumentException::class)
    fun setPercentileImpl(percentileImpl: UnivariateStatistic) {
        try {
            percentileImpl.getClass()
                .getMethod(SET_QUANTILE_METHOD_NAME, arrayOf<java.lang.Class>(java.lang.Double.TYPE))
                .invoke(percentileImpl, arrayOf<Any>(java.lang.Double.valueOf(50.0)))
        } catch (e1: java.lang.NoSuchMethodException) {
            throw MathIllegalArgumentException(
                LocalizedFormats.PERCENTILE_IMPLEMENTATION_UNSUPPORTED_METHOD,
                percentileImpl.getClass().getName(), SET_QUANTILE_METHOD_NAME
            )
        } catch (e2: java.lang.IllegalAccessException) {
            throw MathIllegalArgumentException(
                LocalizedFormats.PERCENTILE_IMPLEMENTATION_CANNOT_ACCESS_METHOD,
                SET_QUANTILE_METHOD_NAME, percentileImpl.getClass().getName()
            )
        } catch (e3: java.lang.reflect.InvocationTargetException) {
            throw java.lang.IllegalArgumentException(e3.cause)
        }
        this.percentileImpl = percentileImpl
    }*/

    /**
     * Returns the currently configured skewness implementation.
     *
     * @return the UnivariateStatistic implementing the skewness
     * @since 1.2
     */
    
    fun getSkewnessImpl(): UnivariateStatistic {
        return skewnessImpl
    }

    /**
     *
     * Sets the implementation for the skewness.
     *
     * @param skewnessImpl the UnivariateStatistic instance to use
     * for computing the skewness
     * @since 1.2
     */
    
    fun setSkewnessImpl(
        skewnessImpl: UnivariateStatistic
    ) {
        this.skewnessImpl = skewnessImpl
    }

    /**
     * Returns the currently configured variance implementation.
     *
     * @return the UnivariateStatistic implementing the variance
     * @since 1.2
     */
    
    fun getVarianceImpl(): UnivariateStatistic {
        return varianceImpl
    }

    /**
     *
     * Sets the implementation for the variance.
     *
     * @param varianceImpl the UnivariateStatistic instance to use
     * for computing the variance
     * @since 1.2
     */
    
    fun setVarianceImpl(
        varianceImpl: UnivariateStatistic
    ) {
        this.varianceImpl = varianceImpl
    }

    /**
     * Returns the currently configured sum of squares implementation.
     *
     * @return the UnivariateStatistic implementing the sum of squares
     * @since 1.2
     */
    
    fun getSumsqImpl(): UnivariateStatistic {
        return sumsqImpl
    }

    /**
     *
     * Sets the implementation for the sum of squares.
     *
     * @param sumsqImpl the UnivariateStatistic instance to use
     * for computing the sum of squares
     * @since 1.2
     */
    
    fun setSumsqImpl(sumsqImpl: UnivariateStatistic) {
        this.sumsqImpl = sumsqImpl
    }

    /**
     * Returns the currently configured sum implementation.
     *
     * @return the UnivariateStatistic implementing the sum
     * @since 1.2
     */
    
    fun getSumImpl(): UnivariateStatistic {
        return sumImpl
    }

    /**
     *
     * Sets the implementation for the sum.
     *
     * @param sumImpl the UnivariateStatistic instance to use
     * for computing the sum
     * @since 1.2
     */
    
    fun setSumImpl(sumImpl: UnivariateStatistic) {
        this.sumImpl = sumImpl
    }

    /**
     * Returns a copy of this DescriptiveStatistics instance with the same internal state.
     *
     * @return a copy of this
     */
    fun copy(): DescriptiveStatistics {
        val result = DescriptiveStatistics()
        // No try-catch or advertised exception because parms are guaranteed valid
        copy(this, result)
        return result
    }

    companion object {
        /**
         * Represents an infinite window size.  When the [.getWindowSize]
         * returns this value, there is no limit to the number of data values
         * that can be stored in the dataset.
         */
        const val INFINITE_WINDOW = -1

        /** Serialization UID  */
        private const val serialVersionUID = 4133067267405273064L

        /** Name of the setQuantile method.  */
        private const val SET_QUANTILE_METHOD_NAME = "setQuantile"

        /**
         * Copies source to dest.
         *
         * Neither source nor dest can be null.
         *
         * @param source DescriptiveStatistics to copy
         * @param dest DescriptiveStatistics to copy to
         * @throws NullArgumentException if either source or dest is null
         */
        //@Throws(NullArgumentException::class)
        fun copy(source: DescriptiveStatistics, dest: DescriptiveStatistics) {
             MathUtils.checkNotNull(source)
            MathUtils.checkNotNull(dest)
            // Copy data and window size
            dest.eDA = source.eDA.copy()
            dest.windowSize = source.windowSize

            // Copy implementations
            dest.maxImpl = source.maxImpl.copy()
            dest.meanImpl = source.meanImpl.copy()
            dest.minImpl = source.minImpl.copy()
            dest.sumImpl = source.sumImpl.copy()
            dest.varianceImpl = source.varianceImpl.copy()
            dest.sumsqImpl = source.sumsqImpl.copy()
            dest.geometricMeanImpl = source.geometricMeanImpl.copy()
            dest.kurtosisImpl = source.kurtosisImpl
            dest.skewnessImpl = source.skewnessImpl
            dest.percentileImpl = source.percentileImpl
        }
    }
}
