/*
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.jetbrains.kotlinx.statistics.math3.stat.descriptive.moment

import org.jetbrains.kotlinx.statistics.math3.exception.NullArgumentException
import org.jetbrains.kotlinx.statistics.math3.stat.descriptive.AbstractStorelessUnivariateStatistic
import org.jetbrains.kotlinx.statistics.math3.util.MathUtils


/**
 * Computes the first moment (arithmetic mean).  Uses the definitional formula:
 *
 *
 * mean = sum(x_i) / n
 *
 *
 * where `n` is the number of observations.
 *
 *
 * To limit numeric errors, the value of the statistic is computed using the
 * following recursive updating algorithm:
 *
 *
 *
 *  1. Initialize `m = ` the first value
 *  1. For each additional value, update using <br></br>
 * `m = m + (new value - m) / (number of observations)`
 *
 *
 *
 * Returns `Double.NaN` if the dataset is empty. Note that
 * Double.NaN may also be returned if the input includes NaN and / or infinite
 * values.
 *
 *
 * **Note that this implementation is not synchronized.** If
 * multiple threads access an instance of this class concurrently, and at least
 * one of the threads invokes the `increment()` or
 * `clear()` method, it must be synchronized externally.
 *
 */
open class FirstMoment : AbstractStorelessUnivariateStatistic {
    /**
     * {@inheritDoc}
     */
    /** Count of values that have been added  */
    override var n: Long = 0
        protected set
    /**
     * {@inheritDoc}
     */
    /** First moment of values that have been added  */
     var m1 = 0.0
        protected set

    override val result: Double
        get() = m1


    /**
     * Deviation of most recently added value from previous first moment.
     * Retained to prevent repeated computation in higher order moments.
     */
    protected var dev = 0.0

    /**
     * Deviation of most recently added value from previous first moment,
     * normalized by previous sample size.  Retained to prevent repeated
     * computation in higher order moments
     */
    protected var nDev = 0.0

    /**
     * Create a FirstMoment instance
     */
    constructor() {
        n = 0
        m1 = Double.NaN
        dev = Double.NaN
        nDev = Double.NaN
    }

    /**
     * Copy constructor, creates a new `FirstMoment` identical
     * to the `original`
     *
     * @param original the `FirstMoment` instance to copy
     * @throws NullArgumentException if original is null
     */
    constructor(original: FirstMoment) : super() {
        copy(original, this)
    }

    /**
     * {@inheritDoc}
     */
    override fun increment(d: Double) {
        if (n == 0L) {
            m1 = 0.0
        }
        n++
        val n0 = n.toDouble()
        dev = d - m1
        nDev = dev / n0
        m1 += nDev
    }


    /**
     * {@inheritDoc}
     */
    override fun clear() {
        m1 = Double.NaN
        n = 0
        dev = Double.NaN
        nDev = Double.NaN
    }


    /**
     * {@inheritDoc}
     */
    override fun copy(): FirstMoment {
        val result = FirstMoment()
        // No try-catch or advertised exception because args are guaranteed non-null
        copy(this, result)
        return result
    }

    companion object {
        /** Serializable version identifier  */
        private const val serialVersionUID = 6112755307178490473L

        /**
         * Copies source to dest.
         *
         * Neither source nor dest can be null.
         *
         * @param source FirstMoment to copy
         * @param dest FirstMoment to copy to
         * @throws NullArgumentException if either source or dest is null
         */
        //@Throws(org.apache.commons.math3.exception.NullArgumentException::class)
        fun copy(source: FirstMoment, dest: FirstMoment) {
            MathUtils.checkNotNull(source)
            MathUtils.checkNotNull(dest)
            dest.setData(source.dataRef)
            dest.n = source.n
            dest.m1 = source.result
            dest.dev = source.dev
            dest.nDev = source.nDev
        }
    }
}


