/*
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.jetbrains.kotlinx.statistics.math3.stat.descriptive.moment

import org.jetbrains.kotlinx.statistics.math3.exception.NullArgumentException
import org.jetbrains.kotlinx.statistics.math3.util.MathUtils.checkNotNull

/**
 * Computes a statistic related to the Fourth Central Moment.  Specifically,
 * what is computed is the sum of
 *
 *
 * (x_i - xbar) ^ 4,
 *
 *
 * where the x_i are the
 * sample observations and xbar is the sample mean.
 *
 *
 * The following recursive updating formula is used:
 *
 *
 * Let
 *  *  dev = (current obs - previous mean)
 *  *  m2 = previous value of [SecondMoment]
 *  *  m2 = previous value of [ThirdMoment]
 *  *  n = number of observations (including current obs)
 *
 * Then
 *
 *
 * new value = old value - 4 * (dev/n) * m3 + 6 * (dev/n)^2 * m2 + <br></br>
 * [n^2 - 3 * (n-1)] * dev^4 * (n-1) / n^3
 *
 *
 * Returns `Double.NaN` if no data values have been added and
 * returns `0` if there is just one value in the data set. Note that
 * Double.NaN may also be returned if the input includes NaN and / or infinite
 * values.
 *
 *
 * **Note that this implementation is not synchronized.** If
 * multiple threads access an instance of this class concurrently, and at least
 * one of the threads invokes the `increment()` or
 * `clear()` method, it must be synchronized externally.
 *
 */
class FourthMoment : ThirdMoment {
    /**
     * {@inheritDoc}
     */
    /** fourth moment of values that have been added  */
    var m4: Double = Double.NaN
        protected set

    override val result: Double
        get() = m4

    /**
     * Create a FourthMoment instance
     */
    constructor() : super() {
        m4 = Double.NaN
    }

    /**
     * Copy constructor, creates a new `FourthMoment` identical
     * to the `original`
     *
     * @param original the `FourthMoment` instance to copy
     * @throws NullArgumentException if original is null
     */
    constructor(original: FourthMoment) : super() {
        copy(original, this)
    }

    /**
     * {@inheritDoc}
     */
    override fun increment(d: Double) {
        if (n < 1) {
            m4 = 0.0
            m3 = 0.0
            m2 = 0.0
            m1 = 0.0
        }
        val prevM3: Double = m3
        val prevM2: Double = m2
        super.increment(d)
        val n0: Double = n.toDouble()
        m4 = m4 - 4.0 * nDev * prevM3 + 6.0 * nDevSq * prevM2 + (n0 * n0 - 3 * (n0 - 1.0)) * (nDevSq * nDevSq * (n0 - 1) * n0)
    }

    /**
     * {@inheritDoc}
     */
    override fun clear() {
        super.clear()
        m4 = Double.NaN
    }

    /**
     * {@inheritDoc}
     */
    override fun copy(): FourthMoment {
        val m4 = FourthMoment()
        // No try-catch or advertised exception because args are guaranteed non-null
        copy(this, m4)
        return m4
    }

    companion object {
        /** Serializable version identifier  */
        private const val serialVersionUID = 4763990447117157611L

        /**
         * Copies source to dest.
         *
         * Neither source nor dest can be null.
         *
         * @param source FourthMoment to copy
         * @param dest FourthMoment to copy to
         * @throws NullArgumentException if either source or dest is null
         */
        @Throws(NullArgumentException::class)
        fun copy(source: FourthMoment, dest: FourthMoment) {
            checkNotNull(source)
            checkNotNull(dest)
            ThirdMoment.copy(source, dest)
            dest.m4 = source.m4
        }
    }
}
