/*
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.jetbrains.kotlinx.statistics.math3.stat.descriptive.moment

import org.jetbrains.kotlinx.statistics.math3.exception.MathIllegalArgumentException
import org.jetbrains.kotlinx.statistics.math3.exception.NullArgumentException
import org.jetbrains.kotlinx.statistics.math3.stat.descriptive.AbstractStorelessUnivariateStatistic
import org.jetbrains.kotlinx.statistics.math3.stat.descriptive.StorelessUnivariateStatistic
import org.jetbrains.kotlinx.statistics.math3.stat.descriptive.summary.SumOfLogs
import org.jetbrains.kotlinx.statistics.math3.util.MathUtils

/**
 * Returns the [
 * geometric mean ](http://www.xycoon.com/geometric_mean.htm) of the available values.
 *
 *
 * Uses a [SumOfLogs] instance to compute sum of logs and returns
 * ` exp( 1/n  (sum of logs) ).`  Therefore,
 *
 *  * If any of values are < 0, the result is `NaN.`
 *  * If all values are non-negative and less than
 * `Double.POSITIVE_INFINITY`,  but at least one value is 0, the
 * result is `0.`
 *  * If both `Double.POSITIVE_INFINITY` and
 * `Double.NEGATIVE_INFINITY` are among the values, the result is
 * `NaN.`
 *
 *
 *
 * **Note that this implementation is not synchronized.** If
 * multiple threads access an instance of this class concurrently, and at least
 * one of the threads invokes the `increment()` or
 * `clear()` method, it must be synchronized externally.
 *
 *
 */
class GeometricMean : AbstractStorelessUnivariateStatistic {
    /** Wrapped SumOfLogs instance  */
    private var sumOfLogs: StorelessUnivariateStatistic? = null

    /**
     * Create a GeometricMean instance
     */
    constructor() {
        sumOfLogs = SumOfLogs()
    }

    /**
     * Copy constructor, creates a new `GeometricMean` identical
     * to the `original`
     *
     * @param original the `GeometricMean` instance to copy
     * @throws NullArgumentException if original is null
     */
    constructor(original: GeometricMean) : super() {
        copy(original, this)
    }

    /**
     * Create a GeometricMean instance using the given SumOfLogs instance
     * @param sumOfLogs sum of logs instance to use for computation
     */
    constructor(sumOfLogs: SumOfLogs?) {
        this.sumOfLogs = sumOfLogs
    }

    /**
     * {@inheritDoc}
     */
    override fun copy(): GeometricMean {
        val result = GeometricMean()
        // no try-catch or advertised exception because args guaranteed non-null
        copy(this, result)
        return result
    }

    /**
     * {@inheritDoc}
     */
    override fun increment(d: Double) {
        sumOfLogs!!.increment(d)
    }

    override val result: Double
        /**
         * {@inheritDoc}
         */
        get() = if (sumOfLogs!!.n > 0) {
            kotlin.math.exp(sumOfLogs!!.result / sumOfLogs!!.n)
        } else {
            Double.NaN
        }

    /**
     * {@inheritDoc}
     */
    override fun clear() {
        sumOfLogs!!.clear()
    }

    /**
     * Returns the geometric mean of the entries in the specified portion
     * of the input array.
     *
     *
     * See [GeometricMean] for details on the computing algorithm.
     *
     *
     * Throws `IllegalArgumentException` if the array is null.
     *
     * @param values input array containing the values
     * @param begin first array element to include
     * @param length the number of elements to include
     * @return the geometric mean or Double.NaN if length = 0 or
     * any of the values are &lt;= 0.
     * @throws MathIllegalArgumentException if the input array is null or the array
     * index parameters are not valid
     */
    //@Throws(MathIllegalArgumentException::class)
    override fun evaluate(
        values: DoubleArray, begin: Int, length: Int
    ): Double {
        return kotlin.math.exp(
            sumOfLogs!!.evaluate(values, begin, length) / length
        )
    }

    override val n: Long
        /**
         * {@inheritDoc}
         */
        get() = sumOfLogs!!.n

    //@set:Throws(MathIllegalStateException::class)
    var sumLogImpl: StorelessUnivariateStatistic?
        /**
         * Returns the currently configured sum of logs implementation
         *
         * @return the StorelessUnivariateStatistic implementing the log sum
         */
        get() = sumOfLogs
        /**
         *
         * Sets the implementation for the sum of logs.
         *
         * This method must be activated before any data has been added - i.e.,
         * before [increment][.increment] has been used to add data;
         * otherwise an IllegalStateException will be thrown.
         *
         * @param sumLogImpl the StorelessUnivariateStatistic instance to use
         * for computing the log sum
         * @throws MathIllegalStateException if data has already been added
         * (i.e if n > 0)
         */
        set(sumLogImpl) {
            checkEmpty()
            sumOfLogs = sumLogImpl
        }

    /**
     * Throws MathIllegalStateException if n > 0.
     * @throws MathIllegalStateException if data has been added to this statistic
     */
    //@Throws(MathIllegalStateException::class)
    private fun checkEmpty() {
        if (n > 0) {
            TODO()
            /*throw MathIllegalStateException(
                LocalizedFormats.VALUES_ADDED_BEFORE_CONFIGURING_STATISTIC,
                n
            )*/
        }
    }

    companion object {
        /** Serializable version identifier  */
        private const val serialVersionUID = -8178734905303459453L

        /**
         * Copies source to dest.
         *
         * Neither source nor dest can be null.
         *
         * @param source GeometricMean to copy
         * @param dest GeometricMean to copy to
         * @throws NullArgumentException if either source or dest is null
         */
        @Throws(NullArgumentException::class)
        fun copy(source: GeometricMean, dest: GeometricMean) {
            MathUtils.checkNotNull(source)
            MathUtils.checkNotNull(dest)
            dest.setData(source.dataRef)
            dest.sumOfLogs = source.sumOfLogs!!.copy()
        }
    }
}
