/*
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.jetbrains.kotlinx.statistics.math3.stat.descriptive.moment

import org.jetbrains.kotlinx.statistics.math3.exception.MathIllegalArgumentException
import org.jetbrains.kotlinx.statistics.math3.exception.NullArgumentException
import org.jetbrains.kotlinx.statistics.math3.stat.descriptive.AbstractStorelessUnivariateStatistic
import org.jetbrains.kotlinx.statistics.math3.util.MathUtils
import kotlin.math.pow
import kotlin.math.sqrt

/**
 * Computes the Kurtosis of the available values.
 *
 *
 * We use the following (unbiased) formula to define kurtosis:
 *
 *
 * kurtosis = { [n(n+1) / (n -1)(n - 2)(n-3)] sum[(x_i - mean)^4] / std^4 } - [3(n-1)^2 / (n-2)(n-3)]
 *
 *
 * where n is the number of values, mean is the [Mean] and std is the
 * [StandardDeviation]
 *
 *
 * Note that this statistic is undefined for n < 4.  `Double.Nan`
 * is returned when there is not sufficient data to compute the statistic.
 * Note that Double.NaN may also be returned if the input includes NaN
 * and / or infinite values.
 *
 *
 * **Note that this implementation is not synchronized.** If
 * multiple threads access an instance of this class concurrently, and at least
 * one of the threads invokes the `increment()` or
 * `clear()` method, it must be synchronized externally.
 *
 */
class Kurtosis : AbstractStorelessUnivariateStatistic {
    /**Fourth Moment on which this statistic is based  */
    protected lateinit var moment: FourthMoment

    /**
     * Determines whether or not this statistic can be incremented or cleared.
     *
     *
     * Statistics based on (constructed from) external moments cannot
     * be incremented or cleared.
     */
    protected var incMoment = false

    /**
     * Construct a Kurtosis
     */
    constructor() {
        incMoment = true
        moment = FourthMoment()
    }

    /**
     * Construct a Kurtosis from an external moment
     *
     * @param m4 external Moment
     */
    constructor(m4: FourthMoment) {
        incMoment = false
        moment = m4
    }

    /**
     * Copy constructor, creates a new `Kurtosis` identical
     * to the `original`
     *
     * @param original the `Kurtosis` instance to copy
     * @throws NullArgumentException if original is null
     */
    constructor(original: Kurtosis) {
        copy(original, this)
    }

    /**
     * {@inheritDoc}
     *
     * Note that when [.Kurtosis] is used to
     * create a Variance, this method does nothing. In that case, the
     * FourthMoment should be incremented directly.
     */
    override fun increment(d: Double) {
        if (incMoment) {
            moment.increment(d)
        }
    }

    override val result: Double
        /**
         * {@inheritDoc}
         */
        get() {
            var kurtosis = Double.NaN
            if (moment.n > 3) {
                val variance: Double = moment.m2 / (moment.n - 1)
                kurtosis = if (moment.n <= 3 || variance < 10E-20) {
                    0.0
                } else {
                    val n: Double = moment.n.toDouble()
                    (n * (n + 1) * moment.result -
                            3 * moment.m2 * moment.m2 * (n - 1)) /
                            ((n - 1) * (n - 2) * (n - 3) * variance * variance)
                }
            }
            return kurtosis
        }

    /**
     * {@inheritDoc}
     */
    override fun clear() {
        if (incMoment) {
            moment.clear()
        }
    }

    override val n: Long
        /**
         * {@inheritDoc}
         */
        get() = moment.n
    /* UnvariateStatistic Approach  */
    /**
     * Returns the kurtosis of the entries in the specified portion of the
     * input array.
     *
     *
     * See [Kurtosis] for details on the computing algorithm.
     *
     *
     * Throws `IllegalArgumentException` if the array is null.
     *
     * @param values the input array
     * @param begin index of the first array element to include
     * @param length the number of elements to include
     * @return the kurtosis of the values or Double.NaN if length is less than 4
     * @throws MathIllegalArgumentException if the input array is null or the array
     * index parameters are not valid
     */
    //@Throws(MathIllegalArgumentException::class)
    override fun evaluate(values: DoubleArray, begin: Int, length: Int): Double {
        // Initialize the kurtosis
        var kurt = Double.NaN
        if (test(values, begin, length) && length > 3) {

            // Compute the mean and standard deviation
            val variance = Variance()
            variance.incrementAll(values, begin, length)

            val mean = variance.moment.m1
            val stdDev: Double = sqrt(variance.result)

            // Sum the ^4 of the distance from the mean divided by the
            // standard deviation
            var accum3 = 0.0
            for (i in begin until begin + length) {
                accum3 += (values[i] - mean).pow( 4.0)
            }
            accum3 /= stdDev.pow(4.0)

            // Get N
            val n0 = length.toDouble()
            val coefficientOne = n0 * (n0 + 1) / ((n0 - 1) * (n0 - 2) * (n0 - 3))
            val termTwo: Double = 3 * (n0 - 1).pow(2.0) / ((n0 - 2) * (n0 - 3))

            // Calculate kurtosis
            kurt = coefficientOne * accum3 - termTwo
        }
        return kurt
    }

    /**
     * {@inheritDoc}
     */
    override fun copy(): Kurtosis {
        val result = Kurtosis()
        // No try-catch because args are guaranteed non-null
        copy(this, result)
        return result
    }

    companion object {
        /** Serializable version identifier  */
        private const val serialVersionUID = 2784465764798260919L

        /**
         * Copies source to dest.
         *
         * Neither source nor dest can be null.
         *
         * @param source Kurtosis to copy
         * @param dest Kurtosis to copy to
         * @throws NullArgumentException if either source or dest is null
         */
        @Throws(NullArgumentException::class)
        fun copy(source: Kurtosis, dest: Kurtosis) {
            MathUtils.checkNotNull(source)
            MathUtils.checkNotNull(dest)
            dest.setData(source.dataRef)
            dest.moment = source.moment.copy()
            dest.incMoment = source.incMoment
        }
    }
}
