/*
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.jetbrains.kotlinx.statistics.math3.stat.descriptive.moment

import org.jetbrains.kotlinx.statistics.math3.stat.descriptive.WeightedEvaluation
import org.jetbrains.kotlinx.statistics.math3.exception.MathIllegalArgumentException
import org.jetbrains.kotlinx.statistics.math3.exception.NullArgumentException
import org.jetbrains.kotlinx.statistics.math3.stat.descriptive.AbstractStorelessUnivariateStatistic
import org.jetbrains.kotlinx.statistics.math3.stat.descriptive.summary.Sum
import org.jetbrains.kotlinx.statistics.math3.util.MathUtils

/**
 *
 * Computes the arithmetic mean of a set of values. Uses the definitional
 * formula:
 *
 *
 * mean = sum(x_i) / n
 *
 *
 * where `n` is the number of observations.
 *
 *
 * When [.increment] is used to add data incrementally from a
 * stream of (unstored) values, the value of the statistic that
 * [.getResult] returns is computed using the following recursive
 * updating algorithm:
 *
 *  1. Initialize `m = ` the first value
 *  1. For each additional value, update using <br></br>
 * `m = m + (new value - m) / (number of observations)`
 *
 *
 *  If [.evaluate] is used to compute the mean of an array
 * of stored values, a two-pass, corrected algorithm is used, starting with
 * the definitional formula computed using the array of stored values and then
 * correcting this by adding the mean deviation of the data values from the
 * arithmetic mean. See, e.g. "Comparison of Several Algorithms for Computing
 * Sample Means and Variances," Robert F. Ling, Journal of the American
 * Statistical Association, Vol. 69, No. 348 (Dec., 1974), pp. 859-866.
 *
 *
 * Returns `Double.NaN` if the dataset is empty. Note that
 * Double.NaN may also be returned if the input includes NaN and / or infinite
 * values.
 *
 * **Note that this implementation is not synchronized.** If
 * multiple threads access an instance of this class concurrently, and at least
 * one of the threads invokes the `increment()` or
 * `clear()` method, it must be synchronized externally.
 *
 */
class Mean : AbstractStorelessUnivariateStatistic, WeightedEvaluation {
    /** First moment on which this statistic is based.  */
    protected var moment: FirstMoment? = null

    /**
     * Determines whether or not this statistic can be incremented or cleared.
     *
     *
     * Statistics based on (constructed from) external moments cannot
     * be incremented or cleared.
     */
    protected var incMoment = false

    /** Constructs a Mean.  */
    constructor() {
        incMoment = true
        moment = FirstMoment()
    }

    /**
     * Constructs a Mean with an External Moment.
     *
     * @param m1 the moment
     */
    constructor(m1: FirstMoment) {
        moment = m1
        incMoment = false
    }

    /**
     * Copy constructor, creates a new `Mean` identical
     * to the `original`
     *
     * @param original the `Mean` instance to copy
     * @throws NullArgumentException if original is null
     */
    constructor(original: Mean) {
        copy(original, this)
    }

    /**
     * {@inheritDoc}
     *
     * Note that when [.Mean] is used to
     * create a Mean, this method does nothing. In that case, the
     * FirstMoment should be incremented directly.
     */
    override fun increment(d: Double) {
        if (incMoment) {
            moment!!.increment(d)
        }
    }

    override fun incrementAll(values: DoubleArray, start: Int, length: Int) {
        TODO("Not yet implemented")
    }

    /**
     * {@inheritDoc}
     */
    override fun clear() {
        if (incMoment) {
            moment!!.clear()
        }
    }

    override val result: Double
        /**
         * {@inheritDoc}
         */
        get() = moment!!.m1
    override val n: Long
        /**
         * {@inheritDoc}
         */
        get() = moment!!.n

    /**
     * Returns the arithmetic mean of the entries in the specified portion of
     * the input array, or `Double.NaN` if the designated subarray
     * is empty.
     *
     *
     * Throws `IllegalArgumentException` if the array is null.
     *
     *
     * See [Mean] for details on the computing algorithm.
     *
     * @param values the input array
     * @param begin index of the first array element to include
     * @param length the number of elements to include
     * @return the mean of the values or Double.NaN if length = 0
     * @throws MathIllegalArgumentException if the array is null or the array index
     * parameters are not valid
     */
    //@Throws(MathIllegalArgumentException::class)
    override fun evaluate(values: DoubleArray, begin: Int, length: Int): Double {
        if (test(values, begin, length)) {
            val sum = Sum()
            val sampleSize = length.toDouble()

            // Compute initial estimate using definitional formula
            val xbar: Double = sum.evaluate(values, begin, length) / sampleSize

            // Compute correction factor in second pass
            var correction = 0.0
            for (i in begin until begin + length) {
                correction += values[i] - xbar
            }
            return xbar + correction / sampleSize
        }
        return Double.NaN
    }

    /**
     * Returns the weighted arithmetic mean of the entries in the specified portion of
     * the input array, or `Double.NaN` if the designated subarray
     * is empty.
     *
     *
     * Throws `IllegalArgumentException` if either array is null.
     *
     *
     * See [Mean] for details on the computing algorithm. The two-pass algorithm
     * described above is used here, with weights applied in computing both the original
     * estimate and the correction factor.
     *
     *
     * Throws `IllegalArgumentException` if any of the following are true:
     *  * the values array is null
     *  * the weights array is null
     *  * the weights array does not have the same length as the values array
     *  * the weights array contains one or more infinite values
     *  * the weights array contains one or more NaN values
     *  * the weights array contains negative values
     *  * the start and length arguments do not determine a valid array
     *
     *
     * @param values the input array
     * @param weights the weights array
     * @param begin index of the first array element to include
     * @param length the number of elements to include
     * @return the mean of the values or Double.NaN if length = 0
     * @throws MathIllegalArgumentException if the parameters are not valid
     * @since 2.1
     */
    /**
     * Returns the weighted arithmetic mean of the entries in the input array.
     *
     *
     * Throws `MathIllegalArgumentException` if either array is null.
     *
     *
     * See [Mean] for details on the computing algorithm. The two-pass algorithm
     * described above is used here, with weights applied in computing both the original
     * estimate and the correction factor.
     *
     *
     * Throws `MathIllegalArgumentException` if any of the following are true:
     *  * the values array is null
     *  * the weights array is null
     *  * the weights array does not have the same length as the values array
     *  * the weights array contains one or more infinite values
     *  * the weights array contains one or more NaN values
     *  * the weights array contains negative values
     *
     *
     * @param values the input array
     * @param weights the weights array
     * @return the mean of the values or Double.NaN if length = 0
     * @throws MathIllegalArgumentException if the parameters are not valid
     * @since 2.1
     */
    //@Throws(MathIllegalArgumentException::class)
     override fun evaluate(
        values: DoubleArray, weights: DoubleArray,
        begin: Int, length: Int
    ): Double {
        if (test(values, weights, begin, length)) {
            val sum = Sum()

            // Compute initial estimate using definitional formula
            val sumw: Double = sum.evaluate(weights, begin, length)
            val xbarw: Double = sum.evaluate(values, weights, begin, length) / sumw

            // Compute correction factor in second pass
            var correction = 0.0
            for (i in begin until begin + length) {
                correction += weights[i] * (values[i] - xbarw)
            }
            return xbarw + correction / sumw
        }
        return Double.NaN
    }

    /**
     * {@inheritDoc}
     */
    override fun copy(): Mean {
        val result = Mean()
        // No try-catch or advertised exception because args are guaranteed non-null
        copy(this, result)
        return result
    }

    companion object {
        /** Serializable version identifier  */
        private const val serialVersionUID = -1296043746617791564L

        /**
         * Copies source to dest.
         *
         * Neither source nor dest can be null.
         *
         * @param source Mean to copy
         * @param dest Mean to copy to
         * @throws NullArgumentException if either source or dest is null
         */
        @Throws(NullArgumentException::class)
        fun copy(source: Mean, dest: Mean) {
            MathUtils.checkNotNull(source)
            MathUtils.checkNotNull(dest)
            dest.dataRef = source.dataRef?.copyOf()
            dest.incMoment = source.incMoment
            dest.moment = source.moment!!.copy()
        }
    }

    override fun evaluate(values: DoubleArray, weights: DoubleArray): Double {
        return evaluate(values, weights, 0 , values.size)
    }

}
