/*
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.jetbrains.kotlinx.statistics.math3.stat.descriptive.moment

import org.jetbrains.kotlinx.statistics.math3.exception.NullArgumentException
import org.jetbrains.kotlinx.statistics.math3.util.MathUtils.checkNotNull

/**
 * Computes a statistic related to the Second Central Moment.  Specifically,
 * what is computed is the sum of squared deviations from the sample mean.
 *
 *
 * The following recursive updating formula is used:
 *
 *
 * Let
 *  *  dev = (current obs - previous mean)
 *  *  n = number of observations (including current obs)
 *
 * Then
 *
 *
 * new value = old value + dev^2 * (n -1) / n.
 *
 *
 * Returns `Double.NaN` if no data values have been added and
 * returns `0` if there is just one value in the data set.
 * Note that Double.NaN may also be returned if the input includes NaN
 * and / or infinite values.
 *
 *
 * **Note that this implementation is not synchronized.** If
 * multiple threads access an instance of this class concurrently, and at least
 * one of the threads invokes the `increment()` or
 * `clear()` method, it must be synchronized externally.
 *
 */
open class SecondMoment : FirstMoment {
    /**
     * {@inheritDoc}
     */
    /** second moment of values that have been added  */
     var m2: Double
        protected set

    override val result: Double
        get() = m2

    /**
     * Create a SecondMoment instance
     */
    constructor() : super() {
        m2 = Double.NaN
    }

    /**
     * Copy constructor, creates a new `SecondMoment` identical
     * to the `original`
     *
     * @param original the `SecondMoment` instance to copy
     * @throws NullArgumentException if original is null
     */
    constructor(original: SecondMoment) : super(original) {
        m2 = original.m2
    }

    /**
     * {@inheritDoc}
     */
    override fun increment(d: Double) {
        if (n < 1) {
            m2 = 0.0
            m1 = m2
        }
        super.increment(d)
        m2 += (n.toDouble() - 1) * dev * nDev
    }

    /**
     * {@inheritDoc}
     */
    override fun clear() {
        super.clear()
        m2 = Double.NaN
    }

    /**
     * {@inheritDoc}
     */
    override fun copy(): SecondMoment {
        val m2 = SecondMoment()
        // no try-catch or advertised NAE because args are guaranteed non-null
        copy(this, m2)
        return m2
    }

    companion object {
        /** Serializable version identifier  */
        private const val serialVersionUID = 3942403127395076445L

        /**
         * Copies source to dest.
         *
         * Neither source nor dest can be null.
         *
         * @param source SecondMoment to copy
         * @param dest SecondMoment to copy to
         * @throws NullArgumentException if either source or dest is null
         */
        @Throws(NullArgumentException::class)
        fun copy(source: SecondMoment, dest: SecondMoment) {
            checkNotNull(source)
            checkNotNull(dest)
            FirstMoment.copy(source, dest)
            dest.m2 = source.m2
        }
    }
}
