/*
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.jetbrains.kotlinx.statistics.math3.stat.descriptive.moment

import org.jetbrains.kotlinx.statistics.math3.exception.MathIllegalArgumentException
import org.jetbrains.kotlinx.statistics.math3.exception.NullArgumentException
import org.jetbrains.kotlinx.statistics.math3.stat.descriptive.AbstractStorelessUnivariateStatistic
import org.jetbrains.kotlinx.statistics.math3.util.MathUtils

/**
 * Computes the skewness of the available values.
 *
 *
 * We use the following (unbiased) formula to define skewness:
 *
 *
 * skewness = [n / (n -1) (n - 2)] sum[(x_i - mean)^3] / std^3
 *
 *
 * where n is the number of values, mean is the [Mean] and std is the
 * [StandardDeviation]
 *
 *
 * Note that this statistic is undefined for n < 3.  `Double.Nan`
 * is returned when there is not sufficient data to compute the statistic.
 * Double.NaN may also be returned if the input includes NaN and / or
 * infinite values.
 *
 *
 * **Note that this implementation is not synchronized.** If
 * multiple threads access an instance of this class concurrently, and at least
 * one of the threads invokes the `increment()` or
 * `clear()` method, it must be synchronized externally.
 *
 */
class Skewness : AbstractStorelessUnivariateStatistic {
    /** Third moment on which this statistic is based  */
    protected lateinit var moment: ThirdMoment

    /**
     * Determines whether or not this statistic can be incremented or cleared.
     *
     *
     * Statistics based on (constructed from) external moments cannot
     * be incremented or cleared.
     */
    protected var incMoment = false

    /**
     * Constructs a Skewness
     */
    constructor() {
        incMoment = true
        moment = ThirdMoment()
    }

    /**
     * Constructs a Skewness with an external moment
     * @param m3 external moment
     */
    constructor(m3: ThirdMoment) {
        incMoment = false
        moment = m3
    }

    /**
     * Copy constructor, creates a new `Skewness` identical
     * to the `original`
     *
     * @param original the `Skewness` instance to copy
     * @throws NullArgumentException if original is null
     */
    constructor(original: Skewness) {
        copy(original, this)
    }

    /**
     * {@inheritDoc}
     *
     * Note that when [.Skewness] is used to
     * create a Skewness, this method does nothing. In that case, the
     * ThirdMoment should be incremented directly.
     */
    override fun increment(d: Double) {
        if (incMoment) {
            moment.increment(d)
        }
    }

    override val result: Double
        /**
         * Returns the value of the statistic based on the values that have been added.
         *
         *
         * See [Skewness] for the definition used in the computation.
         *
         * @return the skewness of the available values.
         */
        get() {
            if (moment.n < 3) {
                return Double.NaN
            }
            val variance = moment.m2 / (moment!!.n - 1)
            return if (variance < 10E-20) {
                0.0
            } else {
                val n0 = moment!!.n.toDouble()
                n0 * moment!!.m3 /
                        ((n0 - 1) * (n0 - 2) * kotlin.math.sqrt(variance) * variance)
            }
        }
    override val n: Long
        /**
         * {@inheritDoc}
         */
        get() = moment!!.n

    /**
     * {@inheritDoc}
     */
    override fun clear() {
        if (incMoment) {
            moment!!.clear()
        }
    }

    /**
     * Returns the Skewness of the entries in the specifed portion of the
     * input array.
     *
     *
     * See [Skewness] for the definition used in the computation.
     *
     *
     * Throws `IllegalArgumentException` if the array is null.
     *
     * @param values the input array
     * @param begin the index of the first array element to include
     * @param length the number of elements to include
     * @return the skewness of the values or Double.NaN if length is less than
     * 3
     * @throws MathIllegalArgumentException if the array is null or the array index
     * parameters are not valid
     */
    //@Throws(MathIllegalArgumentException::class)
    override fun evaluate(
        values: DoubleArray, begin: Int,
        length: Int
    ): Double {

        // Initialize the skewness
        var skew = Double.NaN
        if (test(values, begin, length) && length > 2) {
            val mean = Mean()
            // Get the mean and the standard deviation
            val m = mean.evaluate(values, begin, length)

            // Calc the std, this is implemented here instead
            // of using the standardDeviation method eliminate
            // a duplicate pass to get the mean
            var accum = 0.0
            var accum2 = 0.0
            for (i in begin until begin + length) {
                val d = values[i] - m
                accum += d * d
                accum2 += d
            }
            val variance = (accum - accum2 * accum2 / length) / (length - 1)
            var accum3 = 0.0
            for (i in begin until begin + length) {
                val d = values[i] - m
                accum3 += d * d * d
            }
            accum3 /= variance * kotlin.math.sqrt(variance)

            // Get N
            val n0 = length.toDouble()

            // Calculate skewness
            skew = n0 / ((n0 - 1) * (n0 - 2)) * accum3
        }
        return skew
    }

    /**
     * {@inheritDoc}
     */
    override fun copy(): Skewness {
        val result = Skewness()
        // No try-catch or advertised exception because args are guaranteed non-null
        copy(this, result)
        return result
    }

    companion object {
        /** Serializable version identifier  */
        private const val serialVersionUID = 7101857578996691352L

        /**
         * Copies source to dest.
         *
         * Neither source nor dest can be null.
         *
         * @param source Skewness to copy
         * @param dest Skewness to copy to
         * @throws NullArgumentException if either source or dest is null
         */
        @Throws(NullArgumentException::class)
        fun copy(source: Skewness, dest: Skewness) {
            MathUtils.checkNotNull(source)
            MathUtils.checkNotNull(dest)
            dest.setData(source.dataRef)
            dest.moment = ThirdMoment(source.moment.copy())
            dest.incMoment = source.incMoment
        }
    }
}
