/*
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.jetbrains.kotlinx.statistics.math3.stat.descriptive.moment

import org.jetbrains.kotlinx.statistics.math3.exception.NullArgumentException
import org.jetbrains.kotlinx.statistics.math3.util.MathUtils.checkNotNull

/**
 * Computes a statistic related to the Third Central Moment.  Specifically,
 * what is computed is the sum of cubed deviations from the sample mean.
 *
 *
 * The following recursive updating formula is used:
 *
 *
 * Let
 *  *  dev = (current obs - previous mean)
 *  *  m2 = previous value of [SecondMoment]
 *  *  n = number of observations (including current obs)
 *
 * Then
 *
 *
 * new value = old value - 3 * (dev/n) * m2 + (n-1) * (n -2) * (dev^3/n^2)
 *
 *
 * Returns `Double.NaN` if no data values have been added and
 * returns `0` if there is just one value in the data set.
 * Note that Double.NaN may also be returned if the input includes NaN
 * and / or infinite values.
 *
 *
 * **Note that this implementation is not synchronized.** If
 * multiple threads access an instance of this class concurrently, and at least
 * one of the threads invokes the `increment()` or
 * `clear()` method, it must be synchronized externally.
 *
 */
open class ThirdMoment : SecondMoment {
    /**
     * {@inheritDoc}
     */
    /** third moment of values that have been added  */
    var m3: Double = Double.NaN
        protected set

    override val result: Double
        get() = m3

    /**
     * Square of deviation of most recently added value from previous first
     * moment, normalized by previous sample size.  Retained to prevent
     * repeated computation in higher order moments.  nDevSq = nDev * nDev.
     */
    protected var nDevSq = 0.0

    /**
     * Create a FourthMoment instance
     */
    constructor() : super() {
        m3 = Double.NaN
        nDevSq = Double.NaN
    }

    /**
     * Copy constructor, creates a new `ThirdMoment` identical
     * to the `original`
     *
     * @param original the `ThirdMoment` instance to copy
     * @throws NullArgumentException if orginal is null
     */
    constructor(original: ThirdMoment) {
        copy(original, this)
    }

    /**
     * {@inheritDoc}
     */
    override fun increment(d: Double) {
        if (n < 1) {
            m1 = 0.0
            m2 = m1
            m3 = m2
        }
        val prevM2: Double = m2
        super.increment(d)
        nDevSq = nDev * nDev
        val n0 = n.toDouble()
        m3 = m3 - 3.0 * nDev * prevM2 + (n0 - 1) * (n0 - 2) * nDevSq * dev
    }

    /**
     * {@inheritDoc}
     */
    override fun clear() {
        super.clear()
        m3 = Double.NaN
        nDevSq = Double.NaN
    }

    /**
     * {@inheritDoc}
     */
    override fun copy(): ThirdMoment {
        val m3 = ThirdMoment()
        // No try-catch or advertised exception because args are guaranteed non-null
        copy(this, m3)
        return m3
    }

    companion object {
        /** Serializable version identifier  */
        private const val serialVersionUID = -7818711964045118679L

        /**
         * Copies source to dest.
         *
         * Neither source nor dest can be null.
         *
         * @param source ThirdMoment to copy
         * @param dest ThirdMoment to copy to
         * @throws NullArgumentException if either source or dest is null
         */
        @Throws(NullArgumentException::class)
        fun copy(source: ThirdMoment, dest: ThirdMoment) {
            checkNotNull(source)
            checkNotNull(dest)
            SecondMoment.copy(source, dest)
            dest.m3 = source.m3
            dest.nDevSq = source.nDevSq
        }
    }
}
