/*
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.jetbrains.kotlinx.statistics.math3.stat.descriptive.moment

import org.jetbrains.kotlinx.statistics.math3.exception.MathIllegalArgumentException
import org.jetbrains.kotlinx.statistics.math3.exception.NullArgumentException
import org.jetbrains.kotlinx.statistics.math3.stat.descriptive.AbstractStorelessUnivariateStatistic
import org.jetbrains.kotlinx.statistics.math3.stat.descriptive.WeightedEvaluation
import org.jetbrains.kotlinx.statistics.math3.util.MathUtils
import kotlin.jvm.JvmOverloads

/**
 * Computes the variance of the available values.  By default, the unbiased
 * "sample variance" definitional formula is used:
 *
 *
 * variance = sum((x_i - mean)^2) / (n - 1)
 *
 *
 * where mean is the [Mean] and `n` is the number
 * of sample observations.
 *
 *
 * The definitional formula does not have good numerical properties, so
 * this implementation does not compute the statistic using the definitional
 * formula.
 *  *  The `getResult` method computes the variance using
 * updating formulas based on West's algorithm, as described in
 * [ Chan, T. F. and
 * J. G. Lewis 1979, *Communications of the ACM*,
 * vol. 22 no. 9, pp. 526-531.](http://doi.acm.org/10.1145/359146.359152)
 *  *  The `evaluate` methods leverage the fact that they have the
 * full array of values in memory to execute a two-pass algorithm.
 * Specifically, these methods use the "corrected two-pass algorithm" from
 * Chan, Golub, Levesque, *Algorithms for Computing the Sample Variance*,
 * American Statistician, vol. 37, no. 3 (1983) pp. 242-247.
 * Note that adding values using `increment` or
 * `incrementAll` and then executing `getResult` will
 * sometimes give a different, less accurate, result than executing
 * `evaluate` with the full array of values. The former approach
 * should only be used when the full array of values is not available.
 *
 *
 * The "population variance"  ( sum((x_i - mean)^2) / n ) can also
 * be computed using this statistic.  The `isBiasCorrected`
 * property determines whether the "population" or "sample" value is
 * returned by the `evaluate` and `getResult` methods.
 * To compute population variances, set this property to `false.`
 *
 *
 *
 * **Note that this implementation is not synchronized.** If
 * multiple threads access an instance of this class concurrently, and at least
 * one of the threads invokes the `increment()` or
 * `clear()` method, it must be synchronized externally.
 *
 */
class Variance : AbstractStorelessUnivariateStatistic, WeightedEvaluation {
    /** SecondMoment is used in incremental calculation of Variance */
    lateinit var moment: SecondMoment

    /**
     * Whether or not [.increment] should increment
     * the internal second moment. When a Variance is constructed with an
     * external SecondMoment as a constructor parameter, this property is
     * set to false and increments must be applied to the second moment
     * directly.
     */
    protected var incMoment = true

    /**
     * Whether or not bias correction is applied when computing the
     * value of the statistic. True means that bias is corrected.  See
     * [Variance] for details on the formula.
     */
    private var isBiasCorrected = true

    /**
     * Constructs a Variance with default (true) `isBiasCorrected`
     * property.
     */
    constructor() {
        moment = SecondMoment()
    }

    /**
     * Constructs a Variance based on an external second moment.
     * When this constructor is used, the statistic may only be
     * incremented via the moment, i.e., [.increment]
     * does nothing; whereas `m2.increment(value)` increments
     * both `m2` and the Variance instance constructed from it.
     *
     * @param m2 the SecondMoment (Third or Fourth moments work
     * here as well.)
     */
    constructor(m2: SecondMoment) {
        incMoment = false
        moment = m2
    }

    /**
     * Constructs a Variance with the specified `isBiasCorrected`
     * property
     *
     * @param isBiasCorrected  setting for bias correction - true means
     * bias will be corrected and is equivalent to using the argumentless
     * constructor
     */
    constructor(isBiasCorrected: Boolean) {
        moment = SecondMoment()
        this.isBiasCorrected = isBiasCorrected
    }

    /**
     * Constructs a Variance with the specified `isBiasCorrected`
     * property and the supplied external second moment.
     *
     * @param isBiasCorrected  setting for bias correction - true means
     * bias will be corrected
     * @param m2 the SecondMoment (Third or Fourth moments work
     * here as well.)
     */
    constructor(isBiasCorrected: Boolean, m2: SecondMoment) {
        incMoment = false
        moment = m2
        this.isBiasCorrected = isBiasCorrected
    }

    /**
     * Copy constructor, creates a new `Variance` identical
     * to the `original`
     *
     * @param original the `Variance` instance to copy
     * @throws NullArgumentException if original is null
     */
    constructor(original: Variance) {
        copy(original, this)
    }

    /**
     * {@inheritDoc}
     *
     * If all values are available, it is more accurate to use
     * [.evaluate] rather than adding values one at a time
     * using this method and then executing [.getResult], since
     * `evaluate` leverages the fact that is has the full
     * list of values together to execute a two-pass algorithm.
     * See [Variance].
     *
     *
     * Note also that when [.Variance] is used to
     * create a Variance, this method does nothing. In that case, the
     * SecondMoment should be incremented directly.
     */
    override fun increment(d: Double) {
        if (incMoment) {
            moment.increment(d)
        }
    }

    override val result: Double
        /**
         * {@inheritDoc}
         */
        get() = if (moment.n == 0L) {
            Double.NaN
        } else if (moment.n == 1L) {
            0.0
        } else {
            if (isBiasCorrected) {
                moment.m2 / (moment.n - 1.0)
            } else {
                moment.m2 / moment.n
            }
        }
    override val n: Long
        /**
         * {@inheritDoc}
         */
        get() = moment.n

    /**
     * {@inheritDoc}
     */
    override fun clear() {
        if (incMoment) {
            moment.clear()
        }
    }

    /**
     * Returns the variance of the entries in the input array, or
     * `Double.NaN` if the array is empty.
     *
     *
     * See [Variance] for details on the computing algorithm.
     *
     *
     * Returns 0 for a single-value (i.e. length = 1) sample.
     *
     *
     * Throws `MathIllegalArgumentException` if the array is null.
     *
     *
     * Does not change the internal state of the statistic.
     *
     * @param values the input array
     * @return the variance of the values or Double.NaN if length = 0
     * @throws MathIllegalArgumentException if the array is null
     */
    //@Throws(MathIllegalArgumentException::class)
    override fun evaluate(values: DoubleArray): Double {
        if (values == null) {
            TODO() //throw NullArgumentException(LocalizedFormats.INPUT_ARRAY)
        }
        return evaluate(values, 0, values.size)
    }

    /**
     * Returns the variance of the entries in the specified portion of
     * the input array, or `Double.NaN` if the designated subarray
     * is empty.  Note that Double.NaN may also be returned if the input
     * includes NaN and / or infinite values.
     *
     *
     * See [Variance] for details on the computing algorithm.
     *
     *
     * Returns 0 for a single-value (i.e. length = 1) sample.
     *
     *
     * Does not change the internal state of the statistic.
     *
     *
     * Throws `MathIllegalArgumentException` if the array is null.
     *
     * @param values the input array
     * @param begin index of the first array element to include
     * @param length the number of elements to include
     * @return the variance of the values or Double.NaN if length = 0
     * @throws MathIllegalArgumentException if the array is null or the array index
     * parameters are not valid
     */
    //@Throws(MathIllegalArgumentException::class)
    override fun evaluate(values: DoubleArray, begin: Int, length: Int): Double {
        var `var` = Double.NaN
        if (test(values, begin, length)) {
            clear()
            if (length == 1) {
                `var` = 0.0
            } else if (length > 1) {
                val mean = Mean()
                val m = mean.evaluate(values, begin, length)
                `var` = evaluate(values, m, begin, length)
            }
        }
        return `var`
    }

    /**
     *
     * Returns the weighted variance of the entries in the specified portion of
     * the input array, or `Double.NaN` if the designated subarray
     * is empty.
     *
     *
     * Uses the formula <pre>
     * (weights[i]*(values[i] - weightedMean)<sup>2</sup>)/((weights[i]) - 1)
    </pre> *
     * where weightedMean is the weighted mean
     *
     *
     * This formula will not return the same result as the unweighted variance when all
     * weights are equal, unless all weights are equal to 1. The formula assumes that
     * weights are to be treated as "expansion values," as will be the case if for example
     * the weights represent frequency counts. To normalize weights so that the denominator
     * in the variance computation equals the length of the input vector minus one, use <pre>
     * `evaluate(values, MathArrays.normalizeArray(weights, values.length)); `
    </pre> *
     *
     *
     * Returns 0 for a single-value (i.e. length = 1) sample.
     *
     *
     * Throws `IllegalArgumentException` if any of the following are true:
     *  * the values array is null
     *  * the weights array is null
     *  * the weights array does not have the same length as the values array
     *  * the weights array contains one or more infinite values
     *  * the weights array contains one or more NaN values
     *  * the weights array contains negative values
     *  * the start and length arguments do not determine a valid array
     *
     *
     *
     * Does not change the internal state of the statistic.
     *
     *
     * Throws `MathIllegalArgumentException` if either array is null.
     *
     * @param values the input array
     * @param weights the weights array
     * @param begin index of the first array element to include
     * @param length the number of elements to include
     * @return the weighted variance of the values or Double.NaN if length = 0
     * @throws MathIllegalArgumentException if the parameters are not valid
     * @since 2.1
     */
    //@Throws(MathIllegalArgumentException::class)
    override fun evaluate(
        values: DoubleArray, weights: DoubleArray,
        begin: Int, length: Int
    ): Double {
        var `var` = Double.NaN
        if (test(values, weights, begin, length)) {
            clear()
            if (length == 1) {
                `var` = 0.0
            } else if (length > 1) {
                val mean = Mean()
                val m = mean.evaluate(values, weights, begin, length)
                `var` = evaluate(values, weights, m, begin, length)
            }
        }
        return `var`
    }

    /**
     *
     *
     * Returns the weighted variance of the entries in the the input array.
     *
     *
     * Uses the formula <pre>
     * (weights[i]*(values[i] - weightedMean)<sup>2</sup>)/((weights[i]) - 1)
    </pre> *
     * where weightedMean is the weighted mean
     *
     *
     * This formula will not return the same result as the unweighted variance when all
     * weights are equal, unless all weights are equal to 1. The formula assumes that
     * weights are to be treated as "expansion values," as will be the case if for example
     * the weights represent frequency counts. To normalize weights so that the denominator
     * in the variance computation equals the length of the input vector minus one, use <pre>
     * `evaluate(values, MathArrays.normalizeArray(weights, values.length)); `
    </pre> *
     *
     *
     * Returns 0 for a single-value (i.e. length = 1) sample.
     *
     *
     * Throws `MathIllegalArgumentException` if any of the following are true:
     *  * the values array is null
     *  * the weights array is null
     *  * the weights array does not have the same length as the values array
     *  * the weights array contains one or more infinite values
     *  * the weights array contains one or more NaN values
     *  * the weights array contains negative values
     *
     *
     *
     * Does not change the internal state of the statistic.
     *
     *
     * Throws `MathIllegalArgumentException` if either array is null.
     *
     * @param values the input array
     * @param weights the weights array
     * @return the weighted variance of the values
     * @throws MathIllegalArgumentException if the parameters are not valid
     * @since 2.1
     */
    //@Throws(MathIllegalArgumentException::class)
    override fun evaluate(values: DoubleArray, weights: DoubleArray): Double {
        return evaluate(values, weights, 0, values.size)
    }
    /**
     * Returns the variance of the entries in the specified portion of
     * the input array, using the precomputed mean value.  Returns
     * `Double.NaN` if the designated subarray is empty.
     *
     *
     * See [Variance] for details on the computing algorithm.
     *
     *
     * The formula used assumes that the supplied mean value is the arithmetic
     * mean of the sample data, not a known population parameter.  This method
     * is supplied only to save computation when the mean has already been
     * computed.
     *
     *
     * Returns 0 for a single-value (i.e. length = 1) sample.
     *
     *
     * Throws `MathIllegalArgumentException` if the array is null.
     *
     *
     * Does not change the internal state of the statistic.
     *
     * @param values the input array
     * @param mean the precomputed mean value
     * @param begin index of the first array element to include
     * @param length the number of elements to include
     * @return the variance of the values or Double.NaN if length = 0
     * @throws MathIllegalArgumentException if the array is null or the array index
     * parameters are not valid
     */
    /**
     * Returns the variance of the entries in the input array, using the
     * precomputed mean value.  Returns `Double.NaN` if the array
     * is empty.
     *
     *
     * See [Variance] for details on the computing algorithm.
     *
     *
     * If `isBiasCorrected` is `true` the formula used
     * assumes that the supplied mean value is the arithmetic mean of the
     * sample data, not a known population parameter.  If the mean is a known
     * population parameter, or if the "population" version of the variance is
     * desired, set `isBiasCorrected` to `false` before
     * invoking this method.
     *
     *
     * Returns 0 for a single-value (i.e. length = 1) sample.
     *
     *
     * Throws `MathIllegalArgumentException` if the array is null.
     *
     *
     * Does not change the internal state of the statistic.
     *
     * @param values the input array
     * @param mean the precomputed mean value
     * @return the variance of the values or Double.NaN if the array is empty
     * @throws MathIllegalArgumentException if the array is null
     */
    @JvmOverloads
    @Throws(MathIllegalArgumentException::class)
    fun evaluate(
        values: DoubleArray, mean: Double,
        begin: Int = 0, length: Int = values.size
    ): Double {
        var `var` = Double.NaN
        if (test(values, begin, length)) {
            if (length == 1) {
                `var` = 0.0
            } else if (length > 1) {
                var accum = 0.0
                var dev = 0.0
                var accum2 = 0.0
                for (i in begin until begin + length) {
                    dev = values[i] - mean
                    accum += dev * dev
                    accum2 += dev
                }
                val len = length.toDouble()
                `var` = if (isBiasCorrected) {
                    (accum - accum2 * accum2 / len) / (len - 1.0)
                } else {
                    (accum - accum2 * accum2 / len) / len
                }
            }
        }
        return `var`
    }
    /**
     * Returns the weighted variance of the entries in the specified portion of
     * the input array, using the precomputed weighted mean value.  Returns
     * `Double.NaN` if the designated subarray is empty.
     *
     *
     * Uses the formula <pre>
     * (weights[i]*(values[i] - mean)<sup>2</sup>)/((weights[i]) - 1)
    </pre> *
     *
     *
     * The formula used assumes that the supplied mean value is the weighted arithmetic
     * mean of the sample data, not a known population parameter. This method
     * is supplied only to save computation when the mean has already been
     * computed.
     *
     *
     * This formula will not return the same result as the unweighted variance when all
     * weights are equal, unless all weights are equal to 1. The formula assumes that
     * weights are to be treated as "expansion values," as will be the case if for example
     * the weights represent frequency counts. To normalize weights so that the denominator
     * in the variance computation equals the length of the input vector minus one, use <pre>
     * `evaluate(values, MathArrays.normalizeArray(weights, values.length), mean); `
    </pre> *
     *
     *
     * Returns 0 for a single-value (i.e. length = 1) sample.
     *
     *
     * Throws `MathIllegalArgumentException` if any of the following are true:
     *  * the values array is null
     *  * the weights array is null
     *  * the weights array does not have the same length as the values array
     *  * the weights array contains one or more infinite values
     *  * the weights array contains one or more NaN values
     *  * the weights array contains negative values
     *  * the start and length arguments do not determine a valid array
     *
     *
     *
     * Does not change the internal state of the statistic.
     *
     * @param values the input array
     * @param weights the weights array
     * @param mean the precomputed weighted mean value
     * @param begin index of the first array element to include
     * @param length the number of elements to include
     * @return the variance of the values or Double.NaN if length = 0
     * @throws MathIllegalArgumentException if the parameters are not valid
     * @since 2.1
     */
    /**
     *
     * Returns the weighted variance of the values in the input array, using
     * the precomputed weighted mean value.
     *
     *
     * Uses the formula <pre>
     * (weights[i]*(values[i] - mean)<sup>2</sup>)/((weights[i]) - 1)
    </pre> *
     *
     *
     * The formula used assumes that the supplied mean value is the weighted arithmetic
     * mean of the sample data, not a known population parameter. This method
     * is supplied only to save computation when the mean has already been
     * computed.
     *
     *
     * This formula will not return the same result as the unweighted variance when all
     * weights are equal, unless all weights are equal to 1. The formula assumes that
     * weights are to be treated as "expansion values," as will be the case if for example
     * the weights represent frequency counts. To normalize weights so that the denominator
     * in the variance computation equals the length of the input vector minus one, use <pre>
     * `evaluate(values, MathArrays.normalizeArray(weights, values.length), mean); `
    </pre> *
     *
     *
     * Returns 0 for a single-value (i.e. length = 1) sample.
     *
     *
     * Throws `MathIllegalArgumentException` if any of the following are true:
     *  * the values array is null
     *  * the weights array is null
     *  * the weights array does not have the same length as the values array
     *  * the weights array contains one or more infinite values
     *  * the weights array contains one or more NaN values
     *  * the weights array contains negative values
     *
     *
     *
     * Does not change the internal state of the statistic.
     *
     * @param values the input array
     * @param weights the weights array
     * @param mean the precomputed weighted mean value
     * @return the variance of the values or Double.NaN if length = 0
     * @throws MathIllegalArgumentException if the parameters are not valid
     * @since 2.1
     */
    @JvmOverloads
    @Throws(MathIllegalArgumentException::class)
    fun evaluate(
        values: DoubleArray, weights: DoubleArray,
        mean: Double, begin: Int = 0, length: Int = values.size
    ): Double {
        var `var` = Double.NaN
        if (test(values, weights, begin, length)) {
            if (length == 1) {
                `var` = 0.0
            } else if (length > 1) {
                var accum = 0.0
                var dev = 0.0
                var accum2 = 0.0
                for (i in begin until begin + length) {
                    dev = values[i] - mean
                    accum += weights[i] * (dev * dev)
                    accum2 += weights[i] * dev
                }
                var sumWts = 0.0
                for (i in begin until begin + length) {
                    sumWts += weights[i]
                }
                `var` = if (isBiasCorrected) {
                    (accum - accum2 * accum2 / sumWts) / (sumWts - 1.0)
                } else {
                    (accum - accum2 * accum2 / sumWts) / sumWts
                }
            }
        }
        return `var`
    }

    /**
     * @return Returns the isBiasCorrected.
     */
    fun isBiasCorrected(): Boolean {
        return isBiasCorrected
    }

    /**
     * @param biasCorrected The isBiasCorrected to set.
     */
    fun setBiasCorrected(biasCorrected: Boolean) {
        isBiasCorrected = biasCorrected
    }

    /**
     * {@inheritDoc}
     */
    override fun copy(): Variance {
        val result = Variance()
        // No try-catch or advertised exception because parameters are guaranteed non-null
        copy(this, result)
        return result
    }

    companion object {
        /** Serializable version identifier  */
        private const val serialVersionUID = -9111962718267217978L

        /**
         * Copies source to dest.
         *
         * Neither source nor dest can be null.
         *
         * @param source Variance to copy
         * @param dest Variance to copy to
         * @throws NullArgumentException if either source or dest is null
         */
        @Throws(NullArgumentException::class)
        fun copy(source: Variance, dest: Variance) {
            MathUtils.checkNotNull(source)
            MathUtils.checkNotNull(dest)
            dest.setData(source.dataRef)
            dest.moment = source.moment.copy()
            dest.isBiasCorrected = source.isBiasCorrected
            dest.incMoment = source.incMoment
        }
    }
}
