/*
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.jetbrains.kotlinx.statistics.math3.stat.descriptive.rank

import org.jetbrains.kotlinx.statistics.math3.stat.descriptive.AbstractStorelessUnivariateStatistic
import org.jetbrains.kotlinx.statistics.math3.util.MathUtils

/**
 * Returns the minimum of the available values.
 *
 *
 *
 *  * The result is `NaN` iff all values are `NaN`
 * (i.e. `NaN` values have no impact on the value of the statistic).
 *  * If any of the values equals `Double.NEGATIVE_INFINITY`,
 * the result is `Double.NEGATIVE_INFINITY.`
 *
 *
 *
 * **Note that this implementation is not synchronized.** If
 * multiple threads access an instance of this class concurrently, and at least
 * one of the threads invokes the `increment()` or
 * `clear()` method, it must be synchronized externally.
 *
 */
class Min : AbstractStorelessUnivariateStatistic {
    /**
     * {@inheritDoc}
     */
    /**Number of values that have been added  */
    override var n: Long = 0
        private set
    /**
     * {@inheritDoc}
     */
    /**Current value of the statistic  */
    override var result = 0.0
        private set

    /**
     * Create a Min instance
     */
    constructor() {
        n = 0
        result = Double.NaN
    }

    /**
     * Copy constructor, creates a new `Min` identical
     * to the `original`
     *
     * @param original the `Min` instance to copy
     * @throws NullArgumentException if original is null
     */
    constructor(original: Min) {
        copy(original, this)
    }

    /**
     * {@inheritDoc}
     */
    override fun increment(d: Double) {
        if (d < result || result.isNaN()) {
            result = d
        }
        n++
    }

    /**
     * {@inheritDoc}
     */
    override fun clear() {
        result = Double.NaN
        n = 0
    }

    /**
     * Returns the minimum of the entries in the specified portion of
     * the input array, or `Double.NaN` if the designated subarray
     * is empty.
     *
     *
     * Throws `MathIllegalArgumentException` if the array is null or
     * the array index parameters are not valid.
     *
     *
     *
     *  * The result is `NaN` iff all values are `NaN`
     * (i.e. `NaN` values have no impact on the value of the statistic).
     *  * If any of the values equals `Double.NEGATIVE_INFINITY`,
     * the result is `Double.NEGATIVE_INFINITY.`
     *
     *
     * @param values the input array
     * @param begin index of the first array element to include
     * @param length the number of elements to include
     * @return the minimum of the values or Double.NaN if length = 0
     * @throws MathIllegalArgumentException if the array is null or the array index
     * parameters are not valid
     */
    //@Throws(MathIllegalArgumentException::class)
    override fun evaluate(values: DoubleArray, begin: Int, length: Int): Double {
        var min = Double.NaN
        if (test(values, begin, length)) {
            min = values[begin]
            for (i in begin until begin + length) {
                if (!values[i].isNaN()) {
                    min = if (min < values[i]) min else values[i]
                }
            }
        }
        return min
    }

    /**
     * {@inheritDoc}
     */
    override fun copy(): Min {
        val result = Min()
        // No try-catch or advertised exception - args are non-null
        copy(this, result)
        return result
    }

    companion object {
        /** Serializable version identifier  */
        private const val serialVersionUID = -2941995784909003131L

        /**
         * Copies source to dest.
         *
         * Neither source nor dest can be null.
         *
         * @param source Min to copy
         * @param dest Min to copy to
         * @throws NullArgumentException if either source or dest is null
         */
        //@Throws(NullArgumentException::class)
        fun copy(source: Min, dest: Min) {
            MathUtils.checkNotNull(source)
            MathUtils.checkNotNull(dest)
            dest.setData(source.dataRef)
            dest.n = source.n
            dest.result = source.result
        }
    }
}
