/*
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.jetbrains.kotlinx.statistics.math3.stat.descriptive.rank

import org.jetbrains.kotlinx.statistics.math3.BitSet
import org.jetbrains.kotlinx.statistics.math3.util.*
import org.jetbrains.kotlinx.statistics.math3.stat.descriptive.AbstractUnivariateStatistic
import org.jetbrains.kotlinx.statistics.math3.stat.ranking.NaNStrategy
import kotlin.jvm.JvmOverloads
import kotlin.math.ceil
import kotlin.math.floor

/**
 * Provides percentile computation.
 *
 *
 * There are several commonly used methods for estimating percentiles (a.k.a.
 * quantiles) based on sample data.  For large samples, the different methods
 * agree closely, but when sample sizes are small, different methods will give
 * significantly different results.  The algorithm implemented here works as follows:
 *
 *  1. Let `n` be the length of the (sorted) array and
 * `0 < p <= 100` be the desired percentile.
 *  1. If ` n = 1 ` return the unique array element (regardless of
 * the value of `p`); otherwise
 *  1. Compute the estimated percentile position
 * ` pos = p * (n + 1) / 100` and the difference, `d`
 * between `pos` and `floor(pos)` (i.e. the fractional
 * part of `pos`).
 *  1.  If `pos < 1` return the smallest element in the array.
 *  1.  Else if `pos >= n` return the largest element in the array.
 *  1.  Else let `lower` be the element in position
 * `floor(pos)` in the array and let `upper` be the
 * next element in the array.  Return `lower + d * (upper - lower)`
 *
 *
 *
 *
 * To compute percentiles, the data must be at least partially ordered.  Input
 * arrays are copied and recursively partitioned using an ordering definition.
 * The ordering used by `Arrays.sort(double[])` is the one determined
 * by [Double.compareTo].  This ordering makes
 * `Double.NaN` larger than any other value (including
 * `Double.POSITIVE_INFINITY`).  Therefore, for example, the median
 * (50th percentile) of
 * `{0, 1, 2, 3, 4, Double.NaN}` evaluates to `2.5.`
 *
 *
 * Since percentile estimation usually involves interpolation between array
 * elements, arrays containing  `NaN` or infinite values will often
 * result in `NaN` or infinite values returned.
 *
 *
 * Further, to include different estimation types such as R1, R2 as mentioned in
 * [Quantile page(wikipedia)](http://en.wikipedia.org/wiki/Quantile),
 * a type specific NaN handling strategy is used to closely match with the
 * typically observed results from popular tools like R(R1-R9), Excel(R7).
 *
 *
 * Since 2.2, Percentile uses only selection instead of complete sorting
 * and caches selection algorithm state between calls to the various
 * `evaluate` methods. This greatly improves efficiency, both for a single
 * percentile and multiple percentile computations. To maximize performance when
 * multiple percentiles are computed based on the same data, users should set the
 * data array once using either one of the [.evaluate] or
 * [.setData] methods and thereafter [.evaluate]
 * with just the percentile provided.
 *
 *
 *
 * **Note that this implementation is not synchronized.** If
 * multiple threads access an instance of this class concurrently, and at least
 * one of the threads invokes the `increment()` or
 * `clear()` method, it must be synchronized externally.
 *
 */
class Percentile : AbstractUnivariateStatistic {
    /** Default KthSelector used with default pivoting strategy  */
    private val kthSelector: KthSelector
    /**
     * Get the estimation [type][EstimationType] used for computation.
     *
     * @return the `estimationType` set
     */
    /** Any of the [EstimationType]s such as [CM][EstimationType.LEGACY] can be used.  */
    val estimationType: EstimationType

    /** NaN Handling of the input as defined by [NaNStrategy]  */
    private val nanStrategy: NaNStrategy

    /** Determines what percentile is computed when evaluate() is activated
     * with no quantile argument  */
    private var quantile = 0.0

    /** Cached pivots.  */
    private var cachedPivots: IntArray? = null
    /**
     * Constructs a Percentile with the specific quantile value and the following
     *
     *  * default method type: [EstimationType.LEGACY]
     *  * default NaN strategy: [NaNStrategy.REMOVED]
     *  * a Kth Selector : [KthSelector]
     *
     * @param quantile the quantile
     * @throws MathIllegalArgumentException  if p is not greater than 0 and less
     * than or equal to 100
     */
    /**
     * Constructs a Percentile with the following defaults.
     *
     *  * default quantile: 50.0, can be reset with [.setQuantile]
     *  * default estimation type: [EstimationType.LEGACY],
     * can be reset with [.withEstimationType]
     *  * default NaN strategy: [NaNStrategy.REMOVED],
     * can be reset with [.withNaNStrategy]
     *  * a KthSelector that makes use of [MedianOf3PivotingStrategy],
     * can be reset with [.withKthSelector]
     *
     */
    @JvmOverloads
    constructor(quantile: Double = 50.0) : this(
        quantile, EstimationType.LEGACY, NaNStrategy.REMOVED,
        KthSelector(MedianOf3PivotingStrategy())
    )

    /**
     * Copy constructor, creates a new `Percentile` identical
     * to the `original`
     *
     * @param original the `Percentile` instance to copy
     * @throws NullArgumentException if original is null
     */
    constructor(original: Percentile) {
        MathUtils.checkNotNull(original)
        estimationType = original.estimationType
        nanStrategy = original.naNStrategy
        kthSelector = original.getKthSelector()
        setData(original.dataRef)
        val originalPivots = original.cachedPivots
        if (originalPivots != null) {
            cachedPivots = IntArray(originalPivots.size)
            originalPivots.copyInto(cachedPivots!!, 0, 0, originalPivots.size)
            //java.lang.System.arraycopy(original.cachedPivots, 0, cachedPivots, 0, original.cachedPivots!!.size)
        }
        setQuantile(original.quantile)
    }

    /**
     * Constructs a Percentile with the specific quantile value,
     * [EstimationType], [NaNStrategy] and [KthSelector].
     *
     * @param quantile the quantile to be computed
     * @param estimationType one of the percentile [estimation types][EstimationType]
     * @param nanStrategy one of [NaNStrategy] to handle with NaNs
     * @param kthSelector a [KthSelector] to use for pivoting during search
     * @throws MathIllegalArgumentException if p is not within (0,100]
     * @throws NullArgumentException if type or NaNStrategy passed is null
     */
    protected constructor(
        quantile: Double,
        estimationType: EstimationType,
        nanStrategy: NaNStrategy,
        kthSelector: KthSelector
    ) {
        setQuantile(quantile)
        cachedPivots = null
        MathUtils.checkNotNull(estimationType)
        MathUtils.checkNotNull(nanStrategy)
        MathUtils.checkNotNull(kthSelector)
        this.estimationType = estimationType
        this.nanStrategy = nanStrategy
        this.kthSelector = kthSelector
    }

    /** {@inheritDoc}  */
    override fun setData(values: DoubleArray?) {
        if (values == null) {
            cachedPivots = null
        } else {
            cachedPivots = IntArray(PIVOTS_HEAP_LENGTH)
            cachedPivots!!.fill(-1)
            //java.util.Arrays.fill(cachedPivots, -1)
        }
        super.setData(values)
    }

    /** {@inheritDoc}  */
    //@Throws(MathIllegalArgumentException::class)
    override fun setData(values: DoubleArray?, begin: Int, length: Int) {
        if (values == null) {
            cachedPivots = null
        } else {
            cachedPivots = IntArray(PIVOTS_HEAP_LENGTH) { -1 }
            //java.util.Arrays.fill(cachedPivots, -1)
        }
        super.setData(values, begin, length)
    }

    /**
     * Returns the result of evaluating the statistic over the stored data.
     *
     *
     * The stored array is the one which was set by previous calls to
     * [.setData]
     *
     * @param p the percentile value to compute
     * @return the value of the statistic applied to the stored data
     * @throws MathIllegalArgumentException if p is not a valid quantile value
     * (p must be greater than 0 and less than or equal to 100)
     */
    //@Throws(MathIllegalArgumentException::class)
    fun evaluate(p: Double): Double {
        return evaluate(dataRef!!, p)
    }

    /**
     * Returns an estimate of the `p`th percentile of the values
     * in the `values` array.
     *
     *
     * Calls to this method do not modify the internal `quantile`
     * state of this statistic.
     *
     *
     *
     *  * Returns `Double.NaN` if `values` has length
     * `0`
     *  * Returns (for any value of `p`) `values[0]`
     * if `values` has length `1`
     *  * Throws `MathIllegalArgumentException` if `values`
     * is null or p is not a valid quantile value (p must be greater than 0
     * and less than or equal to 100)
     *
     *
     *
     * See [Percentile] for a description of the percentile estimation
     * algorithm used.
     *
     * @param values input array of values
     * @param p the percentile value to compute
     * @return the percentile value or Double.NaN if the array is empty
     * @throws MathIllegalArgumentException if `values` is null
     * or p is invalid
     */
    //@Throws(MathIllegalArgumentException::class)
    fun evaluate(values: DoubleArray, p: Double): Double {
        test(values, 0, 0)
        return evaluate(values, 0, values.size, p)
    }

    /**
     * Returns an estimate of the `quantile`th percentile of the
     * designated values in the `values` array.  The quantile
     * estimated is determined by the `quantile` property.
     *
     *
     *
     *  * Returns `Double.NaN` if `length = 0`
     *  * Returns (for any value of `quantile`)
     * `values[begin]` if `length = 1 `
     *  * Throws `MathIllegalArgumentException` if `values`
     * is null, or `start` or `length` is invalid
     *
     *
     *
     * See [Percentile] for a description of the percentile estimation
     * algorithm used.
     *
     * @param values the input array
     * @param start index of the first array element to include
     * @param length the number of elements to include
     * @return the percentile value
     * @throws MathIllegalArgumentException if the parameters are not valid
     */
    //@Throws(MathIllegalArgumentException::class)
    override fun evaluate(values: DoubleArray, start: Int, length: Int): Double {
        return evaluate(values, start, length, quantile)
    }

    /**
     * Returns an estimate of the `p`th percentile of the values
     * in the `values` array, starting with the element in (0-based)
     * position `begin` in the array and including `length`
     * values.
     *
     *
     * Calls to this method do not modify the internal `quantile`
     * state of this statistic.
     *
     *
     *
     *  * Returns `Double.NaN` if `length = 0`
     *  * Returns (for any value of `p`) `values[begin]`
     * if `length = 1 `
     *  * Throws `MathIllegalArgumentException` if `values`
     * is null , `begin` or `length` is invalid, or
     * `p` is not a valid quantile value (p must be greater than 0
     * and less than or equal to 100)
     *
     *
     *
     * See [Percentile] for a description of the percentile estimation
     * algorithm used.
     *
     * @param values array of input values
     * @param p  the percentile to compute
     * @param begin  the first (0-based) element to include in the computation
     * @param length  the number of array elements to include
     * @return  the percentile value
     * @throws MathIllegalArgumentException if the parameters are not valid or the
     * input array is null
     */
    //@Throws(MathIllegalArgumentException::class)
    fun evaluate(
        values: DoubleArray, begin: Int,
        length: Int, p: Double
    ): Double {
        test(values, begin, length)
        if (p > 100 || p <= 0) {
            throw TODO() /* OutOfRangeException(
                LocalizedFormats.OUT_OF_BOUNDS_QUANTILE_VALUE, p, 0, 100
            )*/
        }
        if (length == 0) {
            return Double.NaN
        }
        if (length == 1) {
            return values[begin] // always return single value for n = 1
        }
        val work = getWorkArray(values, begin, length)
        val pivotsHeap = getPivots(values)
        return if (work.isEmpty()) Double.NaN else estimationType.evaluate(work, pivotsHeap, p, kthSelector)
    }


    /**
     * Returns the value of the quantile field (determines what percentile is
     * computed when evaluate() is called with no quantile argument).
     *
     * @return quantile set while construction or [.setQuantile]
     */
    fun getQuantile(): Double {
        return quantile
    }

    /**
     * Sets the value of the quantile field (determines what percentile is
     * computed when evaluate() is called with no quantile argument).
     *
     * @param p a value between 0 < p <= 100
     * @throws MathIllegalArgumentException  if p is not greater than 0 and less
     * than or equal to 100
     */
    //@Throws(MathIllegalArgumentException::class)
    fun setQuantile(p: Double) {
        if (p <= 0 || p > 100) {
            TODO()
            /*throw OutOfRangeException(
                LocalizedFormats.OUT_OF_BOUNDS_QUANTILE_VALUE, p, 0, 100
            )*/
        }
        quantile = p
    }

    /**
     * {@inheritDoc}
     */
    override fun copy(): Percentile {
        return Percentile(this)
    }

    /**
     * Get the work array to operate. Makes use of prior `storedData` if
     * it exists or else do a check on NaNs and copy a subset of the array
     * defined by begin and length parameters. The set [.nanStrategy] will
     * be used to either retain/remove/replace any NaNs present before returning
     * the resultant array.
     *
     * @param values the array of numbers
     * @param begin index to start reading the array
     * @param length the length of array to be read from the begin index
     * @return work array sliced from values in the range [begin,begin+length)
     * @throws MathIllegalArgumentException if values or indices are invalid
     */
    protected fun getWorkArray(values: DoubleArray, begin: Int, length: Int): DoubleArray {
        val work: DoubleArray
        if (values.contentEquals(dataRef)) {
            work = dataRef!!
        } else {
            when (nanStrategy) {
                NaNStrategy.MAXIMAL -> work =
                    replaceAndSlice(values, begin, length, Double.NaN, Double.POSITIVE_INFINITY)

                NaNStrategy.MINIMAL -> work =
                    replaceAndSlice(values, begin, length, Double.NaN, Double.NEGATIVE_INFINITY)

                NaNStrategy.REMOVED -> work = removeAndSlice(values, begin, length, Double.NaN)
                NaNStrategy.FAILED -> {
                    work = copyOf(values, begin, length)
                    MathArrays.checkNotNaN(work)
                }

                else -> work = copyOf(values, begin, length)
            }
        }
        return work
    }

    /**
     * Get pivots which is either cached or a newly created one
     *
     * @param values array containing the input numbers
     * @return cached pivots or a newly created one
     */
    private fun getPivots(values: DoubleArray): IntArray? {
        val pivotsHeap: IntArray?
        if (values.contentEquals(dataRef)) {
            pivotsHeap = cachedPivots
        } else {
            pivotsHeap = IntArray(PIVOTS_HEAP_LENGTH) { -1 }
            // java.util.Arrays.fill(pivotsHeap, -1)
        }
        return pivotsHeap
    }

    /**
     * Build a new instance similar to the current one except for the
     * [estimation type][EstimationType].
     *
     *
     * This method is intended to be used as part of a fluent-type builder
     * pattern. Building finely tune instances should be done as follows:
     *
     * <pre>
     * Percentile customized = new Percentile(quantile).
     * withEstimationType(estimationType).
     * withNaNStrategy(nanStrategy).
     * withKthSelector(kthSelector);
    </pre> *
     *
     *
     * If any of the `withXxx` method is omitted, the default value for
     * the corresponding customization parameter will be used.
     *
     * @param newEstimationType estimation type for the new instance
     * @return a new instance, with changed estimation type
     * @throws NullArgumentException when newEstimationType is null
     */
    fun withEstimationType(newEstimationType: EstimationType): Percentile {
        return Percentile(quantile, newEstimationType, nanStrategy, kthSelector)
    }

    val naNStrategy: NaNStrategy
        /**
         * Get the [NaN Handling][NaNStrategy] strategy used for computation.
         * @return `NaN Handling` strategy set during construction
         */
        get() = nanStrategy

    /**
     * Build a new instance similar to the current one except for the
     * [NaN handling][NaNStrategy] strategy.
     *
     *
     * This method is intended to be used as part of a fluent-type builder
     * pattern. Building finely tune instances should be done as follows:
     *
     * <pre>
     * Percentile customized = new Percentile(quantile).
     * withEstimationType(estimationType).
     * withNaNStrategy(nanStrategy).
     * withKthSelector(kthSelector);
    </pre> *
     *
     *
     * If any of the `withXxx` method is omitted, the default value for
     * the corresponding customization parameter will be used.
     *
     * @param newNaNStrategy NaN strategy for the new instance
     * @return a new instance, with changed NaN handling strategy
     * @throws NullArgumentException when newNaNStrategy is null
     */
    fun withNaNStrategy(newNaNStrategy: NaNStrategy): Percentile {
        return Percentile(quantile, estimationType, newNaNStrategy, kthSelector)
    }

    /**
     * Get the [kthSelector][KthSelector] used for computation.
     * @return the `kthSelector` set
     */
    fun getKthSelector(): KthSelector {
        return kthSelector
    }

    val pivotingStrategy: PivotingStrategyInterface
        /**
         * Get the [PivotingStrategyInterface] used in KthSelector for computation.
         * @return the pivoting strategy set
         */
        get() = kthSelector.getPivotingStrategy()

    /**
     * Build a new instance similar to the current one except for the
     * [kthSelector][KthSelector] instance specifically set.
     *
     *
     * This method is intended to be used as part of a fluent-type builder
     * pattern. Building finely tune instances should be done as follows:
     *
     * <pre>
     * Percentile customized = new Percentile(quantile).
     * withEstimationType(estimationType).
     * withNaNStrategy(nanStrategy).
     * withKthSelector(newKthSelector);
    </pre> *
     *
     *
     * If any of the `withXxx` method is omitted, the default value for
     * the corresponding customization parameter will be used.
     *
     * @param newKthSelector KthSelector for the new instance
     * @return a new instance, with changed KthSelector
     * @throws NullArgumentException when newKthSelector is null
     */
    fun withKthSelector(newKthSelector: KthSelector): Percentile {
        return Percentile(
            quantile, estimationType, nanStrategy,
            newKthSelector
        )
    }

    /**
     * An enum for various estimation strategies of a percentile referred in
     * [wikipedia on quantile](http://en.wikipedia.org/wiki/Quantile)
     * with the names of enum matching those of types mentioned in
     * wikipedia.
     *
     *
     * Each enum corresponding to the specific type of estimation in wikipedia
     * implements  the respective formulae that specializes in the below aspects
     *
     *  * An **index method** to calculate approximate index of the
     * estimate
     *  * An **estimate method** to estimate a value found at the earlier
     * computed index
     *  * A ** minLimit** on the quantile for which first element of sorted
     * input is returned as an estimate
     *  * A ** maxLimit** on the quantile for which last element of sorted
     * input is returned as an estimate
     *
     *
     *
     * Users can now create [Percentile] by explicitly passing this enum;
     * such as by invoking [Percentile.withEstimationType]
     *
     *
     * References:
     *
     *  1.
     * [Wikipedia on quantile](http://en.wikipedia.org/wiki/Quantile)
     *
     *  1.
     * [
 * Hyndman, R. J. and Fan, Y. (1996) Sample quantiles in statistical
 * packages, American Statistician 50, 361–365](https://www.amherst.edu/media/view/129116/.../Sample+Quantiles.pdf)
     *  1.
     * [
 * R-Manual ](http://stat.ethz.ch/R-manual/R-devel/library/stats/html/quantile.html)
     *
     *
     */
    enum class EstimationType
    /**
     * Constructor
     *
     * @param type name of estimation type as per wikipedia
     */(
        /** Simple name such as R-1, R-2 corresponding to those in wikipedia.  */
        val description: String
    ) {
        /**
         * This is the default type used in the [Percentile].This method
         * has the following formulae for index and estimates<br></br>
         * \( \begin{align}
         * &amp;index    = (N+1)p\ \\
         * &amp;estimate = x_{\lceil h\,-\,1/2 \rceil} \\
         * &amp;minLimit = 0 \\
         * &amp;maxLimit = 1 \\
         * \end{align}\)
         */
        LEGACY("Legacy Apache Commons Math") {
            /**
             * {@inheritDoc}.This method in particular makes use of existing
             * Apache Commons Math style of picking up the index.
             */
            override fun index(p: Double, length: Int): Double {
                val minLimit = 0.0
                val maxLimit = 1.0
                return if (p.compareTo(minLimit) == 0) 0.0 else if (
                    p.compareTo(maxLimit) == 0
                ) {
                    length.toDouble()
                } else {
                    p * (length + 1)
                }
            }
        },

        /**
         * The method R_1 has the following formulae for index and estimates<br></br>
         * \( \begin{align}
         * &amp;index= Np + 1/2\,  \\
         * &amp;estimate= x_{\lceil h\,-\,1/2 \rceil} \\
         * &amp;minLimit = 0 \\
         * \end{align}\)
         */
        R_1("R-1") {
            override fun index(p: Double, length: Int): Double {
                val minLimit = 0.0
                return if (p.compareTo(minLimit) == 0) 0.0 else length * p + 0.5
            }

            /**
             * {@inheritDoc}This method in particular for R_1 uses ceil(pos-0.5)
             */
            override fun estimate(
                values: DoubleArray,
                pivotsHeap: IntArray?, pos: Double,
                length: Int, selector: KthSelector
            ): Double {
                return super.estimate(values, pivotsHeap, ceil(pos - 0.5), length, selector)
            }
        },

        /**
         * The method R_2 has the following formulae for index and estimates<br></br>
         * \( \begin{align}
         * &amp;index= Np + 1/2\, \\
         * &amp;estimate=\frac{x_{\lceil h\,-\,1/2 \rceil} +
         * x_{\lfloor h\,+\,1/2 \rfloor}}{2} \\
         * &amp;minLimit = 0 \\
         * &amp;maxLimit = 1 \\
         * \end{align}\)
         */
        R_2("R-2") {
            override fun index(p: Double, length: Int): Double {
                val minLimit = 0.0
                val maxLimit = 1.0
                return if (
                        p.compareTo(maxLimit) == 0
                ) length.toDouble() else if (p.compareTo( minLimit) == 0) 0.0 else length * p + 0.5
            }

            /**
             * {@inheritDoc}This method in particular for R_2 averages the
             * values at ceil(p+0.5) and floor(p-0.5).
             */
            override fun estimate(
                work: DoubleArray,
                pivotsHeap: IntArray?, pos: Double,
                length: Int, selector: KthSelector
            ): Double {
                val low = super.estimate(work, pivotsHeap, ceil(pos - 0.5), length, selector)
                val high = super.estimate(work, pivotsHeap, floor(pos + 0.5), length, selector)
                return (low + high) / 2
            }
        },

        /**
         * The method R_3 has the following formulae for index and estimates<br></br>
         * \( \begin{align}
         * &amp;index= Np \\
         * &amp;estimate= x_{\lfloor h \rceil}\, \\
         * &amp;minLimit = 0.5/N \\
         * \end{align}\)
         */
        R_3("R-3") {
            override fun index(p: Double, length: Int): Double {
                val minLimit = 1.0 / 2 / length
                return if (p.compareTo(minLimit) <= 0) 0.0 else rint(length * p)
            }
        },

        /**
         * The method R_4 has the following formulae for index and estimates<br></br>
         * \( \begin{align}
         * &amp;index= Np\, \\
         * &amp;estimate= x_{\lfloor h \rfloor} + (h -
         * \lfloor h \rfloor) (x_{\lfloor h \rfloor + 1} - x_{\lfloor h
         * \rfloor}) \\
         * &amp;minLimit = 1/N \\
         * &amp;maxLimit = 1 \\
         * \end{align}\)
         */
        R_4("R-4") {
            override fun index(p: Double, length: Int): Double {
                val minLimit = 1.0 / length
                val maxLimit = 1.0
                return if (p.compareTo(minLimit) < 0) 0.0 else if (
                        p.compareTo(maxLimit) == 0
                ) length.toDouble() else length * p
            }
        },

        /**
         * The method R_5 has the following formulae for index and estimates<br></br>
         * \( \begin{align}
         * &amp;index= Np + 1/2\\
         * &amp;estimate= x_{\lfloor h \rfloor} + (h -
         * \lfloor h \rfloor) (x_{\lfloor h \rfloor + 1} - x_{\lfloor h
         * \rfloor}) \\
         * &amp;minLimit = 0.5/N \\
         * &amp;maxLimit = (N-0.5)/N
         * \end{align}\)
         */
        R_5("R-5") {
            override fun index(p: Double, length: Int): Double {
                val minLimit = 1.0 / 2 / length
                val maxLimit = (length - 0.5) / length
                return if (p.compareTo(minLimit) < 0) 0.0 else if (
                        p.compareTo(maxLimit) >= 0
                ) length.toDouble() else length * p + 0.5
            }
        },

        /**
         * The method R_6 has the following formulae for index and estimates<br></br>
         * \( \begin{align}
         * &amp;index= (N + 1)p \\
         * &amp;estimate= x_{\lfloor h \rfloor} + (h -
         * \lfloor h \rfloor) (x_{\lfloor h \rfloor + 1} - x_{\lfloor h
         * \rfloor}) \\
         * &amp;minLimit = 1/(N+1) \\
         * &amp;maxLimit = N/(N+1) \\
         * \end{align}\)
         *
         *
         * **Note:** This method computes the index in a manner very close to
         * the default Commons Math Percentile existing implementation. However
         * the difference to be noted is in picking up the limits with which
         * first element (p&lt;1(N+1)) and last elements (p&gt;N/(N+1))are done.
         * While in default case; these are done with p=0 and p=1 respectively.
         */
        R_6("R-6") {
            override fun index(p: Double, length: Int): Double {
                val minLimit = 1.0 / (length + 1)
                val maxLimit = 1.0 * length / (length + 1)
                return if (p.compareTo(minLimit) < 0) 0.0 else if (
                        p.compareTo(maxLimit) >= 0
                ) length.toDouble() else (length + 1) * p
            }
        },

        /**
         * The method R_7 implements Microsoft Excel style computation has the
         * following formulae for index and estimates.<br></br>
         * \( \begin{align}
         * &amp;index = (N-1)p + 1 \\
         * &amp;estimate = x_{\lfloor h \rfloor} + (h -
         * \lfloor h \rfloor) (x_{\lfloor h \rfloor + 1} - x_{\lfloor h
         * \rfloor}) \\
         * &amp;minLimit = 0 \\
         * &amp;maxLimit = 1 \\
         * \end{align}\)
         */
        R_7("R-7") {
            override fun index(p: Double, length: Int): Double {
                val minLimit = 0.0
                val maxLimit = 1.0
                return if (p.compareTo(minLimit) == 0) 0.0 else if (
                        p.compareTo(maxLimit) == 0
                ) length.toDouble() else 1 + (length - 1) * p
            }
        },

        /**
         * The method R_8 has the following formulae for index and estimates<br></br>
         * \( \begin{align}
         * &amp;index = (N + 1/3)p + 1/3  \\
         * &amp;estimate = x_{\lfloor h \rfloor} + (h -
         * \lfloor h \rfloor) (x_{\lfloor h \rfloor + 1} - x_{\lfloor h
         * \rfloor}) \\
         * &amp;minLimit = (2/3)/(N+1/3) \\
         * &amp;maxLimit = (N-1/3)/(N+1/3) \\
         * \end{align}\)
         *
         *
         * As per Ref [2,3] this approach is most recommended as it provides
         * an approximate median-unbiased estimate regardless of distribution.
         */
        R_8("R-8") {
            override fun index(p: Double, length: Int): Double {
                val minLimit = 2 * (1.0 / 3) / (length + 1.0 / 3)
                val maxLimit = (length - 1.0 / 3) / (length + 1.0 / 3)
                return if (p.compareTo(minLimit) < 0) 0.0 else if (
                        p.compareTo(maxLimit) >= 0
                ) length.toDouble() else (length + 1.0 / 3) * p + 1.0 / 3
            }
        },

        /**
         * The method R_9 has the following formulae for index and estimates<br></br>
         * \( \begin{align}
         * &amp;index = (N + 1/4)p + 3/8\\
         * &amp;estimate = x_{\lfloor h \rfloor} + (h -
         * \lfloor h \rfloor) (x_{\lfloor h \rfloor + 1} - x_{\lfloor h
         * \rfloor}) \\
         * &amp;minLimit = (5/8)/(N+1/4) \\
         * &amp;maxLimit = (N-3/8)/(N+1/4) \\
         * \end{align}\)
         */
        R_9("R-9") {
            override fun index(p: Double, length: Int): Double {
                val minLimit = 5.0 / 8 / (length + 0.25)
                val maxLimit = (length - 3.0 / 8) / (length + 0.25)
                return if (p.compareTo(minLimit) < 0) 0.0 else if (
                        p.compareTo(
                        maxLimit
                    ) >= 0
                ) length.toDouble() else (length + 0.25) * p + 3.0 / 8
            }
        };

        override fun toString(): String {
            return description
        }
        /**
         * Gets the name of the enum
         *
         * @return the name
         */

        /**
         * Finds the index of array that can be used as starting index to
         * [estimate][.estimate]
         * percentile. The calculation of index calculation is specific to each
         * [EstimationType].
         *
         * @param p the p<sup>th</sup> quantile
         * @param length the total number of array elements in the work array
         * @return a computed real valued index as explained in the wikipedia
         */
        protected abstract fun index(p: Double, length: Int): Double

        /**
         * Estimation based on K<sup>th</sup> selection. This may be overridden
         * in specific enums to compute slightly different estimations.
         *
         * @param work array of numbers to be used for finding the percentile
         * @param pos indicated positional index prior computed from calling
         * [.index]
         * @param pivotsHeap an earlier populated cache if exists; will be used
         * @param length size of array considered
         * @param selector a [KthSelector] used for pivoting during search
         * @return estimated percentile
         */
        protected open fun estimate(
            work: DoubleArray, pivotsHeap: IntArray?,
            pos: Double, length: Int,
            selector: KthSelector
        ): Double {
            val fpos: Double = floor(pos)
            val intPos = fpos.toInt()
            val dif = pos - fpos
            if (pos < 1) {
                return selector.select(work, pivotsHeap, 0)
            }
            if (pos >= length) {
                return selector.select(work, pivotsHeap, length - 1)
            }
            val lower: Double = selector.select(work, pivotsHeap, intPos - 1)
            val upper: Double = selector.select(work, pivotsHeap, intPos)
            return lower + dif * (upper - lower)
        }

        /**
         * Evaluate method to compute the percentile for a given bounded array
         * using earlier computed pivots heap.<br></br>
         * This basically calls the [index][.index] and then
         * [estimate][.estimate]
         * functions to return the estimated percentile value.
         *
         * @param work array of numbers to be used for finding the percentile
         * @param pivotsHeap a prior cached heap which can speed up estimation
         * @param p the p<sup>th</sup> quantile to be computed
         * @param selector a [KthSelector] used for pivoting during search
         * @return estimated percentile
         * @throws OutOfRangeException if p is out of range
         * @throws NullArgumentException if work array is null
         */
        fun evaluate(
            work: DoubleArray, pivotsHeap: IntArray?, p: Double,
            selector: KthSelector
        ): Double {
            MathUtils.checkNotNull(work)
            if (p > 100 || p <= 0) {
                throw TODO() /*OutOfRangeException(
                    LocalizedFormats.OUT_OF_BOUNDS_QUANTILE_VALUE,
                    p, 0, 100
                )*/
            }
            return estimate(work, pivotsHeap, index(p / 100.0, work.size), work.size, selector)
        }

        /**
         * Evaluate method to compute the percentile for a given bounded array.
         * This basically calls the [index][.index] and then
         * [estimate][.estimate]
         * functions to return the estimated percentile value. Please
         * note that this method does not make use of cached pivots.
         *
         * @param work array of numbers to be used for finding the percentile
         * @param p the p<sup>th</sup> quantile to be computed
         * @return estimated percentile
         * @param selector a [KthSelector] used for pivoting during search
         * @throws OutOfRangeException if length or p is out of range
         * @throws NullArgumentException if work array is null
         */
        fun evaluate(work: DoubleArray, p: Double, selector: KthSelector): Double {
            return this.evaluate(work, null, p, selector)
        }
    }

    companion object {
        /** Serializable version identifier  */
        private const val serialVersionUID = -8091216485095130416L

        /** Maximum number of partitioning pivots cached (each level double the number of pivots).  */
        private const val MAX_CACHED_LEVELS = 10

        /** Maximum number of cached pivots in the pivots cached array  */
        private const val PIVOTS_HEAP_LENGTH = 0x1 shl MAX_CACHED_LEVELS - 1

        /**
         * Make a copy of the array for the slice defined by array part from
         * [begin, begin+length)
         * @param values the input array
         * @param begin start index of the array to include
         * @param length number of elements to include from begin
         * @return copy of a slice of the original array
         */
        private fun copyOf(values: DoubleArray, begin: Int, length: Int): DoubleArray {
            MathArrays.verifyValues(values, begin, length)
            return MathArrays.copyOfRange(values, begin, begin + length)
        }

        /**
         * Replace every occurrence of a given value with a replacement value in a
         * copied slice of array defined by array part from [begin, begin+length).
         * @param values the input array
         * @param begin start index of the array to include
         * @param length number of elements to include from begin
         * @param original the value to be replaced with
         * @param replacement the value to be used for replacement
         * @return the copy of sliced array with replaced values
         */
        private fun replaceAndSlice(
            values: DoubleArray,
            begin: Int, length: Int,
            original: Double,
            replacement: Double
        ): DoubleArray {
            val temp = copyOf(values, begin, length)
            for (i in 0 until length) {
                temp[i] = if (Precision.equalsIncludingNaN(original, temp[i])) replacement else temp[i]
            }
            return temp
        }

        /**
         * Remove the occurrence of a given value in a copied slice of array
         * defined by the array part from [begin, begin+length).
         * @param values the input array
         * @param begin start index of the array to include
         * @param length number of elements to include from begin
         * @param removedValue the value to be removed from the sliced array
         * @return the copy of the sliced array after removing the removedValue
         */
        private fun removeAndSlice(
            values: DoubleArray,
            begin: Int, length: Int,
            removedValue: Double
        ): DoubleArray {
            MathArrays.verifyValues(values, begin, length)
            val temp: DoubleArray
            //BitSet(length) to indicate where the removedValue is located

            val bits: BitSet = BitSet(length)
            for (i in begin until begin + length) {
                if (Precision.equalsIncludingNaN(removedValue, values[i])) {
                    bits.set(i - begin)
                }
            }
            //Check if empty then create a new copy
            if (bits.isEmpty) {
                temp = copyOf(values, begin, length) // Nothing removed, just copy
            } else if (bits.cardinality == length) {
                temp = DoubleArray(0) // All removed, just empty
            } else {                                   // Some removable, so new
                temp = DoubleArray(length - bits.cardinality)
                var start = begin //start index from source array (i.e values)
                var dest = 0 //dest index in destination array(i.e temp)
                var nextOne = -1 //nextOne is the index of bit set of next one
                var bitSetPtr = 0 //bitSetPtr is start index pointer of bitset
                while (bits.nextSetBit(bitSetPtr).also { nextOne = it } != -1) {
                    val lengthToCopy = nextOne - bitSetPtr
                    values.copyInto(temp, dest, start, lengthToCopy + start)
                    //java.lang.System.arraycopy(values, start, temp, dest, lengthToCopy)
                    dest += lengthToCopy
                    start = begin + bits.nextClearBit(nextOne).also { bitSetPtr = it }
                }
                //Copy any residue past start index till begin+length
                if (start < begin + length) {
                    values.copyInto(temp, dest, start,  begin + length)
                   // java.lang.System.arraycopy(values, start, temp, dest, begin + length - start)
                }
            }
            return temp

        }
    }
}
