/*
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.jetbrains.kotlinx.statistics.math3.stat.descriptive.summary

import org.jetbrains.kotlinx.statistics.math3.stat.descriptive.AbstractStorelessUnivariateStatistic
import org.jetbrains.kotlinx.statistics.math3.util.MathUtils
import kotlin.jvm.JvmOverloads

/**
 * Returns the sum of the available values.
 *
 *
 * If there are no values in the dataset, then 0 is returned.
 * If any of the values are
 * `NaN`, then `NaN` is returned.
 *
 *
 * **Note that this implementation is not synchronized.** If
 * multiple threads access an instance of this class concurrently, and at least
 * one of the threads invokes the `increment()` or
 * `clear()` method, it must be synchronized externally.
 *
 */
class Sum : AbstractStorelessUnivariateStatistic {
    /**
     * {@inheritDoc}
     */
    /**  */
    override var n: Long = 0
        private set
    /**
     * {@inheritDoc}
     */
    /**
     * The currently running sum.
     */
    override var result = 0.0
        private set

    /**
     * Create a Sum instance
     */
    constructor() {
        n = 0
        result = 0.0
    }

    /**
     * Copy constructor, creates a new `Sum` identical
     * to the `original`
     *
     * @param original the `Sum` instance to copy
     * @throws NullArgumentException if original is null
     */
    constructor(original: Sum) {
        copy(original, this)
    }

    /**
     * {@inheritDoc}
     */
    override fun increment(d: Double) {
        result += d
        n++
    }

    override fun incrementAll(values: DoubleArray, start: Int, length: Int) {
        TODO("Not yet implemented")
    }

    /**
     * {@inheritDoc}
     */
    override fun clear() {
        result = 0.0
        n = 0
    }

    /**
     * The sum of the entries in the specified portion of
     * the input array, or 0 if the designated subarray
     * is empty.
     *
     *
     * Throws `MathIllegalArgumentException` if the array is null.
     *
     * @param values the input array
     * @param begin index of the first array element to include
     * @param length the number of elements to include
     * @return the sum of the values or 0 if length = 0
     * @throws MathIllegalArgumentException if the array is null or the array index
     * parameters are not valid
     */
    //@Throws(MathIllegalArgumentException::class)
    override fun evaluate(values: DoubleArray, begin: Int, length: Int): Double {
        var sum = Double.NaN
        if (test(values, begin, length, true)) {
            sum = 0.0
            for (i in begin until begin + length) {
                sum += values[i]
            }
        }
        return sum
    }
    /**
     * The weighted sum of the entries in the specified portion of
     * the input array, or 0 if the designated subarray
     * is empty.
     *
     *
     * Throws `MathIllegalArgumentException` if any of the following are true:
     *  * the values array is null
     *  * the weights array is null
     *  * the weights array does not have the same length as the values array
     *  * the weights array contains one or more infinite values
     *  * the weights array contains one or more NaN values
     *  * the weights array contains negative values
     *  * the start and length arguments do not determine a valid array
     *
     *
     *
     * Uses the formula, <pre>
     * weighted sum = (values[i] * weights[i])
    </pre> *
     *
     * @param values the input array
     * @param weights the weights array
     * @param begin index of the first array element to include
     * @param length the number of elements to include
     * @return the sum of the values or 0 if length = 0
     * @throws MathIllegalArgumentException if the parameters are not valid
     * @since 2.1
     */
    /**
     * The weighted sum of the entries in the the input array.
     *
     *
     * Throws `MathIllegalArgumentException` if any of the following are true:
     *  * the values array is null
     *  * the weights array is null
     *  * the weights array does not have the same length as the values array
     *  * the weights array contains one or more infinite values
     *  * the weights array contains one or more NaN values
     *  * the weights array contains negative values
     *
     *
     *
     * Uses the formula, <pre>
     * weighted sum = (values[i] * weights[i])
    </pre> *
     *
     * @param values the input array
     * @param weights the weights array
     * @return the sum of the values or Double.NaN if length = 0
     * @throws MathIllegalArgumentException if the parameters are not valid
     * @since 2.1
     */
    @JvmOverloads
    //@Throws(MathIllegalArgumentException::class)
    fun evaluate(
        values: DoubleArray, weights: DoubleArray,
        begin: Int = 0, length: Int = values.size
    ): Double {
        var sum = Double.NaN
        if (test(values, weights, begin, length, true)) {
            sum = 0.0
            for (i in begin until begin + length) {
                sum += values[i] * weights[i]
            }
        }
        return sum
    }

    /**
     * {@inheritDoc}
     */
    override fun copy(): Sum {
        val result = Sum()
        // No try-catch or advertised exception because args are valid
        copy(this, result)
        return result
    }

    companion object {
        /** Serializable version identifier  */
        private const val serialVersionUID = -8231831954703408316L

        /**
         * Copies source to dest.
         *
         * Neither source nor dest can be null.
         *
         * @param source Sum to copy
         * @param dest Sum to copy to
         * @throws NullArgumentException if either source or dest is null
         */
        //@Throws(NullArgumentException::class)
        fun copy(source: Sum, dest: Sum) {
            MathUtils.checkNotNull(source)
            MathUtils.checkNotNull(dest)
            dest.setData(source.dataRef)
            dest.n = source.n
            dest.result = source.result
        }
    }
}
