/*
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.jetbrains.kotlinx.statistics.math3.stat.descriptive.summary

import org.jetbrains.kotlinx.statistics.math3.stat.descriptive.AbstractStorelessUnivariateStatistic
import org.jetbrains.kotlinx.statistics.math3.util.MathUtils


/**
 * Returns the sum of the natural logs for this collection of values.
 *
 *
 * Uses [org.apache.commons.math3.util.FastMath.log] to compute the logs.
 * Therefore,
 *
 *  * If any of values are &lt; 0, the result is `NaN.`
 *  * If all values are non-negative and less than
 * `Double.POSITIVE_INFINITY`,  but at least one value is 0, the
 * result is `Double.NEGATIVE_INFINITY.`
 *  * If both `Double.POSITIVE_INFINITY` and
 * `Double.NEGATIVE_INFINITY` are among the values, the result is
 * `NaN.`
 *
 *
 *
 * **Note that this implementation is not synchronized.** If
 * multiple threads access an instance of this class concurrently, and at least
 * one of the threads invokes the `increment()` or
 * `clear()` method, it must be synchronized externally.
 *
 */
class SumOfLogs : AbstractStorelessUnivariateStatistic {
    /**Number of values that have been added  */
    override var n: Long = 0
    /**
     * {@inheritDoc}
     */
    /**
     * The currently running value
     */
    override var result = 0.0
        private set

    /**
     * Create a SumOfLogs instance
     */
    constructor() {
        result = 0.0
        n = 0
    }

    /**
     * Copy constructor, creates a new `SumOfLogs` identical
     * to the `original`
     *
     * @param original the `SumOfLogs` instance to copy
     * @throws NullArgumentException if original is null
     */
    constructor(original: SumOfLogs) {
        copy(original, this)
    }

    /**
     * {@inheritDoc}
     */
    override fun increment(d: Double) {
        result += kotlin.math.ln(d)
        n++
    }


    /**
     * {@inheritDoc}
     */
    override fun clear() {
        result = 0.0
        n = 0
    }

    /**
     * Returns the sum of the natural logs of the entries in the specified portion of
     * the input array, or `Double.NaN` if the designated subarray
     * is empty.
     *
     *
     * Throws `MathIllegalArgumentException` if the array is null.
     *
     *
     * See [SumOfLogs].
     *
     * @param values the input array
     * @param begin index of the first array element to include
     * @param length the number of elements to include
     * @return the sum of the natural logs of the values or 0 if
     * length = 0
     * @throws MathIllegalArgumentException if the array is null or the array index
     * parameters are not valid
     */
    //@Throws(MathIllegalArgumentException::class)
    override fun evaluate(values: DoubleArray, begin: Int, length: Int): Double {
        var sumLog = Double.NaN
        if (test(values, begin, length, true)) {
            sumLog = 0.0
            for (i in begin until begin + length) {
                sumLog += kotlin.math.ln(values[i])
            }
        }
        return sumLog
    }

    /**
     * {@inheritDoc}
     */
    override fun copy(): SumOfLogs {
        val result = SumOfLogs()
        // No try-catch or advertised exception here because args are valid
        copy(this, result)
        return result
    }

    companion object {
        /** Serializable version identifier  */
        private const val serialVersionUID = -370076995648386763L

        /**
         * Copies source to dest.
         *
         * Neither source nor dest can be null.
         *
         * @param source SumOfLogs to copy
         * @param dest SumOfLogs to copy to
         * @throws NullArgumentException if either source or dest is null
         */
        //@Throws(NullArgumentException::class)
        fun copy(source: SumOfLogs, dest: SumOfLogs) {
            MathUtils.checkNotNull(source)
            MathUtils.checkNotNull(dest)
            dest.setData(source.dataRef)
            dest.n = source.n
            dest.result = source.result
        }
    }
}
