/*
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.jetbrains.kotlinx.statistics.math3.stat.descriptive.summary

import org.jetbrains.kotlinx.statistics.math3.stat.descriptive.AbstractStorelessUnivariateStatistic
import org.jetbrains.kotlinx.statistics.math3.util.MathUtils


/**
 * Returns the sum of the squares of the available values.
 *
 *
 * If there are no values in the dataset, then 0 is returned.
 * If any of the values are
 * `NaN`, then `NaN` is returned.
 *
 *
 * **Note that this implementation is not synchronized.** If
 * multiple threads access an instance of this class concurrently, and at least
 * one of the threads invokes the `increment()` or
 * `clear()` method, it must be synchronized externally.
 *
 */
class SumOfSquares : AbstractStorelessUnivariateStatistic {
    /**
     * {@inheritDoc}
     */
    /**  */
    override var n: Long = 0
        private set
    /**
     * {@inheritDoc}
     */
    /**
     * The currently running sumSq
     */
    override var result = 0.0
        private set

    /**
     * Create a SumOfSquares instance
     */
    constructor() {
        n = 0
        result = 0.0
    }

    /**
     * Copy constructor, creates a new `SumOfSquares` identical
     * to the `original`
     *
     * @param original the `SumOfSquares` instance to copy
     * @throws NullArgumentException if original is null
     */
    constructor(original: SumOfSquares) {
        copy(original, this)
    }

    /**
     * {@inheritDoc}
     */
    override fun increment(d: Double) {
        result += d * d
        n++
    }

    /**
     * {@inheritDoc}
     */
    override fun clear() {
        result = 0.0
        n = 0
    }

    /**
     * Returns the sum of the squares of the entries in the specified portion of
     * the input array, or `Double.NaN` if the designated subarray
     * is empty.
     *
     *
     * Throws `MathIllegalArgumentException` if the array is null.
     *
     * @param values the input array
     * @param begin index of the first array element to include
     * @param length the number of elements to include
     * @return the sum of the squares of the values or 0 if length = 0
     * @throws MathIllegalArgumentException if the array is null or the array index
     * parameters are not valid
     */
    //@Throws(MathIllegalArgumentException::class)
    override fun evaluate(values: DoubleArray, begin: Int, length: Int): Double {
        var sumSq = Double.NaN
        if (test(values, begin, length, true)) {
            sumSq = 0.0
            for (i in begin until begin + length) {
                sumSq += values[i] * values[i]
            }
        }
        return sumSq
    }

    /**
     * {@inheritDoc}
     */
    override fun copy(): SumOfSquares {
        val result = SumOfSquares()
        // no try-catch or advertised exception here because args are valid
        copy(this, result)
        return result
    }

    companion object {
        /** Serializable version identifier  */
        private const val serialVersionUID = 1460986908574398008L

        /**
         * Copies source to dest.
         *
         * Neither source nor dest can be null.
         *
         * @param source SumOfSquares to copy
         * @param dest SumOfSquares to copy to
         * @throws NullArgumentException if either source or dest is null
         */
        //@Throws(NullArgumentException::class)
        fun copy(source: SumOfSquares, dest: SumOfSquares) {
            MathUtils.checkNotNull(source)
            MathUtils.checkNotNull(dest)
            dest.setData(source.dataRef)
            dest.n = source.n
            dest.result = source.result
        }
    }
}
