/*
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.jetbrains.kotlinx.statistics.math3.stat.inference

import org.jetbrains.kotlinx.statistics.math3.distribution.TDistribution
import org.jetbrains.kotlinx.statistics.math3.stat.StatUtils
import org.jetbrains.kotlinx.statistics.math3.stat.descriptive.StatisticalSummary
import kotlin.math.abs
import kotlin.math.sqrt

/**
 * An implementation for Student's t-tests.
 *
 *
 * Tests can be:
 *  * One-sample or two-sample
 *  * One-sided or two-sided
 *  * Paired or unpaired (for two-sample tests)
 *  * Homoscedastic (equal variance assumption) or heteroscedastic
 * (for two sample tests)
 *  * Fixed significance level (boolean-valued) or returning p-values.
 *
 *
 *
 * Test statistics are available for all tests.  Methods including "Test" in
 * in their names perform tests, all other methods return t-statistics.  Among
 * the "Test" methods, `double-`valued methods return p-values;
 * `boolean-`valued methods perform fixed significance level tests.
 * Significance levels are always specified as numbers between 0 and 0.5
 * (e.g. tests at the 95% level  use `alpha=0.05`).
 *
 *
 * Input to tests can be either `double[]` arrays or
 * [StatisticalSummary] instances.
 *
 *
 * Uses commons-math [org.apache.commons.math3.distribution.TDistribution]
 * implementation to estimate exact p-values.
 *
 */
class TTest {
    /**
     * Computes a paired, 2-sample t-statistic based on the data in the input
     * arrays.  The t-statistic returned is equivalent to what would be returned by
     * computing the one-sample t-statistic [.t], with
     * `mu = 0` and the sample array consisting of the (signed)
     * differences between corresponding entries in `sample1` and
     * `sample2.`
     *
     *
     * **Preconditions**:
     *  * The input arrays must have the same length and their common length
     * must be at least 2.
     *
     *
     * @param sample1 array of sample data values
     * @param sample2 array of sample data values
     * @return t statistic
     * //@Throws NullArgumentException if the arrays are `null`
     * //@Throws NoDataException if the arrays are empty
     * //@Throws DimensionMismatchException if the length of the arrays is not equal
     * //@Throws NumberIsTooSmallException if the length of the arrays is &lt; 2
     */
    /*//@Throws(
        NullArgumentException::class,
        NoDataException::class,
        DimensionMismatchException::class,
        NumberIsTooSmallException::class
    )*/
    fun pairedT(sample1: DoubleArray, sample2: DoubleArray): Double {
        checkSampleData(sample1)
        checkSampleData(sample2)
        val meanDifference: Double = StatUtils.meanDifference(sample1, sample2)
        return t(
            meanDifference, 0.0,
            StatUtils.varianceDifference(sample1, sample2, meanDifference),
            sample1.size.toDouble()
        )
    }

    /**
     * Returns the *observed significance level*, or
     * * p-value*, associated with a paired, two-sample, two-tailed t-test
     * based on the data in the input arrays.
     *
     *
     * The number returned is the smallest significance level
     * at which one can reject the null hypothesis that the mean of the paired
     * differences is 0 in favor of the two-sided alternative that the mean paired
     * difference is not equal to 0. For a one-sided test, divide the returned
     * value by 2.
     *
     *
     * This test is equivalent to a one-sample t-test computed using
     * [.tTest] with `mu = 0` and the sample
     * array consisting of the signed differences between corresponding elements of
     * `sample1` and `sample2.`
     *
     *
     * **Usage Note:**<br></br>
     * The validity of the p-value depends on the assumptions of the parametric
     * t-test procedure, as discussed
     * [
 * here](http://www.basic.nwu.edu/statguidefiles/ttest_unpaired_ass_viol.html)
     *
     *
     * **Preconditions**:
     *  * The input array lengths must be the same and their common length must
     * be at least 2.
     *
     *
     * @param sample1 array of sample data values
     * @param sample2 array of sample data values
     * @return p-value for t-test
     * //@Throws NullArgumentException if the arrays are `null`
     * //@Throws NoDataException if the arrays are empty
     * //@Throws DimensionMismatchException if the length of the arrays is not equal
     * //@Throws NumberIsTooSmallException if the length of the arrays is &lt; 2
     * //@Throws MaxCountExceededException if an error occurs computing the p-value
     */
    /*//@Throws(
        NullArgumentException::class,
        NoDataException::class,
        DimensionMismatchException::class,
        NumberIsTooSmallException::class,
        MaxCountExceededException::class
    )*/
    fun pairedTTest(sample1: DoubleArray, sample2: DoubleArray): Double {
        val meanDifference: Double = StatUtils.meanDifference(sample1, sample2)
        return tTest(
            meanDifference, 0.0,
            StatUtils.varianceDifference(sample1, sample2, meanDifference),
            sample1.size.toDouble()
        )
    }

    /**
     * Performs a paired t-test evaluating the null hypothesis that the
     * mean of the paired differences between `sample1` and
     * `sample2` is 0 in favor of the two-sided alternative that the
     * mean paired difference is not equal to 0, with significance level
     * `alpha`.
     *
     *
     * Returns `true` iff the null hypothesis can be rejected with
     * confidence `1 - alpha`.  To perform a 1-sided test, use
     * `alpha * 2`
     *
     *
     * **Usage Note:**<br></br>
     * The validity of the test depends on the assumptions of the parametric
     * t-test procedure, as discussed
     * [
 * here](http://www.basic.nwu.edu/statguidefiles/ttest_unpaired_ass_viol.html)
     *
     *
     * **Preconditions**:
     *  * The input array lengths must be the same and their common length
     * must be at least 2.
     *
     *  *  ` 0 < alpha < 0.5 `
     *
     *
     * @param sample1 array of sample data values
     * @param sample2 array of sample data values
     * @param alpha significance level of the test
     * @return true if the null hypothesis can be rejected with
     * confidence 1 - alpha
     * //@Throws NullArgumentException if the arrays are `null`
     * //@Throws NoDataException if the arrays are empty
     * //@Throws DimensionMismatchException if the length of the arrays is not equal
     * //@Throws NumberIsTooSmallException if the length of the arrays is &lt; 2
     * //@Throws OutOfRangeException if `alpha` is not in the range (0, 0.5]
     * //@Throws MaxCountExceededException if an error occurs computing the p-value
     */
    /*//@Throws(
        NullArgumentException::class,
        NoDataException::class,
        DimensionMismatchException::class,
        NumberIsTooSmallException::class,
        OutOfRangeException::class,
        MaxCountExceededException::class
    )*/
    fun pairedTTest(
        sample1: DoubleArray,
        sample2: DoubleArray,
        alpha: Double
    ): Boolean {
        checkSignificanceLevel(alpha)
        return pairedTTest(sample1, sample2) < alpha
    }

    /**
     * Computes a [
 * t statistic ](http://www.itl.nist.gov/div898/handbook/prc/section2/prc22.htm#formula) given observed values and a comparison constant.
     *
     *
     * This statistic can be used to perform a one sample t-test for the mean.
     *
     *
     * **Preconditions**:
     *  * The observed array length must be at least 2.
     *
     *
     * @param mu comparison constant
     * @param observed array of values
     * @return t statistic
     * //@Throws NullArgumentException if `observed` is `null`
     * //@Throws NumberIsTooSmallException if the length of `observed` is &lt; 2
     */
    //@Throws(NullArgumentException::class, NumberIsTooSmallException::class)
    fun t(mu: Double, observed: DoubleArray): Double {
        checkSampleData(observed)
        // No try-catch or advertised exception because args have just been checked
        return t(
            StatUtils.mean(observed), mu, StatUtils.variance(observed),
            observed.size.toDouble()
        )
    }

    /**
     * Computes a [
 * t statistic ](http://www.itl.nist.gov/div898/handbook/prc/section2/prc22.htm#formula) to use in comparing the mean of the dataset described by
     * `sampleStats` to `mu`.
     *
     *
     * This statistic can be used to perform a one sample t-test for the mean.
     *
     *
     * **Preconditions**:
     *  * `observed.n  2`.
     *
     *
     * @param mu comparison constant
     * @param sampleStats DescriptiveStatistics holding sample summary statitstics
     * @return t statistic
     * //@Throws NullArgumentException if `sampleStats` is `null`
     * //@Throws NumberIsTooSmallException if the number of samples is &lt; 2
     */
    //@Throws(NullArgumentException::class, NumberIsTooSmallException::class)
    fun t(mu: Double, sampleStats: StatisticalSummary): Double {
        checkSampleData(sampleStats)
        return t(
            sampleStats.mean, mu, sampleStats.variance,
            sampleStats.n.toDouble()
        )
    }

    /**
     * Computes a 2-sample t statistic,  under the hypothesis of equal
     * subpopulation variances.  To compute a t-statistic without the
     * equal variances hypothesis, use [.t].
     *
     *
     * This statistic can be used to perform a (homoscedastic) two-sample
     * t-test to compare sample means.
     *
     *
     * The t-statistic is
     *
     *
     * &nbsp;&nbsp;`  t = (m1 - m2) / (sqrt(1/n1 +1/n2) sqrt(var))`
     *
     *
     * where **`n1`** is the size of first sample;
     * **` n2`** is the size of second sample;
     * **` m1`** is the mean of first sample;
     * **` m2`** is the mean of second sample
     *
     * and **`var`** is the pooled variance estimate:
     *
     *
     * `var = sqrt(((n1 - 1)var1 + (n2 - 1)var2) / ((n1-1) + (n2-1)))`
     *
     *
     * with **`var1`** the variance of the first sample and
     * **`var2`** the variance of the second sample.
     *
     *
     * **Preconditions**:
     *  * The observed array lengths must both be at least 2.
     *
     *
     * @param sample1 array of sample data values
     * @param sample2 array of sample data values
     * @return t statistic
     * //@Throws NullArgumentException if the arrays are `null`
     * //@Throws NumberIsTooSmallException if the length of the arrays is &lt; 2
     */
    //@Throws(NullArgumentException::class, NumberIsTooSmallException::class)
    fun homoscedasticT(sample1: DoubleArray, sample2: DoubleArray): Double {
        checkSampleData(sample1)
        checkSampleData(sample2)
        // No try-catch or advertised exception because args have just been checked
        return homoscedasticT(
            StatUtils.mean(sample1), StatUtils.mean(sample2),
            StatUtils.variance(sample1), StatUtils.variance(sample2),
            sample1.size.toDouble(), sample2.size.toDouble()
        )
    }

    /**
     * Computes a 2-sample t statistic, without the hypothesis of equal
     * subpopulation variances.  To compute a t-statistic assuming equal
     * variances, use [.homoscedasticT].
     *
     *
     * This statistic can be used to perform a two-sample t-test to compare
     * sample means.
     *
     *
     * The t-statistic is
     *
     *
     * &nbsp;&nbsp; `  t = (m1 - m2) / sqrt(var1/n1 + var2/n2)`
     *
     *
     * where **`n1`** is the size of the first sample
     * **` n2`** is the size of the second sample;
     * **` m1`** is the mean of the first sample;
     * **` m2`** is the mean of the second sample;
     * **` var1`** is the variance of the first sample;
     * **` var2`** is the variance of the second sample;
     *
     *
     * **Preconditions**:
     *  * The observed array lengths must both be at least 2.
     *
     *
     * @param sample1 array of sample data values
     * @param sample2 array of sample data values
     * @return t statistic
     * //@Throws NullArgumentException if the arrays are `null`
     * //@Throws NumberIsTooSmallException if the length of the arrays is &lt; 2
     */
    //@Throws(NullArgumentException::class, NumberIsTooSmallException::class)
    fun t(sample1: DoubleArray, sample2: DoubleArray): Double {
        checkSampleData(sample1)
        checkSampleData(sample2)
        // No try-catch or advertised exception because args have just been checked
        return t(
            StatUtils.mean(sample1), StatUtils.mean(sample2),
            StatUtils.variance(sample1), StatUtils.variance(sample2),
            sample1.size.toDouble(), sample2.size.toDouble()
        )
    }

    /**
     * Computes a 2-sample t statistic , comparing the means of the datasets
     * described by two [StatisticalSummary] instances, without the
     * assumption of equal subpopulation variances.  Use
     * [.homoscedasticT] to
     * compute a t-statistic under the equal variances assumption.
     *
     *
     * This statistic can be used to perform a two-sample t-test to compare
     * sample means.
     *
     *
     * The returned  t-statistic is
     *
     *
     * &nbsp;&nbsp; `  t = (m1 - m2) / sqrt(var1/n1 + var2/n2)`
     *
     *
     * where **`n1`** is the size of the first sample;
     * **` n2`** is the size of the second sample;
     * **` m1`** is the mean of the first sample;
     * **` m2`** is the mean of the second sample
     * **` var1`** is the variance of the first sample;
     * **` var2`** is the variance of the second sample
     *
     *
     * **Preconditions**:
     *  * The datasets described by the two Univariates must each contain
     * at least 2 observations.
     *
     *
     * @param sampleStats1 StatisticalSummary describing data from the first sample
     * @param sampleStats2 StatisticalSummary describing data from the second sample
     * @return t statistic
     * //@Throws NullArgumentException if the sample statistics are `null`
     * //@Throws NumberIsTooSmallException if the number of samples is &lt; 2
     */
    //@Throws(NullArgumentException::class, NumberIsTooSmallException::class)
    fun t(
        sampleStats1: StatisticalSummary,
        sampleStats2: StatisticalSummary
    ): Double {
        checkSampleData(sampleStats1)
        checkSampleData(sampleStats2)
        return t(
            sampleStats1.mean, sampleStats2.mean,
            sampleStats1.variance, sampleStats2.variance,
            sampleStats1.n.toDouble(), sampleStats2.n.toDouble()
        )
    }

    /**
     * Computes a 2-sample t statistic, comparing the means of the datasets
     * described by two [StatisticalSummary] instances, under the
     * assumption of equal subpopulation variances.  To compute a t-statistic
     * without the equal variances assumption, use
     * [.t].
     *
     *
     * This statistic can be used to perform a (homoscedastic) two-sample
     * t-test to compare sample means.
     *
     *
     * The t-statistic returned is
     *
     *
     * &nbsp;&nbsp;`  t = (m1 - m2) / (sqrt(1/n1 +1/n2) sqrt(var))`
     *
     *
     * where **`n1`** is the size of first sample;
     * **` n2`** is the size of second sample;
     * **` m1`** is the mean of first sample;
     * **` m2`** is the mean of second sample
     * and **`var`** is the pooled variance estimate:
     *
     *
     * `var = sqrt(((n1 - 1)var1 + (n2 - 1)var2) / ((n1-1) + (n2-1)))`
     *
     *
     * with **`var1`** the variance of the first sample and
     * **`var2`** the variance of the second sample.
     *
     *
     * **Preconditions**:
     *  * The datasets described by the two Univariates must each contain
     * at least 2 observations.
     *
     *
     * @param sampleStats1 StatisticalSummary describing data from the first sample
     * @param sampleStats2 StatisticalSummary describing data from the second sample
     * @return t statistic
     * //@Throws NullArgumentException if the sample statistics are `null`
     * //@Throws NumberIsTooSmallException if the number of samples is &lt; 2
     */
    //@Throws(NullArgumentException::class, NumberIsTooSmallException::class)
    fun homoscedasticT(
        sampleStats1: StatisticalSummary,
        sampleStats2: StatisticalSummary
    ): Double {
        checkSampleData(sampleStats1)
        checkSampleData(sampleStats2)
        return homoscedasticT(
            sampleStats1.mean, sampleStats2.mean,
            sampleStats1.variance, sampleStats2.variance,
            sampleStats1.n.toDouble(), sampleStats2.n.toDouble()
        )
    }

    /**
     * Returns the *observed significance level*, or
     * *p-value*, associated with a one-sample, two-tailed t-test
     * comparing the mean of the input array with the constant `mu`.
     *
     *
     * The number returned is the smallest significance level
     * at which one can reject the null hypothesis that the mean equals
     * `mu` in favor of the two-sided alternative that the mean
     * is different from `mu`. For a one-sided test, divide the
     * returned value by 2.
     *
     *
     * **Usage Note:**<br></br>
     * The validity of the test depends on the assumptions of the parametric
     * t-test procedure, as discussed
     * [here](http://www.basic.nwu.edu/statguidefiles/ttest_unpaired_ass_viol.html)
     *
     *
     * **Preconditions**:
     *  * The observed array length must be at least 2.
     *
     *
     * @param mu constant value to compare sample mean against
     * @param sample array of sample data values
     * @return p-value
     * //@Throws NullArgumentException if the sample array is `null`
     * //@Throws NumberIsTooSmallException if the length of the array is &lt; 2
     * //@Throws MaxCountExceededException if an error occurs computing the p-value
     */
    //@Throws(NullArgumentException::class, NumberIsTooSmallException::class, MaxCountExceededException::class)
    fun tTest(mu: Double, sample: DoubleArray): Double {
        checkSampleData(sample)
        // No try-catch or advertised exception because args have just been checked
        return tTest(
            StatUtils.mean(sample), mu, StatUtils.variance(sample),
            sample.size.toDouble()
        )
    }

    /**
     * Performs a [
 * two-sided t-test](http://www.itl.nist.gov/div898/handbook/eda/section3/eda353.htm) evaluating the null hypothesis that the mean of the population from
     * which `sample` is drawn equals `mu`.
     *
     *
     * Returns `true` iff the null hypothesis can be
     * rejected with confidence `1 - alpha`.  To
     * perform a 1-sided test, use `alpha * 2`
     *
     *
     * **Examples:**<br></br>
     *  1. To test the (2-sided) hypothesis `sample mean = mu ` at
     * the 95% level, use <br></br>`tTest(mu, sample, 0.05) `
     *
     *  1. To test the (one-sided) hypothesis ` sample mean < mu `
     * at the 99% level, first verify that the measured sample mean is less
     * than `mu` and then use
     * <br></br>`tTest(mu, sample, 0.02) `
     *
     *
     *
     * **Usage Note:**<br></br>
     * The validity of the test depends on the assumptions of the one-sample
     * parametric t-test procedure, as discussed
     * [here](http://www.basic.nwu.edu/statguidefiles/sg_glos.html#one-sample)
     *
     *
     * **Preconditions**:
     *  * The observed array length must be at least 2.
     *
     *
     * @param mu constant value to compare sample mean against
     * @param sample array of sample data values
     * @param alpha significance level of the test
     * @return p-value
     * //@Throws NullArgumentException if the sample array is `null`
     * //@Throws NumberIsTooSmallException if the length of the array is &lt; 2
     * //@Throws OutOfRangeException if `alpha` is not in the range (0, 0.5]
     * //@Throws MaxCountExceededException if an error computing the p-value
     */
    /*//@Throws(
        NullArgumentException::class,
        NumberIsTooSmallException::class,
        OutOfRangeException::class,
        MaxCountExceededException::class
    )*/
    fun tTest(mu: Double, sample: DoubleArray, alpha: Double): Boolean {
        checkSignificanceLevel(alpha)
        return tTest(mu, sample) < alpha
    }

    /**
     * Returns the *observed significance level*, or
     * *p-value*, associated with a one-sample, two-tailed t-test
     * comparing the mean of the dataset described by `sampleStats`
     * with the constant `mu`.
     *
     *
     * The number returned is the smallest significance level
     * at which one can reject the null hypothesis that the mean equals
     * `mu` in favor of the two-sided alternative that the mean
     * is different from `mu`. For a one-sided test, divide the
     * returned value by 2.
     *
     *
     * **Usage Note:**<br></br>
     * The validity of the test depends on the assumptions of the parametric
     * t-test procedure, as discussed
     * [
 * here](http://www.basic.nwu.edu/statguidefiles/ttest_unpaired_ass_viol.html)
     *
     *
     * **Preconditions**:
     *  * The sample must contain at least 2 observations.
     *
     *
     * @param mu constant value to compare sample mean against
     * @param sampleStats StatisticalSummary describing sample data
     * @return p-value
     * //@Throws NullArgumentException if `sampleStats` is `null`
     * //@Throws NumberIsTooSmallException if the number of samples is &lt; 2
     * //@Throws MaxCountExceededException if an error occurs computing the p-value
     */
    //@Throws(NullArgumentException::class, NumberIsTooSmallException::class, MaxCountExceededException::class)
    fun tTest(mu: Double, sampleStats: StatisticalSummary): Double {
        checkSampleData(sampleStats)
        return tTest(
            sampleStats.mean, mu, sampleStats.variance,
            sampleStats.n.toDouble()
        )
    }

    /**
     * Performs a [
 * two-sided t-test](http://www.itl.nist.gov/div898/handbook/eda/section3/eda353.htm) evaluating the null hypothesis that the mean of the
     * population from which the dataset described by `stats` is
     * drawn equals `mu`.
     *
     *
     * Returns `true` iff the null hypothesis can be rejected with
     * confidence `1 - alpha`.  To  perform a 1-sided test, use
     * `alpha * 2.`
     *
     *
     * **Examples:**<br></br>
     *  1. To test the (2-sided) hypothesis `sample mean = mu ` at
     * the 95% level, use <br></br>`tTest(mu, sampleStats, 0.05) `
     *
     *  1. To test the (one-sided) hypothesis ` sample mean < mu `
     * at the 99% level, first verify that the measured sample mean is less
     * than `mu` and then use
     * <br></br>`tTest(mu, sampleStats, 0.02) `
     *
     *
     *
     * **Usage Note:**<br></br>
     * The validity of the test depends on the assumptions of the one-sample
     * parametric t-test procedure, as discussed
     * [here](http://www.basic.nwu.edu/statguidefiles/sg_glos.html#one-sample)
     *
     *
     * **Preconditions**:
     *  * The sample must include at least 2 observations.
     *
     *
     * @param mu constant value to compare sample mean against
     * @param sampleStats StatisticalSummary describing sample data values
     * @param alpha significance level of the test
     * @return p-value
     * //@Throws NullArgumentException if `sampleStats` is `null`
     * //@Throws NumberIsTooSmallException if the number of samples is &lt; 2
     * //@Throws OutOfRangeException if `alpha` is not in the range (0, 0.5]
     * //@Throws MaxCountExceededException if an error occurs computing the p-value
     */
    /*//@Throws(
        NullArgumentException::class,
        NumberIsTooSmallException::class,
        OutOfRangeException::class,
        MaxCountExceededException::class
    )*/
    fun tTest(
        mu: Double,
        sampleStats: StatisticalSummary,
        alpha: Double
    ): Boolean {
        checkSignificanceLevel(alpha)
        return tTest(mu, sampleStats) < alpha
    }

    /**
     * Returns the *observed significance level*, or
     * *p-value*, associated with a two-sample, two-tailed t-test
     * comparing the means of the input arrays.
     *
     *
     * The number returned is the smallest significance level
     * at which one can reject the null hypothesis that the two means are
     * equal in favor of the two-sided alternative that they are different.
     * For a one-sided test, divide the returned value by 2.
     *
     *
     * The test does not assume that the underlying popuation variances are
     * equal  and it uses approximated degrees of freedom computed from the
     * sample data to compute the p-value.  The t-statistic used is as defined in
     * [.t] and the Welch-Satterthwaite approximation
     * to the degrees of freedom is used,
     * as described
     * [
 * here.](http://www.itl.nist.gov/div898/handbook/prc/section3/prc31.htm)  To perform the test under the assumption of equal subpopulation
     * variances, use [.homoscedasticTTest].
     *
     *
     * **Usage Note:**<br></br>
     * The validity of the p-value depends on the assumptions of the parametric
     * t-test procedure, as discussed
     * [
 * here](http://www.basic.nwu.edu/statguidefiles/ttest_unpaired_ass_viol.html)
     *
     *
     * **Preconditions**:
     *  * The observed array lengths must both be at least 2.
     *
     *
     * @param sample1 array of sample data values
     * @param sample2 array of sample data values
     * @return p-value for t-test
     * //@Throws NullArgumentException if the arrays are `null`
     * //@Throws NumberIsTooSmallException if the length of the arrays is &lt; 2
     * //@Throws MaxCountExceededException if an error occurs computing the p-value
     */
    //@Throws(NullArgumentException::class, NumberIsTooSmallException::class, MaxCountExceededException::class)
    fun tTest(sample1: DoubleArray, sample2: DoubleArray): Double {
        checkSampleData(sample1)
        checkSampleData(sample2)
        // No try-catch or advertised exception because args have just been checked
        return tTest(
            StatUtils.mean(sample1), StatUtils.mean(sample2),
            StatUtils.variance(sample1), StatUtils.variance(sample2),
            sample1.size.toDouble(), sample2.size.toDouble()
        )
    }

    /**
     * Returns the *observed significance level*, or
     * *p-value*, associated with a two-sample, two-tailed t-test
     * comparing the means of the input arrays, under the assumption that
     * the two samples are drawn from subpopulations with equal variances.
     * To perform the test without the equal variances assumption, use
     * [.tTest].
     *
     *
     * The number returned is the smallest significance level
     * at which one can reject the null hypothesis that the two means are
     * equal in favor of the two-sided alternative that they are different.
     * For a one-sided test, divide the returned value by 2.
     *
     *
     * A pooled variance estimate is used to compute the t-statistic.  See
     * [.homoscedasticT]. The sum of the sample sizes
     * minus 2 is used as the degrees of freedom.
     *
     *
     * **Usage Note:**<br></br>
     * The validity of the p-value depends on the assumptions of the parametric
     * t-test procedure, as discussed
     * [
 * here](http://www.basic.nwu.edu/statguidefiles/ttest_unpaired_ass_viol.html)
     *
     *
     * **Preconditions**:
     *  * The observed array lengths must both be at least 2.
     *
     *
     * @param sample1 array of sample data values
     * @param sample2 array of sample data values
     * @return p-value for t-test
     * //@Throws NullArgumentException if the arrays are `null`
     * //@Throws NumberIsTooSmallException if the length of the arrays is &lt; 2
     * //@Throws MaxCountExceededException if an error occurs computing the p-value
     */
    //@Throws(NullArgumentException::class, NumberIsTooSmallException::class, MaxCountExceededException::class)
    fun homoscedasticTTest(sample1: DoubleArray, sample2: DoubleArray): Double {
        checkSampleData(sample1)
        checkSampleData(sample2)
        // No try-catch or advertised exception because args have just been checked
        return homoscedasticTTest(
            StatUtils.mean(sample1),
            StatUtils.mean(sample2),
            StatUtils.variance(sample1),
            StatUtils.variance(sample2),
            sample1.size.toDouble(), sample2.size.toDouble()
        )
    }

    /**
     * Performs a
     * [
 * two-sided t-test](http://www.itl.nist.gov/div898/handbook/eda/section3/eda353.htm) evaluating the null hypothesis that `sample1`
     * and `sample2` are drawn from populations with the same mean,
     * with significance level `alpha`.  This test does not assume
     * that the subpopulation variances are equal.  To perform the test assuming
     * equal variances, use
     * [.homoscedasticTTest].
     *
     *
     * Returns `true` iff the null hypothesis that the means are
     * equal can be rejected with confidence `1 - alpha`.  To
     * perform a 1-sided test, use `alpha * 2`
     *
     *
     * See [.t] for the formula used to compute the
     * t-statistic.  Degrees of freedom are approximated using the
     * [
 * Welch-Satterthwaite approximation.](http://www.itl.nist.gov/div898/handbook/prc/section3/prc31.htm)
     *
     *
     * **Examples:**<br></br>
     *  1. To test the (2-sided) hypothesis `mean 1 = mean 2 ` at
     * the 95% level,  use
     * <br></br>`tTest(sample1, sample2, 0.05). `
     *
     *  1. To test the (one-sided) hypothesis ` mean 1 < mean 2 `,
     * at the 99% level, first verify that the measured  mean of `sample 1`
     * is less than the mean of `sample 2` and then use
     * <br></br>`tTest(sample1, sample2, 0.02) `
     *
     *
     *
     * **Usage Note:**<br></br>
     * The validity of the test depends on the assumptions of the parametric
     * t-test procedure, as discussed
     * [
 * here](http://www.basic.nwu.edu/statguidefiles/ttest_unpaired_ass_viol.html)
     *
     *
     * **Preconditions**:
     *  * The observed array lengths must both be at least 2.
     *
     *  *  ` 0 < alpha < 0.5 `
     *
     *
     * @param sample1 array of sample data values
     * @param sample2 array of sample data values
     * @param alpha significance level of the test
     * @return true if the null hypothesis can be rejected with
     * confidence 1 - alpha
     * //@Throws NullArgumentException if the arrays are `null`
     * //@Throws NumberIsTooSmallException if the length of the arrays is &lt; 2
     * //@Throws OutOfRangeException if `alpha` is not in the range (0, 0.5]
     * //@Throws MaxCountExceededException if an error occurs computing the p-value
     */
    /*//@Throws(
        NullArgumentException::class,
        NumberIsTooSmallException::class,
        OutOfRangeException::class,
        MaxCountExceededException::class
    )*/
    fun tTest(
        sample1: DoubleArray,
        sample2: DoubleArray,
        alpha: Double
    ): Boolean {
        checkSignificanceLevel(alpha)
        return tTest(sample1, sample2) < alpha
    }

    /**
     * Performs a
     * [
 * two-sided t-test](http://www.itl.nist.gov/div898/handbook/eda/section3/eda353.htm) evaluating the null hypothesis that `sample1`
     * and `sample2` are drawn from populations with the same mean,
     * with significance level `alpha`,  assuming that the
     * subpopulation variances are equal.  Use
     * [.tTest] to perform the test without
     * the assumption of equal variances.
     *
     *
     * Returns `true` iff the null hypothesis that the means are
     * equal can be rejected with confidence `1 - alpha`.  To
     * perform a 1-sided test, use `alpha * 2.`  To perform the test
     * without the assumption of equal subpopulation variances, use
     * [.tTest].
     *
     *
     * A pooled variance estimate is used to compute the t-statistic. See
     * [.t] for the formula. The sum of the sample
     * sizes minus 2 is used as the degrees of freedom.
     *
     *
     * **Examples:**<br></br>
     *  1. To test the (2-sided) hypothesis `mean 1 = mean 2 ` at
     * the 95% level, use <br></br>`tTest(sample1, sample2, 0.05). `
     *
     *  1. To test the (one-sided) hypothesis ` mean 1 < mean 2, `
     * at the 99% level, first verify that the measured mean of
     * `sample 1` is less than the mean of `sample 2`
     * and then use
     * <br></br>`tTest(sample1, sample2, 0.02) `
     *
     *
     *
     * **Usage Note:**<br></br>
     * The validity of the test depends on the assumptions of the parametric
     * t-test procedure, as discussed
     * [
 * here](http://www.basic.nwu.edu/statguidefiles/ttest_unpaired_ass_viol.html)
     *
     *
     * **Preconditions**:
     *  * The observed array lengths must both be at least 2.
     *
     *  *  ` 0 < alpha < 0.5 `
     *
     *
     * @param sample1 array of sample data values
     * @param sample2 array of sample data values
     * @param alpha significance level of the test
     * @return true if the null hypothesis can be rejected with
     * confidence 1 - alpha
     * //@Throws NullArgumentException if the arrays are `null`
     * //@Throws NumberIsTooSmallException if the length of the arrays is &lt; 2
     * //@Throws OutOfRangeException if `alpha` is not in the range (0, 0.5]
     * //@Throws MaxCountExceededException if an error occurs computing the p-value
     */
    /*//@Throws(
        NullArgumentException::class,
        NumberIsTooSmallException::class,
        OutOfRangeException::class,
        MaxCountExceededException::class
    )*/
    fun homoscedasticTTest(
        sample1: DoubleArray,
        sample2: DoubleArray,
        alpha: Double
    ): Boolean {
        checkSignificanceLevel(alpha)
        return homoscedasticTTest(sample1, sample2) < alpha
    }

    /**
     * Returns the *observed significance level*, or
     * *p-value*, associated with a two-sample, two-tailed t-test
     * comparing the means of the datasets described by two StatisticalSummary
     * instances.
     *
     *
     * The number returned is the smallest significance level
     * at which one can reject the null hypothesis that the two means are
     * equal in favor of the two-sided alternative that they are different.
     * For a one-sided test, divide the returned value by 2.
     *
     *
     * The test does not assume that the underlying population variances are
     * equal  and it uses approximated degrees of freedom computed from the
     * sample data to compute the p-value.   To perform the test assuming
     * equal variances, use
     * [.homoscedasticTTest].
     *
     *
     * **Usage Note:**<br></br>
     * The validity of the p-value depends on the assumptions of the parametric
     * t-test procedure, as discussed
     * [
 * here](http://www.basic.nwu.edu/statguidefiles/ttest_unpaired_ass_viol.html)
     *
     *
     * **Preconditions**:
     *  * The datasets described by the two Univariates must each contain
     * at least 2 observations.
     *
     *
     * @param sampleStats1  StatisticalSummary describing data from the first sample
     * @param sampleStats2  StatisticalSummary describing data from the second sample
     * @return p-value for t-test
     * //@Throws NullArgumentException if the sample statistics are `null`
     * //@Throws NumberIsTooSmallException if the number of samples is &lt; 2
     * //@Throws MaxCountExceededException if an error occurs computing the p-value
     */
    //@Throws(NullArgumentException::class, NumberIsTooSmallException::class, MaxCountExceededException::class)
    fun tTest(
        sampleStats1: StatisticalSummary,
        sampleStats2: StatisticalSummary
    ): Double {
        checkSampleData(sampleStats1)
        checkSampleData(sampleStats2)
        return tTest(
            sampleStats1.mean, sampleStats2.mean,
            sampleStats1.variance, sampleStats2.variance,
            sampleStats1.n.toDouble(), sampleStats2.n.toDouble()
        )
    }

    /**
     * Returns the *observed significance level*, or
     * *p-value*, associated with a two-sample, two-tailed t-test
     * comparing the means of the datasets described by two StatisticalSummary
     * instances, under the hypothesis of equal subpopulation variances. To
     * perform a test without the equal variances assumption, use
     * [.tTest].
     *
     *
     * The number returned is the smallest significance level
     * at which one can reject the null hypothesis that the two means are
     * equal in favor of the two-sided alternative that they are different.
     * For a one-sided test, divide the returned value by 2.
     *
     *
     * See [.homoscedasticT] for the formula used to
     * compute the t-statistic. The sum of the  sample sizes minus 2 is used as
     * the degrees of freedom.
     *
     *
     * **Usage Note:**<br></br>
     * The validity of the p-value depends on the assumptions of the parametric
     * t-test procedure, as discussed
     * [here](http://www.basic.nwu.edu/statguidefiles/ttest_unpaired_ass_viol.html)
     *
     *
     * **Preconditions**:
     *  * The datasets described by the two Univariates must each contain
     * at least 2 observations.
     *
     *
     * @param sampleStats1  StatisticalSummary describing data from the first sample
     * @param sampleStats2  StatisticalSummary describing data from the second sample
     * @return p-value for t-test
     * //@Throws NullArgumentException if the sample statistics are `null`
     * //@Throws NumberIsTooSmallException if the number of samples is &lt; 2
     * //@Throws MaxCountExceededException if an error occurs computing the p-value
     */
    //@Throws(NullArgumentException::class, NumberIsTooSmallException::class, MaxCountExceededException::class)
    fun homoscedasticTTest(
        sampleStats1: StatisticalSummary,
        sampleStats2: StatisticalSummary
    ): Double {
        checkSampleData(sampleStats1)
        checkSampleData(sampleStats2)
        return homoscedasticTTest(
            sampleStats1.mean,
            sampleStats2.mean,
            sampleStats1.variance,
            sampleStats2.variance,
            sampleStats1.n.toDouble(), sampleStats2.n.toDouble()
        )
    }

    /**
     * Performs a
     * [
 * two-sided t-test](http://www.itl.nist.gov/div898/handbook/eda/section3/eda353.htm) evaluating the null hypothesis that
     * `sampleStats1` and `sampleStats2` describe
     * datasets drawn from populations with the same mean, with significance
     * level `alpha`.   This test does not assume that the
     * subpopulation variances are equal.  To perform the test under the equal
     * variances assumption, use
     * [.homoscedasticTTest].
     *
     *
     * Returns `true` iff the null hypothesis that the means are
     * equal can be rejected with confidence `1 - alpha`.  To
     * perform a 1-sided test, use `alpha * 2`
     *
     *
     * See [.t] for the formula used to compute the
     * t-statistic.  Degrees of freedom are approximated using the
     * [
 * Welch-Satterthwaite approximation.](http://www.itl.nist.gov/div898/handbook/prc/section3/prc31.htm)
     *
     *
     * **Examples:**<br></br>
     *  1. To test the (2-sided) hypothesis `mean 1 = mean 2 ` at
     * the 95%, use
     * <br></br>`tTest(sampleStats1, sampleStats2, 0.05) `
     *
     *  1. To test the (one-sided) hypothesis ` mean 1 < mean 2 `
     * at the 99% level,  first verify that the measured mean of
     * `sample 1` is less than  the mean of `sample 2`
     * and then use
     * <br></br>`tTest(sampleStats1, sampleStats2, 0.02) `
     *
     *
     *
     * **Usage Note:**<br></br>
     * The validity of the test depends on the assumptions of the parametric
     * t-test procedure, as discussed
     * [
 * here](http://www.basic.nwu.edu/statguidefiles/ttest_unpaired_ass_viol.html)
     *
     *
     * **Preconditions**:
     *  * The datasets described by the two Univariates must each contain
     * at least 2 observations.
     *
     *  *  ` 0 < alpha < 0.5 `
     *
     *
     * @param sampleStats1 StatisticalSummary describing sample data values
     * @param sampleStats2 StatisticalSummary describing sample data values
     * @param alpha significance level of the test
     * @return true if the null hypothesis can be rejected with
     * confidence 1 - alpha
     * //@Throws NullArgumentException if the sample statistics are `null`
     * //@Throws NumberIsTooSmallException if the number of samples is &lt; 2
     * //@Throws OutOfRangeException if `alpha` is not in the range (0, 0.5]
     * //@Throws MaxCountExceededException if an error occurs computing the p-value
     */
    /*//@Throws(
        NullArgumentException::class,
        NumberIsTooSmallException::class,
        OutOfRangeException::class,
        MaxCountExceededException::class
    )*/
    fun tTest(
        sampleStats1: StatisticalSummary,
        sampleStats2: StatisticalSummary,
        alpha: Double
    ): Boolean {
        checkSignificanceLevel(alpha)
        return tTest(sampleStats1, sampleStats2) < alpha
    }
    //----------------------------------------------- Protected methods
    /**
     * Computes approximate degrees of freedom for 2-sample t-test.
     *
     * @param v1 first sample variance
     * @param v2 second sample variance
     * @param n1 first sample n
     * @param n2 second sample n
     * @return approximate degrees of freedom
     */
    protected fun df(v1: Double, v2: Double, n1: Double, n2: Double): Double {
        return (v1 / n1 + v2 / n2) * (v1 / n1 + v2 / n2) /
                (v1 * v1 / (n1 * n1 * (n1 - 1.0)) + v2 * v2 /
                        (n2 * n2 * (n2 - 1.0)))
    }

    /**
     * Computes t test statistic for 1-sample t-test.
     *
     * @param m sample mean
     * @param mu constant to test against
     * @param v sample variance
     * @param n sample n
     * @return t test statistic
     */
    protected fun t(
        m: Double, mu: Double,
        v: Double, n: Double
    ): Double {
        return (m - mu) / sqrt(v / n)
    }

    /**
     * Computes t test statistic for 2-sample t-test.
     *
     *
     * Does not assume that subpopulation variances are equal.
     *
     * @param m1 first sample mean
     * @param m2 second sample mean
     * @param v1 first sample variance
     * @param v2 second sample variance
     * @param n1 first sample n
     * @param n2 second sample n
     * @return t test statistic
     */
    protected fun t(
        m1: Double, m2: Double,
        v1: Double, v2: Double,
        n1: Double, n2: Double
    ): Double {
        return (m1 - m2) / sqrt(v1 / n1 + v2 / n2)
    }

    /**
     * Computes t test statistic for 2-sample t-test under the hypothesis
     * of equal subpopulation variances.
     *
     * @param m1 first sample mean
     * @param m2 second sample mean
     * @param v1 first sample variance
     * @param v2 second sample variance
     * @param n1 first sample n
     * @param n2 second sample n
     * @return t test statistic
     */
    protected fun homoscedasticT(
        m1: Double, m2: Double,
        v1: Double, v2: Double,
        n1: Double, n2: Double
    ): Double {
        val pooledVariance = ((n1 - 1) * v1 + (n2 - 1) * v2) / (n1 + n2 - 2)
        return (m1 - m2) / sqrt(pooledVariance * (1.0 / n1 + 1.0 / n2))
    }

    /**
     * Computes p-value for 2-sided, 1-sample t-test.
     *
     * @param m sample mean
     * @param mu constant to test against
     * @param v sample variance
     * @param n sample n
     * @return p-value
     * //@Throws MaxCountExceededException if an error occurs computing the p-value
     * //@Throws MathIllegalArgumentException if n is not greater than 1
     */
    //@Throws(MaxCountExceededException::class, MathIllegalArgumentException::class)
    protected fun tTest(
        m: Double, mu: Double,
        v: Double, n: Double
    ): Double {
        val t: Double = abs(t(m, mu, v, n))
        // pass a null rng to avoid unneeded overhead as we will not sample from this distribution
        //val distribution = TDistribution(null, n - 1)
        return 2.0 * TDistribution.cumulativeProbability(-t, n-1)
    }

    /**
     * Computes p-value for 2-sided, 2-sample t-test.
     *
     *
     * Does not assume subpopulation variances are equal. Degrees of freedom
     * are estimated from the data.
     *
     * @param m1 first sample mean
     * @param m2 second sample mean
     * @param v1 first sample variance
     * @param v2 second sample variance
     * @param n1 first sample n
     * @param n2 second sample n
     * @return p-value
     * //@Throws MaxCountExceededException if an error occurs computing the p-value
     * //@Throws NotStrictlyPositiveException if the estimated degrees of freedom is not
     * strictly positive
     */
    //@Throws(MaxCountExceededException::class, NotStrictlyPositiveException::class)
    protected fun tTest(
        m1: Double, m2: Double,
        v1: Double, v2: Double,
        n1: Double, n2: Double
    ): Double {
        val t: Double = abs(t(m1, m2, v1, v2, n1, n2))
        val degreesOfFreedom = df(v1, v2, n1, n2)
        // pass a null rng to avoid unneeded overhead as we will not sample from this distribution
        //val distribution = TDistribution(null, degreesOfFreedom)
        return 2.0 * TDistribution.cumulativeProbability(-t, degreesOfFreedom)
    }

    /**
     * Computes p-value for 2-sided, 2-sample t-test, under the assumption
     * of equal subpopulation variances.
     *
     *
     * The sum of the sample sizes minus 2 is used as degrees of freedom.
     *
     * @param m1 first sample mean
     * @param m2 second sample mean
     * @param v1 first sample variance
     * @param v2 second sample variance
     * @param n1 first sample n
     * @param n2 second sample n
     * @return p-value
     * //@Throws MaxCountExceededException if an error occurs computing the p-value
     * //@Throws NotStrictlyPositiveException if the estimated degrees of freedom is not
     * strictly positive
     */
    //@Throws(MaxCountExceededException::class, NotStrictlyPositiveException::class)
    protected fun homoscedasticTTest(
        m1: Double, m2: Double,
        v1: Double, v2: Double,
        n1: Double, n2: Double
    ): Double {
        val t: Double = abs(homoscedasticT(m1, m2, v1, v2, n1, n2))
        val degreesOfFreedom = n1 + n2 - 2
        // pass a null rng to avoid unneeded overhead as we will not sample from this distribution
        //val distribution = TDistribution(degreesOfFreedom)
        return 2.0 * TDistribution.cumulativeProbability(-t, degreesOfFreedom)
    }

    /**
     * Check significance level.
     *
     * @param alpha significance level
     * //@Throws OutOfRangeException if the significance level is out of bounds.
     */
    //@Throws(OutOfRangeException::class)
    private fun checkSignificanceLevel(alpha: Double) {
        if (alpha <= 0 || alpha > 0.5) {
            throw TODO()/*OutOfRangeException(
                LocalizedFormats.SIGNIFICANCE_LEVEL,
                alpha, 0.0, 0.5
            )*/
        }
    }

    /**
     * Check sample data.
     *
     * @param data Sample data.
     * //@Throws NullArgumentException if `data` is `null`.
     * //@Throws NumberIsTooSmallException if there is not enough sample data.
     */
    ////@Throws(NullArgumentException::class, NumberIsTooSmallException::class)
    private fun checkSampleData(data: DoubleArray) {
        /*if (data == null) {
            throw NullArgumentException()
        }*/
        if (data.size < 2) {
            throw TODO()/*NumberIsTooSmallException(
                LocalizedFormats.INSUFFICIENT_DATA_FOR_T_STATISTIC,
                data.size, 2, true
            )*/
        }
    }

    /**
     * Check sample data.
     *
     * @param stat Statistical summary.
     * //@Throws NullArgumentException if `data` is `null`.
     * //@Throws NumberIsTooSmallException if there is not enough sample data.
     */
    ////@Throws(NullArgumentException::class, NumberIsTooSmallException::class)
    private fun checkSampleData(stat: StatisticalSummary) {
        /*if (stat == null) {
            throw NullArgumentException()
        }*/
        
        if (stat.n < 2) {
            TODO()
            /*throw NumberIsTooSmallException(
                LocalizedFormats.INSUFFICIENT_DATA_FOR_T_STATISTIC,
                stat.n, 2, true
            )*/
        }
    }
}
