package org.jetbrains.kotlinx.statistics.regression

fun sumOfSquaredDeviations(xVals: DoubleArray, meanX: Double): Double {
    return sumOfDeviationProducts(xVals, xVals, meanX, meanX)
}

fun sumOfDeviationProducts(xVals: DoubleArray, yVals: DoubleArray, meanX: Double, meanY: Double): Double {
    return (xVals zip yVals).sumOf { (x, y) -> (x - meanX) * (y - meanY) }
}

class LinearRegression private constructor (
    n: Int,
    meanX: Double,
    sumXX: Double,
    model: (Double) -> Double,
    standardErrorOfEstimate: Double,
    tCritical: Double
) : RegressionEvaluator(n, meanX, sumXX, model, standardErrorOfEstimate, tCritical) {
    companion object {
        fun fit(xs: List<Double?>, ys: List<Double?>, confidenceLevel: Double): LinearRegression? {
            check(xs, ys, confidenceLevel)

            // Prepare data
            val (xVals, yVals) = xs.filterNotNull().toDoubleArray() to ys.filterNotNull().toDoubleArray()
            val n = xVals.size
            val degreesOfFreedom = n - 2.0

            // Check computability
            if (n <= 1) {
                return null
            }

            // Calculate standard stats
            val meanX = xVals.average()
            val sumXX = sumOfSquaredDeviations(xVals, meanX)

            // Prepare model
            val meanY = yVals.average()
            val sumXY = sumOfDeviationProducts(xVals, yVals, meanX, meanY)
            val slope = sumXY / sumXX
            val intercept = meanY - slope * meanX
            val model: (Double) -> Double = { x -> slope * x + intercept }

            return LinearRegression(
                n,
                meanX,
                sumXX,
                model,
                calcStandardErrorOfEstimate(xVals, yVals, model, degreesOfFreedom),
                calcTCritical(degreesOfFreedom, confidenceLevel)
            )
        }
    }
}