package org.jetbrains.kotlinx.statistics.regression

import org.jetbrains.kotlinx.statistics.math3.analysis.polynomials.PolynomialFunction
import org.jetbrains.kotlinx.statistics.math3.analysis.polynomials.times

private fun averageByX(lst: List<Pair<Double, Double>>): Pair<List<Double>, List<Double>> {

    if (lst.isEmpty())
        return Pair(ArrayList<Double>(), ArrayList<Double>())

    val tx = ArrayList<Double>()
    val ty = ArrayList<Double>()
    var (prevX, sumY) = lst.first()
    var countY = 1

    for ((x, y) in lst.asSequence().drop(1)) {
        if (x == prevX) {
            sumY += y
            ++countY
        } else {
            tx.add(prevX)
            ty.add(sumY.div(countY))
            prevX = x
            sumY = y
            countY = 1
        }
    }

    tx.add(prevX)
    ty.add(sumY.div(countY))

    return Pair(tx, ty)
}


private fun finitePairs(xs: List<Double?>, ys: List<Double?>): ArrayList<Pair<Double, Double>> {
    val res = ArrayList<Pair<Double, Double>>()

    for ((x, y) in xs.asSequence().zip(ys.asSequence())) {
      //  if (SeriesUtil.allFinite(x, y)) {
            res.add(Pair(x!!, y!!))
      //  }
    }

    return res
}

fun averageByX(xs: List<Double?>, ys: List<Double?>): Pair<DoubleArray, DoubleArray> {
    val tp = finitePairs(xs, ys)
    tp.sortBy { it.first }
    val res = averageByX(tp)
    return Pair(res.first.toDoubleArray(), res.second.toDoubleArray())
}

class PolynomialRegression private constructor (
    n: Int,
    meanX: Double,
    sumXX: Double,
    model: (Double) -> Double,
    standardErrorOfEstimate: Double,
    tCritical: Double
) : RegressionEvaluator(n, meanX, sumXX, model, standardErrorOfEstimate, tCritical) {
    companion object {
        fun fit(xs: List<Double?>, ys: List<Double?>, confidenceLevel: Double, deg: Int): PolynomialRegression? {
            check(xs, ys, confidenceLevel)
            require(deg >= 2) { "Degree of polynomial must be at least 2" }

            // Prepare data
            val (xVals, yVals) = averageByX(xs, ys)
            val n = xVals.size
            val degreesOfFreedom = n - deg - 1.0

            // Check computability
            if (n <= deg) {
                return null
            }

            // Calculate standard stats
            val meanX = xVals.average()
            val sumXX = sumOfSquaredDeviations(xVals, meanX)

            // Prepare model
            val polynomial = calculatePolynomial(deg, xVals, yVals)
            val model: (Double) -> Double = { x -> polynomial.value(x) }

            return PolynomialRegression(
                n,
                meanX,
                sumXX,
                model,
                calcStandardErrorOfEstimate(xVals, yVals, model, degreesOfFreedom),
                calcTCritical(degreesOfFreedom, confidenceLevel)
            )
        }

        private fun calculatePolynomial(deg: Int, xVals: DoubleArray, yVals: DoubleArray): PolynomialFunction {
            val fpg = ForsythePolynomialGenerator(xVals)
            var res = PolynomialFunction(doubleArrayOf(0.0))

            for (i in 0..deg) {
                val p = fpg.getPolynomial(i)
                val s = coefficient(p, xVals, yVals)
                res += s * p
            }

            return res
        }

        private fun coefficient(p: PolynomialFunction, xVals: DoubleArray, yVals: DoubleArray): Double {
            var ww = 0.0
            var w = 0.0
            for (i in xVals.indices) {
                val x = xVals[i]
                val y = yVals[i]
                val pval = p.value(x)

                ww += pval * pval
                w += y * pval
            }

            return w / ww
        }
    }
}