package org.jetbrains.kotlinx.statistics.kandy.layers

import org.jetbrains.kotlinx.dataframe.DataColumn
import org.jetbrains.kotlinx.dataframe.DataFrame
import org.jetbrains.kotlinx.dataframe.api.asIterable
import org.jetbrains.kotlinx.dataframe.api.column
import org.jetbrains.kotlinx.dataframe.columns.ColumnAccessor
import org.jetbrains.kotlinx.kandy.dsl.internal.dataframe.DataFramePlotBuilder
import org.jetbrains.kotlinx.kandy.dsl.internal.LayerCreatorScope
import org.jetbrains.kotlinx.statistics.kandy.layers.context.BoxplotLayerBuilder
import org.jetbrains.kotlinx.statistics.kandy.util.internalDataFrame
import org.jetbrains.kotlinx.statistics.plotting.boxplot.models.StatBoxplotFrame
import org.jetbrains.kotlinx.statistics.plotting.boxplot.statBoxplot
import org.jetbrains.kotlinx.statistics.plotting.boxplotOutliers.models.StatBoxplotOutliersFrame
import org.jetbrains.kotlinx.statistics.plotting.boxplotOutliers.statBoxplotOutliers

@Suppress("INVISIBLE_MEMBER")
@PublishedApi
internal fun <X> boxplot(
    parent: LayerCreatorScope,
    statBoxplotFrame: StatBoxplotFrame<X>,
    statBoxplotOutliersFrame: StatBoxplotOutliersFrame<X>,
    xAxisName: String,
    yAxisName: String,
    block: BoxplotLayerBuilder<X>.() -> Unit
) {
    BoxplotLayerBuilder<X>(parent, statBoxplotFrame, statBoxplotOutliersFrame).apply {
        this.boxes {
            x(Stat.x)
            yMin(Stat.min)
            lower(Stat.lower)
            middle(Stat.middle)
            upper(Stat.upper)
            yMax(Stat.max)

            x.axis.name = xAxisName
            y.axis.name = yAxisName
        }

        outliers {
            x(Stat.x)
            y(Stat.y)
        }
    }.apply(block).apply {
        parent.createLayer(boxes, {})
        if (outliers.show) {
            parent.createLayer(outliers, {})
        }
    }
}

@PublishedApi
internal fun <T> boxAndOutliers(
    x: DataColumn<T>,
    y: DataColumn<Number>,
    whiskerIQRRatio: Double,
): Pair<StatBoxplotFrame<T>, StatBoxplotOutliersFrame<T>> {
    return statBoxplot(x, y, whiskerIQRRatio) to statBoxplotOutliers(x, y, whiskerIQRRatio)
}

@PublishedApi
internal fun <T> boxAndOutliers(
    df: DataFrame<*>,
    x: ColumnAccessor<T>,
    y: ColumnAccessor<Number>,
    whiskerIQRRatio: Double,
): Pair<StatBoxplotFrame<T>, StatBoxplotOutliersFrame<T>> {
    return df.statBoxplot(x, y, whiskerIQRRatio) to df.statBoxplotOutliers(x, y, whiskerIQRRatio)
}

@PublishedApi
internal fun boxAndOutliers(
    y: Iterable<Number>,
    whiskerIQRRatio: Double,
    name: String
): Pair<StatBoxplotFrame<String>, StatBoxplotOutliersFrame<String>> {
    return statBoxplot(y, whiskerIQRRatio, name) to statBoxplotOutliers(y, whiskerIQRRatio, name)
}


public fun DataFramePlotBuilder<*>.boxplot(
    sample: Iterable<Number>,
    whiskerIQRRatio: Double = 1.5,
    xName: String = "x",
    yName: String = "y",
    block: BoxplotLayerBuilder<String>.() -> Unit = {}
) {
    val (boxes, outliers) = boxAndOutliers(sample, whiskerIQRRatio, xName)
    boxplot(this, boxes, outliers, xName, yName, block)
}

public fun DataFramePlotBuilder<*>.boxplot(
    y: String,
    whiskerIQRRatio: Double = 1.5,
    xName: String = "x",
    block: BoxplotLayerBuilder<String>.() -> Unit = {}
) {
    val (boxes, outliers) = boxAndOutliers((internalDataFrame[y] as DataColumn<Number>).asIterable(), whiskerIQRRatio, xName)
    boxplot(this, boxes, outliers, xName, y, block)
}

public fun DataFramePlotBuilder<*>.boxplot(
    y: ColumnAccessor<Number>,
    whiskerIQRRatio: Double = 1.5,
    xName: String = "x",
    block: BoxplotLayerBuilder<String>.() -> Unit = {}
) {
    val (boxes, outliers) = boxAndOutliers(internalDataFrame[y].asIterable(), whiskerIQRRatio, xName)
    boxplot(this, boxes, outliers, xName, y.name(), block)
}

public fun DataFramePlotBuilder<*>.boxplot(
    y: DataColumn<Number>,
    whiskerIQRRatio: Double = 1.5,
    xName: String = "x",
    block: BoxplotLayerBuilder<String>.() -> Unit = {}
) {
    val (boxes, outliers) = boxAndOutliers(y.asIterable(), whiskerIQRRatio, xName)
    boxplot(this, boxes, outliers, xName, y.name(), block)
}

public fun <T> DataFramePlotBuilder<*>.boxplot(
    x: ColumnAccessor<T>,
    y: ColumnAccessor<Number>,
    whiskerIQRRatio: Double = 1.5,
    block: BoxplotLayerBuilder<T>.() -> Unit = {}
) {
    val (boxes, outliers) = boxAndOutliers(internalDataFrame, x, y, whiskerIQRRatio)
    boxplot(this, boxes, outliers, x.name(), y.name(), block)
}

public fun DataFramePlotBuilder<*>.boxplot(
    x: String,
    y: String,
    whiskerIQRRatio: Double = 1.5,
    block: BoxplotLayerBuilder<Any?>.() -> Unit = {}
) {
    boxplot(column(x), column(y), whiskerIQRRatio, block)
}

public fun <T> DataFramePlotBuilder<*>.boxplot(
    x: DataColumn<T>,
    y: DataColumn<Number>,
    whiskerIQRRatio: Double = 1.5,
    block: BoxplotLayerBuilder<T>.() -> Unit = {}
) {
    val (boxes, outliers) = boxAndOutliers(x, y, whiskerIQRRatio)
    boxplot(this, boxes, outliers, x.name(), y.name(), block)
}
