package org.jetbrains.kotlinx.statistics.kandy.layers

import kotlinx.datetime.Instant
import kotlinx.datetime.LocalDate
import kotlinx.datetime.LocalTime
import org.jetbrains.kotlinx.dataframe.DataColumn
import org.jetbrains.kotlinx.dataframe.columns.ColumnAccessor
import org.jetbrains.kotlinx.dataframe.columns.ColumnReference
import org.jetbrains.kotlinx.kandy.dsl.categorical
import org.jetbrains.kotlinx.kandy.dsl.internal.dataframe.DataFramePlotBuilder
import org.jetbrains.kotlinx.kandy.dsl.internal.LayerCreatorScope
import org.jetbrains.kotlinx.kandy.ir.scale.Scale
import org.jetbrains.kotlinx.kandy.letsplot.scales.guide.LegendType
import org.jetbrains.kotlinx.kandy.letsplot.tooltips.tooltips
import org.jetbrains.kotlinx.statistics.kandy.layers.context.CandlestickBuilder
import org.jetbrains.kotlinx.statistics.kandy.util.internalDataFrame
import org.jetbrains.kotlinx.statistics.plotting.candlestick.models.StatCandlestickFrame
import org.jetbrains.kotlinx.statistics.plotting.candlestick.statCandlestick
import java.time.LocalDateTime
import kotlin.reflect.typeOf

import org.jetbrains.kotlinx.kandy.letsplot.internal.*

@Suppress("INVISIBLE_MEMBER")
@PublishedApi
internal fun <T> candlestick(parent: LayerCreatorScope, data: StatCandlestickFrame<T>, axisName: String, block: CandlestickBuilder<T>.() -> Unit) {
    parent.createLayer(CandlestickBuilder(parent, data)) {
        x(Stat.x)
        yMin(Stat.min)
        lower(Stat.lower)
        middle(List(Stat.x.size()) { null })
        upper(Stat.upper)
        yMax(Stat.max)
        x.axis.name = axisName
        y.axis.name = ""
        val type = Stat.x.type()
        // TODO
        val xFormat = when (type) {
            typeOf<LocalDate>(), typeOf<LocalDateTime>(), typeOf<Instant>() -> "%d/%m/%y"
            typeOf<LocalTime>() -> "%H:%M%"
            else -> null
        }
        val formats: Map<ColumnReference<*>, String> = xFormat?.let { mapOf(Stat.x to it) } ?: mapOf()
        tooltips(title = "@x") {
            //line(axisName, Stat.x.tooltipValue(xFormat))
            Stat.x.tooltipValue(xFormat)
            line("open", Stat.open.tooltipValue(".2f"))
            line("high", Stat.max.tooltipValue(".2f"))
            line("low", Stat.min.tooltipValue(".2f"))
            line("close", Stat.close.tooltipValue(".2f"))
        }
        whiskerWidth = 0.0
        block()
        val (fillOnInc, fillOnDec) = increase.fillColor to decrease.fillColor
        if (FILL !in bindingCollector.mappings) {
            fillColor(Stat.isIncreased) {
                scale = categorical(true to fillOnInc, false to fillOnDec)
                legend.type = LegendType.None
            }
        }
        val (colorOnInc, colorOnDec) = increase.borderLine.color to decrease.borderLine.color
        if (COLOR !in bindingCollector.mappings) {
            borderLine.color(Stat.isIncreased) {
                scale = Scale.categorical(true to colorOnInc, false to colorOnDec)
                legend.type = LegendType.None
            }
        }
        val (ltOnInc, ltOnDec) = increase.borderLine.type to decrease.borderLine.type
        if (LINE_TYPE !in bindingCollector.mappings) {
            borderLine.type(Stat.isIncreased) {
                scale = categorical(true to ltOnInc, false to ltOnDec)
                legend.type = LegendType.None
            }
        }
        val (lwOnInc, lwOnDec) = increase.borderLine.width to decrease.borderLine.width
        if (SIZE !in bindingCollector.mappings) {
            borderLine.width(Stat.isIncreased) {
                scale = Scale.categorical(true to lwOnInc, false to lwOnDec)
                legend.type = LegendType.None
            }
        }
        val (aOnInc, aOnDec) = increase.alpha to decrease.alpha
        if (ALPHA !in bindingCollector.mappings) {
            alpha(Stat.isIncreased) {
                scale = Scale.categorical(true to aOnInc, false to aOnDec)
                legend.type = LegendType.None
            }
        }
    }
}

public fun<X> DataFramePlotBuilder<*>.candlestick(
    x: Iterable<X>,
    open: Iterable<Number>,
    high: Iterable<Number>,
    low: Iterable<Number>,
    close: Iterable<Number>,
    block: CandlestickBuilder<X>.() -> Unit = {}
) {
    candlestick(this, statCandlestick(x, open, high, low, close), "x", block)
}

public fun<X> DataFramePlotBuilder<*>.candlestick(
    x: ColumnAccessor<X>,
    open: ColumnAccessor<Number>,
    high: ColumnAccessor<Number>,
    low: ColumnAccessor<Number>,
    close: ColumnAccessor<Number>,
    block: CandlestickBuilder<X>.() -> Unit = {}
) {
    candlestick(this, internalDataFrame.statCandlestick(x, open, high, low, close), x.name(), block)
}

public fun DataFramePlotBuilder<*>.candlestick(
    x: String,
    open: String,
    high: String,
    low: String,
    close: String,
    block: CandlestickBuilder<Any?>.() -> Unit = {}
) {
   candlestick(this, internalDataFrame.statCandlestick(x, open, high, low, close), x, block)
}

public fun<X> DataFramePlotBuilder<*>.candlestick(
    x: DataColumn<X>,
    open: DataColumn<Number>,
    high: DataColumn<Number>,
    low: DataColumn<Number>,
    close: DataColumn<Number>,
    block: CandlestickBuilder<X>.() -> Unit = {}
) {
    candlestick(this, statCandlestick(x, open, high, low, close), x.name(), block)
}
