package org.jetbrains.kotlinx.statistics.kandy.statplots

import kotlinx.datetime.Instant
import kotlinx.datetime.LocalDate
import kotlinx.datetime.LocalTime
import org.jetbrains.kotlinx.dataframe.DataFrame
import org.jetbrains.kotlinx.dataframe.columns.ColumnAccessor
import org.jetbrains.kotlinx.dataframe.columns.ColumnReference
import org.jetbrains.kotlinx.kandy.dsl.categorical
import org.jetbrains.kotlinx.kandy.ir.Plot
import org.jetbrains.kotlinx.kandy.ir.scale.Scale
import org.jetbrains.kotlinx.kandy.letsplot.feature.Position
import org.jetbrains.kotlinx.kandy.letsplot.feature.position
import org.jetbrains.kotlinx.kandy.letsplot.internal.*
import org.jetbrains.kotlinx.kandy.letsplot.scales.guide.LegendType
import org.jetbrains.kotlinx.kandy.letsplot.tooltips.tooltips
import org.jetbrains.kotlinx.statistics.selectors.StatXNumericSelector
import org.jetbrains.kotlinx.statistics.kandy.statplots.context.CandlestickPlotContext
import org.jetbrains.kotlinx.statistics.kandy.statplots.models.CandlestickPlot
import org.jetbrains.kotlinx.statistics.plotting.candlestick.statCandlestick
import org.jetbrains.kotlinx.statistics.selectors.StatCandlestickSelectorDSL
import org.jetbrains.kotlinx.statistics.selectors.StatCandlestickSelector
import org.jetbrains.kotlinx.statistics.selectors.StatXSelector
import org.jetbrains.kotlinx.statistics.selectors.StatXSelectorDSL
import java.time.LocalDateTime
import kotlin.reflect.typeOf



fun<X> candlestick(
    x: Iterable<X>,
    open: Iterable<Number>,
    high: Iterable<Number>,
    low: Iterable<Number>,
    close: Iterable<Number>,
    xName: String = "x",
    yName: String = "y",
): CandlestickPlot<X> {
    return CandlestickPlot(statCandlestick(x, open, high, low, close), xName, yName)
}

fun<X> DataFrame<*>.candlestick(
    x: ColumnAccessor<X>,
    open: ColumnAccessor<Number>,
    high: ColumnAccessor<Number>,
    low: ColumnAccessor<Number>,
    close: ColumnAccessor<Number>,
    yName: String = "y",
): CandlestickPlot<X> {
    return CandlestickPlot(statCandlestick(x, open, high, low, close), x.name(), yName)
}

fun DataFrame<*>.candlestick(
    xColumnName: String,
    openColumnName: String,
    highColumnName: String,
    lowColumnName: String,
    closeColumnName: String,
    yName: String = "y",
): CandlestickPlot<Any?> {
    return CandlestickPlot(statCandlestick(xColumnName, openColumnName, highColumnName, lowColumnName, closeColumnName), xColumnName, yName)
}


fun<T, X> DataFrame<T>.candlestick(
    yName: String = "y",
    selector: StatCandlestickSelectorDSL<T, X>
): CandlestickPlot<X> {
    val appliedSelector = StatCandlestickSelector<T, X>(this).apply(selector)
    return CandlestickPlot(statCandlestick(selector), appliedSelector.x.name(), yName)
}

@Suppress("INVISIBLE_MEMBER")
fun<X> CandlestickPlot<X>.configure(block: CandlestickPlotContext<X>.() -> Unit): Plot {
    return CandlestickPlotContext(this).apply {
        x(Stat.x)
        yMin(Stat.min)
        lower(Stat.lower)
        middle(List(Stat.x.size()){ null })
        upper(Stat.upper)
        yMax(Stat.max)
        x.axis.name = xName
        y.axis.name = yName
        val type = Stat.x.type()
        // TODO
        val xFormat = when (type) {
            typeOf<LocalDate>(), typeOf<LocalDateTime>(), typeOf<Instant>() -> "%d/%m/%y"
            typeOf<LocalTime>() -> "%H:%M%"
            else -> null
        }
        val formats: Map<ColumnReference<*>, String> = xFormat?.let { mapOf(Stat.x to it) } ?: mapOf()
        tooltips(title = "@x") {
            //line(axisName, Stat.x.tooltipValue(xFormat))
            Stat.x.tooltipValue(xFormat)
            line("open", Stat.open.tooltipValue(".2f"))
            line("high", Stat.max.tooltipValue(".2f"))
            line("low", Stat.min.tooltipValue(".2f"))
            line("close", Stat.close.tooltipValue(".2f"))
        }
        whiskerWidth = 0.0
        block()
        val (fillOnInc, fillOnDec) = increase.fillColor to decrease.fillColor
        if (FILL !in bindingCollector.mappings) {
            fillColor(Stat.isIncreased) {
                scale = categorical(true to fillOnInc, false to fillOnDec)
                legend.type = LegendType.None
            }
        }
        val (colorOnInc, colorOnDec) = increase.borderLine.color to decrease.borderLine.color
        if (COLOR !in bindingCollector.mappings) {
            borderLine.color(Stat.isIncreased) {
                scale = Scale.categorical(true to colorOnInc, false to colorOnDec)
                legend.type = LegendType.None
            }
        }
        val (ltOnInc, ltOnDec) = increase.borderLine.type to decrease.borderLine.type
        if (LINE_TYPE !in bindingCollector.mappings) {
            borderLine.type(Stat.isIncreased) {
                scale = categorical(true to ltOnInc, false to ltOnDec)
                legend.type = LegendType.None
            }
        }
        val (lwOnInc, lwOnDec) = increase.borderLine.width to decrease.borderLine.width
        if (SIZE !in bindingCollector.mappings) {
            borderLine.width(Stat.isIncreased) {
                scale = Scale.categorical(true to lwOnInc, false to lwOnDec)
                legend.type = LegendType.None
            }
        }
        val (aOnInc, aOnDec) = increase.alpha to decrease.alpha
        if (ALPHA !in bindingCollector.mappings) {
            alpha(Stat.isIncreased) {
                scale = Scale.categorical(true to aOnInc, false to aOnDec)
                legend.type = LegendType.None
            }
        }
    }.toPlot()
}
