package org.jetbrains.kotlinx.statistics.plotting.bin

import org.jetbrains.kotlinx.dataframe.DataColumn
import org.jetbrains.kotlinx.dataframe.DataFrame
import org.jetbrains.kotlinx.dataframe.api.*
import org.jetbrains.kotlinx.dataframe.columns.ColumnAccessor
import org.jetbrains.kotlinx.statistics.binning.*
import org.jetbrains.kotlinx.statistics.dataframe.binning.binBy
import org.jetbrains.kotlinx.statistics.selectors.StatXNumericSelector
import org.jetbrains.kotlinx.statistics.plotting.bin.models.StatBin
import org.jetbrains.kotlinx.statistics.plotting.bin.models.StatBinFrame

fun statBin(
    x: Iterable<Number>,
    weights: Iterable<Number>? = null,
    binsOption: BinsOption = BinsOption.byNumber(20),
    binsAlign: BinsAlign = BinsAlign.center(0.0),
    binsRange: Pair<Number, Number>? = null,
): StatBinFrame {
    val valuesDouble = x.map { it.toDouble() }
    val weightsDouble = weights?.map { it.toDouble() } ?: List(valuesDouble.size) { 1.0 }
    val rangeLength = valuesDouble.let {
        it.max() - it.min()
    }
    // density plot area should be == 1
    val binTotalCount = BinStatUtil.binCountAndWidth(rangeLength, binsOption).count
    val normalBinWidth = rangeLength / binTotalCount
    val densityNormalizingFactor = if (normalBinWidth > 0) 1.0 / normalBinWidth else 1.0
    val totalCount: Int = valuesDouble.size
    val totalWeight = weightsDouble.sum()
    val df = dataFrameOf("x" to valuesDouble, "weights" to weightsDouble) as DataFrame<Double>
    return StatBinFrame(dataFrameOf(df.binBy(binsOption, binsAlign, binsRange) { column("x") }.aggregate {
        val count = count()
        count into "count"
        // todo fix in df: empty df sum
        val countWeighted = if ("weights" in columnNames()) {
            sum("weights").toDouble()
        } else {
            count.toDouble()
        }
        countWeighted into "countWeighted"
        val density = count.toDouble() / totalCount * densityNormalizingFactor
        density into "density"
        val densityWeighted = countWeighted / totalWeight * densityNormalizingFactor
        densityWeighted into "densityWeighted"
    }.convert("bin").with { (it as Bin<Double>).center() }.rename("bin").into("x").asColumnGroup("Stat")) as DataFrame<StatBin>)
}

fun statBin(
    x: DataColumn<Number>,
    weights: DataColumn<Number>? = null,
    binsOption: BinsOption = BinsOption.byNumber(20),
    binsAlign: BinsAlign = BinsAlign.center(0.0),
    binsRange: Pair<Number, Number>? = null,
): StatBinFrame = statBin(x.asIterable(), weights?.asIterable(), binsOption, binsAlign, binsRange)

// todo other selecting options
fun <T> DataFrame<T>.statBin(
    binsOption: BinsOption = BinsOption.byNumber(20),
    binsAlign: BinsAlign = BinsAlign.center(0.0),
    binsRange: Pair<Number, Number>? = null,
    selector: StatXNumericSelector<T>.() -> Unit,
): StatBinFrame {
    val appliedSelector = StatXNumericSelector(this).apply(selector)
    return statBin(appliedSelector.x, appliedSelector.weights, binsOption, binsAlign, binsRange)
}

fun <T> DataFrame<T>.statBin(
    x: ColumnAccessor<Number>,
    weights: ColumnAccessor<Number>? = null,
    binsOption: BinsOption = BinsOption.byNumber(20),
    binsAlign: BinsAlign = BinsAlign.center(0.0),
    binsRange: Pair<Number, Number>? = null,
): StatBinFrame {
    return statBin(
        get(x.toColumnAccessor()),
        weights?.let { get(it) },
        binsOption,
        binsAlign,
        binsRange
    )
}

fun <T> DataFrame<T>.statBin(
    x: String,
    weights: String? = null,
    binsOption: BinsOption = BinsOption.byNumber(20),
    binsAlign: BinsAlign = BinsAlign.center(0.0),
    binsRange: Pair<Number, Number>? = null,
): StatBinFrame {
    return statBin(
        column(x),
        weights?.let { column(it) },
        binsOption,
        binsAlign,
        binsRange
    )
}
