package org.jetbrains.kotlinx.statistics.plotting.boxplot

import org.jetbrains.kotlinx.dataframe.DataColumn
import org.jetbrains.kotlinx.dataframe.DataFrame
import org.jetbrains.kotlinx.dataframe.api.*
import org.jetbrains.kotlinx.dataframe.columns.ColumnAccessor
import org.jetbrains.kotlinx.statistics.math3.stat.descriptive.DescriptiveStatistics
import org.jetbrains.kotlinx.statistics.plotting.boxplot.models.*
import org.jetbrains.kotlinx.statistics.selectors.StatXYNumSelector
import org.jetbrains.kotlinx.statistics.selectors.StatXYNumSelectorDSL

private fun<T> boxplot(x: T, y: Iterable<Number>, whiskerIQRRatio: Double): BoxplotStatistics<T> {
    val yDouble = y.map {it.toDouble()}
    val stats = DescriptiveStatistics(y.map {it.toDouble()}.toDoubleArray())

    val middle = stats.getPercentile(50.0)
    val lowerHinge = stats.getPercentile(25.0)
    val upperHinge = stats.getPercentile(75.0)
    val IQR = upperHinge - lowerHinge
    val lowerFence = lowerHinge - IQR * whiskerIQRRatio
    val upperFence = upperHinge + IQR * whiskerIQRRatio

    var lowerWhisker = lowerFence
    var upperWhisker = upperFence
    if (lowerFence.isFinite() && upperFence.isFinite()) {
        val boxed = yDouble.filter { y -> y in lowerFence..upperFence }
            lowerWhisker = boxed.min()
            upperWhisker = boxed.max()
    }

    return BoxplotStatistics(x, lowerWhisker, lowerHinge, middle, upperHinge, upperWhisker)
}

fun statBoxplot(y: Iterable<Number>, whiskerIQRRatio: Double = 1.5, name: String = ""): StatBoxplotFrame<String> {
    val box = boxplot(name, y, whiskerIQRRatio)
    return StatBoxplotFrame.fromList(listOf( box))
}

fun statBoxplot(y: DataColumn<Number>, whiskerIQRRatio: Double = 1.5): StatBoxplotFrame<String> {
   return statBoxplot(y.asIterable(), whiskerIQRRatio, y.name())
}

fun <T> DataFrame<T>.statBoxplot(y: ColumnAccessor<Number>, whiskerIQRRatio: Double = 1.5): StatBoxplotFrame<String> {
    return statBoxplot(get(y).asIterable(), whiskerIQRRatio, y.name())
}

fun <T> DataFrame<T>.statBoxplot(y: String, whiskerIQRRatio: Double = 1.5): StatBoxplotFrame<String> {
    return statBoxplot(column(y), whiskerIQRRatio)
}

fun <T, X> DataFrame<T>.statBoxplot(x: ColumnAccessor<X>, y:ColumnAccessor<Number>, whiskerIQRRatio: Double = 1.5): StatBoxplotFrame<X> {
    val boxes = groupBy(x).map {
        val xs = key[x]
        val ys = it.group[y].toList()
        boxplot(xs, ys, whiskerIQRRatio)
    }
    return StatBoxplotFrame.fromList(boxes)
}

fun <T> DataFrame<T>.statBoxplot(x: String, y:String, whiskerIQRRatio: Double = 1.5): StatBoxplotFrame<Any?> {
    return statBoxplot(column(x), column(y), whiskerIQRRatio)
}

fun <T, X> DataFrame<T>.statBoxplot(whiskerIQRRatio: Double = 1.5, selector: StatXYNumSelectorDSL<T, X>): StatBoxplotFrame<Any?> {
   val appliedSelector = StatXYNumSelector<T,X>(this).apply(selector)
    return statBoxplot(appliedSelector.x, appliedSelector.y, whiskerIQRRatio)
}

fun<T> statBoxplot(x: DataColumn<T>, y: DataColumn<Number>,whiskerIQRRatio: Double = 1.5): StatBoxplotFrame<T> {
    return dataFrameOf(x, y).statBoxplot(x.toColumnAccessor(), y.toColumnAccessor(), whiskerIQRRatio)
}
