package org.jetbrains.kotlinx.statistics.plotting.density

import org.jetbrains.kotlinx.dataframe.DataColumn
import org.jetbrains.kotlinx.dataframe.DataFrame
import org.jetbrains.kotlinx.dataframe.api.*
import org.jetbrains.kotlinx.dataframe.columns.ColumnAccessor
import org.jetbrains.kotlinx.dataframe.columns.ColumnGroup
import org.jetbrains.kotlinx.statistics.plotting.density.models.DensityStatistics
import org.jetbrains.kotlinx.statistics.plotting.density.models.StatDensity
import org.jetbrains.kotlinx.statistics.plotting.density.models.StatDensityFrame
import org.jetbrains.kotlinx.statistics.selectors.StatXNumericSelector

fun statDensity(
    x: Iterable<Number>,
    weights: Iterable<Number>? = null,
    n: Int = 512,
    trim: Boolean = false,
    adjust: Double = 1.0,
    kernel: Kernel = Kernel.GAUSSIAN,
    fullScanMax: Int = 5000,
    bandWidth: BandWidth = BandWidth.Method.NRD0,
    range: Pair<Number, Number>? = null,
): StatDensityFrame {
    val xList = x.map { it.toDouble() }
    val weightsList = weights?.map { it.toDouble() } ?: List(xList.size) {1.0}

    val (xSorted, weightsSorted) = xList.zip(weightsList).sortedBy { it.first }.unzip()

    val rangeX = if (trim && range != null) {
        range.let { it.first.toDouble() to it.second.toDouble() }
    } else {
        xSorted.min() to xSorted.max()
    }
    val statX = DensityStatUtil.createStepValues(rangeX, n)
    val statDensity = ArrayList<Double>()
    val statDensityWeighted = ArrayList<Double>()
    val statCount = ArrayList<Double>()
    val statCountWeighted = ArrayList<Double>()
    val statScaled = ArrayList<Double>()
    val statScaledWeighted = ArrayList<Double>()
    val statDensityFunction = DensityStatUtil.densityFunction(
        xSorted, weightsSorted,
        bandWidth, adjust, kernel, fullScanMax
    )

    val nTotal = weightsSorted.sum()
    for (x in statX) {
        val d = statDensityFunction(x, false)
        val dW = statDensityFunction(x, true)
        statCount.add(d)
        statDensity.add(d / nTotal)
        statCountWeighted.add(dW)
        statDensityWeighted.add(dW / nTotal)
    }

    val maxm = statCount.maxOrNull()!!
    for (d in statCount) {
        statScaled.add(d / maxm)
    }

    val maxmW = statCountWeighted.maxOrNull()!!
    for (d in statCountWeighted) {
        statScaledWeighted.add(d / maxmW)
    }

    // val statQuantile = DensityStatUtil.calculateStatQuantile(statX, statCount, quantiles)

    return StatDensityFrame(
        dataFrameOf(
            dataFrameOf(
                "x" to statX,
                "density" to statDensity,
                "densityWeighted" to statDensityWeighted,
                "scaled" to statScaled,
                "scaledWeighted" to statScaledWeighted,
               // "count" to statCount,
                //"countWeighted" to statCountWeighted,
            ).asColumnGroup("Stat") as ColumnGroup<DensityStatistics>
        ) as DataFrame<StatDensity>
    )
}

fun statDensity(
    x: DataColumn<Number>,
    weights: DataColumn<Number>? = null,
    n: Int = 512,
    trim: Boolean = false,
    adjust: Double = 1.0,
    kernel: Kernel = Kernel.GAUSSIAN,
    fullScanMax: Int = 5000,
    bandWidth: BandWidth = BandWidth.Method.NRD0,
    range: Pair<Number, Number>? = null,
): StatDensityFrame = statDensity(
    x.asIterable(),
    weights?.asIterable(),
    n, trim, adjust, kernel, fullScanMax, bandWidth, range
)

// todo other selecting options
fun <T> DataFrame<T>.statDensity(
    n: Int = 512,
    trim: Boolean = false,
    adjust: Double = 1.0,
    kernel: Kernel = Kernel.GAUSSIAN,
    fullScanMax: Int = 5000,
    bandWidth: BandWidth = BandWidth.Method.NRD0,
    range: Pair<Number, Number>? = null,
    selector: StatXNumericSelector<T>.() -> Unit,
): StatDensityFrame {
    val appliedSelector = StatXNumericSelector(this).apply(selector)
    return statDensity(appliedSelector.x, appliedSelector.weights, n, trim, adjust, kernel, fullScanMax, bandWidth, range)
}

fun <T> DataFrame<T>.statDensity(
    x: ColumnAccessor<Number>,
    weights: ColumnAccessor<Number>? = null,
    n: Int = 512,
    trim: Boolean = false,
    adjust: Double = 1.0,
    kernel: Kernel = Kernel.GAUSSIAN,
    fullScanMax: Int = 5000,
    bandWidth: BandWidth = BandWidth.Method.NRD0,
    range: Pair<Number, Number>? = null,
): StatDensityFrame {
    return statDensity(
        get(x.toColumnAccessor()),
        weights?.let { get(it) },
        n, trim, adjust, kernel, fullScanMax, bandWidth, range
    )
}

fun <T> DataFrame<T>.statDensity(
    x: String,
    weights: String? = null,
    n: Int = 512,
    trim: Boolean = false,
    adjust: Double = 1.0,
    kernel: Kernel = Kernel.GAUSSIAN,
    fullScanMax: Int = 5000,
    bandWidth: BandWidth = BandWidth.Method.NRD0,
    range: Pair<Number, Number>? = null,
): StatDensityFrame {
    return statDensity(
        column(x),
        weights?.let { column(it) },
       n, trim, adjust, kernel, fullScanMax, bandWidth, range
    )
}
