package org.jetbrains.kotlinx.statistics.plotting.smooth

import org.jetbrains.kotlinx.dataframe.DataColumn
import org.jetbrains.kotlinx.dataframe.DataFrame
import org.jetbrains.kotlinx.dataframe.api.asColumnGroup
import org.jetbrains.kotlinx.dataframe.api.asIterable
import org.jetbrains.kotlinx.dataframe.api.column
import org.jetbrains.kotlinx.dataframe.api.dataFrameOf
import org.jetbrains.kotlinx.dataframe.columns.ColumnAccessor
import org.jetbrains.kotlinx.dataframe.columns.ColumnGroup
import org.jetbrains.kotlinx.statistics.plotting.smooth.models.SmoothStatistics
import org.jetbrains.kotlinx.statistics.plotting.smooth.models.StatSmooth
import org.jetbrains.kotlinx.statistics.plotting.smooth.models.StatSmoothFrame
import org.jetbrains.kotlinx.statistics.regression.LinearRegression
import org.jetbrains.kotlinx.statistics.regression.LocalPolynomialRegression
import org.jetbrains.kotlinx.statistics.regression.PolynomialRegression
import org.jetbrains.kotlinx.statistics.selectors.StatXNumYNumSelector
import org.jetbrains.kotlinx.statistics.selectors.StatXNumYNumSelectorDSL
import kotlin.random.Random

/*private fun extendSampleIndices(initialSize: Int, extendedSize: Int): List<Int> {
    val buffer = IntArray(extendedSize)
    for (i in 0 until initialSize) {
        buffer[i] = i
    }
    for (i in initialSize until extendedSize) {
        buffer[i] = Random.nextInt(initialSize)
    }
    return buffer.sorted()
}

private fun extendSample(sample: List<Double>, indices: List<Int>): List<Double> {
    return List(indices.size) { sample[indices[it]] }
}*/

private fun  List<Double>.takeIndices(indices: List<Int>): List<Double> {
    return indices.map { this[it] }
}

fun statSmooth(
    x: Iterable<Number>,
    y: Iterable<Number>,
    method: SmoothMethod = SmoothMethod.LOESS(),
    smootherPointCount: Int = 100
): StatSmoothFrame {
    val xDouble = x.map { it.toDouble() }
    val yDouble = y.map { it.toDouble() }
    val regression = when (method) {
        is SmoothMethod.LOESS -> {
            val criticalSize = method.loessCriticalSize
            if (xDouble.size >= criticalSize) {
                val indices = xDouble.indices.shuffled().take(criticalSize)
                LocalPolynomialRegression.fit(
                    xDouble.takeIndices(indices),
                    yDouble.takeIndices(indices),
                    method.confidenceLevel, method.span
                )
            } else {
                LocalPolynomialRegression.fit(xDouble, yDouble, method.confidenceLevel, method.span)
            }
        }

        is SmoothMethod.Linear -> LinearRegression.fit(xDouble, yDouble, method.confidenceLevel)
        is SmoothMethod.Polynomial -> PolynomialRegression.fit(xDouble, yDouble, method.confidenceLevel, method.degree)
    }!! // TODO

    val min = xDouble.min()
    val max = xDouble.max()
    val stepX = (max - min) / (smootherPointCount - 1)

    val statX = ArrayList<Double>()
    val statY = ArrayList<Double>()
    val statMinY = ArrayList<Double>()
    val statMaxY = ArrayList<Double>()
    val statSE = ArrayList<Double>()

    for (i in 0 until smootherPointCount) {
        val x_i = min + i * stepX
        val eval = regression.evalX(x_i.coerceIn(min, max))
        statX.add(x_i)
        statY.add(eval.y)
        statMinY.add(eval.ymin)
        statMaxY.add(eval.ymax)
        statSE.add(eval.se)
    }
    return StatSmoothFrame(
        dataFrameOf(
            dataFrameOf(
                "x" to statX,
                "y" to statY,
                "yMin" to statMinY,
                "yMax" to statMaxY,
                "se" to statSE
            ).asColumnGroup("Stat")
                    as ColumnGroup<SmoothStatistics>
        ) as DataFrame<StatSmooth>
    )
}

fun statSmooth(
    x: DataColumn<Number>,
    y: DataColumn<Number>,
    method: SmoothMethod = SmoothMethod.LOESS(),
    smootherPointCount: Int = 100
): StatSmoothFrame {
    return statSmooth(x.asIterable(), y.asIterable(), method, smootherPointCount)
}

fun <T> DataFrame<T>.statSmooth(
    method: SmoothMethod = SmoothMethod.LOESS(),
    smootherPointCount: Int = 100,
    selector: StatXNumYNumSelectorDSL<T>,
): StatSmoothFrame {
    val appliedSelector = StatXNumYNumSelector<T>(this).apply(selector)
    return statSmooth(appliedSelector.x, appliedSelector.y, method, smootherPointCount)
}

fun <T> DataFrame<T>.statSmooth(
    x: ColumnAccessor<Number>,
    y: ColumnAccessor<Number>,
    method: SmoothMethod = SmoothMethod.LOESS(),
    smootherPointCount: Int = 100,
): StatSmoothFrame {
    return statSmooth(get(x), get(y), method, smootherPointCount)
}

fun <T> DataFrame<T>.statSmooth(
    x: String,
    y: String,
    method: SmoothMethod = SmoothMethod.LOESS(),
    smootherPointCount: Int = 100,
): StatSmoothFrame {
    return statSmooth(column(x), column(y), method, smootherPointCount)
}
